/*
 *  This is the LCDproc driver for CrystalFontz LCD using Packet protocol.
 *  It support the CrystalFontz 633 USB/Serial, the 631 USB and the 635 USB
 *  (get yours from http://crystalfontz.com)
 *
 *  Applicable Data Sheets
 *  http://www.crystalfontz.com/products/631/CFA-631_v1.0.pdf
 *  http://www.crystalfontz.com/products/633/CFA_633_0_6.PDF
 *  http://www.crystalfontz.com/products/635/CFA_635_1_0.pdf
 *
 *  Copyright (C) 2002 David GLAUDE
 *  Portions Copyright (C) 2005 Peter Marschall
 *  Portions Copyright (C) 2005 Nicolas Croiset <ncroiset@vdldiffusion.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */

/*
 * Driver status
 * 04/04/2002: Working driver
 * 05/06/2002: Reading of return value
 * 02/09/2002: KeyPad handling and return string
 * 03/09/2002: New icon incorporated
 * 27/01/2003: Adapted for CFontz 631
 * 10/04/2004: added temp and fan functions
 *   /07/2005: packet que, hw probe, LCD lines
 * 16/05/2005: Adapted for CFontz 635
 * 21/12/2005: data logfile with proc, lm-sensors
 * THINGS NOT DONE:
 * + No support for multiple instance (require private structure)
 * + No cache of custom char usage (like in MtxOrb)
 *
 * THINGS DONE:
 * + Stopping the live reporting (of temperature)
 * + Stopping the reporting of temp and fan (is it necessary after reboot)
 * + Use of library for hbar and vbar (good but library could be better)
 * + Support for keypad (Using a KeyRing)
 * + BigNum (for CF635 only: it is a 4-line display)
 * + Output support (LED control on a CF635 only)
 *
 * THINGS TO DO:
 * + Make the caching at least for heartbeat icon
 * + Create and use the library (for custom char handling)
 *
*/
/*
 * ===========================================================================
 * 633 Test code for linux.
 * Copyright 2002, David GLAUDE
 * Partial copyright
 * 2001 Crystalfontz America, Inc. 
 * brent@crystalfontz.com
 * Written by Brent A. Crosby 
 * www.crystalfontz.com
 * ===========================================================================
 */
/*
 *  This is a multithreaded que based fan and temperature enabed
 *  LCDproc driver for CrystalFontz 631, 633, and 635 LCDs
 *  Copyright (C) 2005 Stewart Putnam
 *  Portions Copyright (c) 2002, Joris Robijn
 *
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif


#define TRY_AGAIN 0
#define GOOD_MSG 1
#define GIVE_UP 2

#define DEBUG 1

#include <termios.h>
#include <time.h>
#include <dirent.h>
#include <limits.h>
#include <dlfcn.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <signal.h>
#include <wait.h>
// appears uneeded, which manpage said to include ?
#include <sched.h>

#include <sys/types.h>
#include <sys/time.h>
#ifdef TIMINGREPORT
#ifdef LIVE_TIMING
#include <signal.h>
#endif
#endif

#include "lcd.h"
#include "CFontz633.h"
#include "report.h"
#include "lcd_lib.h"
#include "CFontz-charmap.h"
#include "timing.h"
#include <sensors/sensors.h>
#include <sensors/chips.h>
#include <sensors/error.h>
#include "../../clients/lcdproc/machine.h"
#include "../../shared/sockets.h"
#ifdef CLIENT_CHILD
#include "../screen.h"
#include "../widget.h"
#include "../clients.h"
#endif

/*********************************************************
 * create any custom lines as globals here
 * and increase MAXLINES above accordingly
 * and create ptr below within #ifdef PROTOCOL _ CHEATING
 * ******************************************************/
#ifdef MY_CUSTOM_LINE
char *watts;
int wattsptr;
float watts_est=0;
#endif // MY_CUSTOM_LINE

#ifdef CLIENT_CHILD
#ifdef PROTOCOL_CHEATING
int iowritestohere[MAXLINES];
#endif // PROTOCOL_CHEATING

int order[MAXLINES];  // internal use for ordering created lines
int displinenum[MAXLINES];  // lookup for custom line updating;
char *orderline[MAXLINES];
unsigned char t_or_f_num[MAXLINES]; 
#endif // CLIENT_CHILD

#define CF633_KEY_UP      1
#define CF633_KEY_DOWN      2
#define CF633_KEY_LEFT      3
#define CF633_KEY_RIGHT      4
#define CF633_KEY_ENTER      5
#define CF633_KEY_ESCAPE    6
#define CF633_KEY_UP_RELEASE    7
#define CF633_KEY_DOWN_RELEASE    8
#define CF633_KEY_LEFT_RELEASE    9
#define CF633_KEY_RIGHT_RELEASE    10
#define CF633_KEY_ENTER_RELEASE    11
#define CF633_KEY_ESCAPE_RELEASE  12
#define CF631_KEY_UL_PRESS    13
#define CF631_KEY_UR_PRESS    14
#define CF631_KEY_LL_PRESS    15
#define CF631_KEY_LR_PRESS     16
#define CF631_KEY_UL_RELEASE    17
#define CF631_KEY_UR_RELEASE    18
#define CF631_KEY_LL_RELEASE    19
#define CF631_KEY_LR_RELEASE    20

#define CELLWIDTH  DEFAULT_CELL_WIDTH
#define CELLHEIGHT  DEFAULT_CELL_HEIGHT


/* LEDs dispatch */
#define CF635_NUM_LEDs  8

/* Vars for the server core */
MODULE_EXPORT char *api_version = API_VERSION;
MODULE_EXPORT int stay_in_foreground = 0;
MODULE_EXPORT int supports_multiple = 0;
MODULE_EXPORT char *symbol_prefix = "CFontz633_";

const char *pk_nm[MAX_COMMAND+2]= {"CF633_Ping_Command",
  "CF633_Get_Hardware_And_Firmware_Version","CF633_Write_User_Flash_Area",
  "CF633_Read_User_Flash_Area","CF633_Store_Current_State_As_Boot_State",
  "CF633_Reboot","CF633_Clear_LCD_Screen",
  "CF633_Set_LCD_Contents_Line_One","CF633_Set_LCD_Contents_Line_Two",
  "CF633_Set_LCD_Special_Character_Data",
  "CF633_Read_Eight_Bytes_of_LCD_Memory","CF633_Set_LCD_Cursor_Position",
  "CF633_Set_LCD_Cursor_Style","CF633_Set_LCD_Contrast",
  "CF633_Set_LCD_And_Keypad_Backlight","CF633_Query_Last_Fan_Pulse_Count",
  "CF633_Set_Up_Fan_Reporting","CF633_Set_Fan_Power",
  "CF633_Read_DOW_Device_Information","CF633_Set_Up_Temperature_Reporting"
  ,"CF633_Arbitrary_DOW_Transaction",
  "CF633_Set_Up_Live_Fan_or_Temperature_Display",
  "CF633_Send_Command_Directly_to_the_LCD_Controller",
  "CF633_Configure_Key_Reporting", "CF633_Read_Keypad_Polled_Mode",
  "CF633_Set_Fan_Power_Failsafe","CF633_Set_Fan_Tachometer_Glitch_Filter",
  "CF633_Query_Fan_Power_and_Failsafe_Mask",
  "CF633_Set_ATX_Power_Switch_Functionality", 
  "CF633_Enable_Disable_and_Reset_the_Watchdog",
  "CF633_Read_Reporting_and_Status", "CF633_Send_Data_to_LCD",
  "CF633_Key_Legends", "CF633_Set_Baud_Rate",
  "CF633_Set_GPIO_Pin", "CF633_Read_GPIO_Pin",
  "CF633_Dummy_Unused" };

/* Internal functions */
/* static void CFontz633_linewrap (int on); */
/* static void CFontz633_autoscroll (int on);  */
static void CFontz633_hidecursor (Driver *drvthis);
static void CFontz633_reboot (Driver *drvthis, struct timeval *sent_reboot);
static void CFontz633_rbooted (Driver *drvthis, struct timeval *sent_reboot);
static void CFontz633_init_vbar (Driver *drvthis);
static void CFontz633_init_hbar (Driver *drvthis);
static void CFontz633_no_live_report(Driver *drvthis);
static void CFontz633_hardware_clear (Driver *drvthis);
static void CFontz633_no_temp_report(Driver *drvthis);
static void CFontz633_no_fan_report(Driver *drvthis);
void CFontz633_await_packet (Driver *drvthis, COMMAND_PACKET *out, int send);
static void set_fan_fsg (Driver* drvthis);
static void CFontz633_set_temp_report(Driver* drvthis, unsigned char out[4]);
static int CFontz633_packet_neq(COMMAND_PACKET *packet1, COMMAND_PACKET *packet2);
int CFontz633_ds18_init (Driver* drvthis);
int CFontz633_fan_init (Driver* drvthis);
void query_hw_ver (Driver *drvthis);
#ifdef STACK_USAGE
int check_clone_stack_usage (Driver *drvthis, char *stack_name, void *start, int block_size);
#endif

void send_packet(Driver *drvthis, COMMAND_PACKET *out);
void commit_packet(Driver* drvthis,  COMMAND_PACKET *packet);
double time_diff (Driver *drvthis, struct timeval past, struct timeval now);
int           check_for_packet(Driver *drvthis, unsigned char expected_length);
void          show_packet (Driver* drvthis, int rep_level, COMMAND_PACKET *packet);

#ifdef TIMING_REPORT
void timing_report(Driver *drvthis);
void clear_timing (Driver *drvthis);
#ifdef LIVE_TIMING
static void install_signal(Driver *drvthis);
static void live_timing(int signal_num);
static void live_timing_reset(int signal_num);
#endif
#endif

void spawn_clones (Driver *drvthis);

void CFontz633_data_col (Driver *drvthis);
int fan_loop(Driver * drvthis);
static void textlog (Driver *drvthis); 
static void purge_que (Driver *drvthis); 
static void manage_fans  (Driver *drvthis);
void set_fan_rpt (Driver *drvthis);
void set_fan_pow (Driver *drvthis);

int listen_packets (Driver * drvthis);
static void sensors_packet(Driver *drvthis);

int datalog_cli (Driver * drvthis);
#ifdef DO_ULT
void store_load(Driver *drvthis);
static int sort_jiffpid (void *a, void *b);
int machine_get_jiffies(Driver *drvthis, LinkedList *procs);
void store_tops(Driver *drvthis);
static int sort_jiffies (void *a, void *b);
#endif // DO_ULT
#ifdef DO_LM
void store_lm(Driver *drvthis);
int find_build_req_lm(Driver *drvthis);
#endif //DO_LM
#ifdef DO_HDD
static void hddtemp_names(Driver *drvthis);
static void hddtemp_temps(Driver *drvthis);
#endif // DO_HDD
static void client_loop(Driver * drvthis);

#ifdef MY_CUSTOM_LINE
static void update_wattsline (Driver *drvthis, unsigned char unused);
#endif // MY_CUSTOM_LINE

#ifdef CLIENT_CHILD
// glabals used by cli child
int lines=0; // # of lines actually created
int cf_dsk;
int lcd_wid = 0;
int lcd_hgt = 0;
int lcd_cellwid = 0;
int lcd_cellhgt = 0;
// updateline is an array of pointers
// not a function
static void (*updateline[MAXLINES])(Driver *, unsigned char);

static void set_up_templine (Driver *drvthis, unsigned char temp);
static void set_up_fanline (Driver *drvthis, unsigned char fan);
#ifdef DO_ULT
static void set_up_topnameline (Driver *drvthis, unsigned char topn);
#endif // DO_ULT
#ifdef DO_LM
static void set_up_lmline (Driver *drvthis, unsigned char lmn);
#endif //DO_LM
#ifdef DO_HDD
static void set_up_hddline (Driver *drvthis, unsigned char lmn);
#endif // DO_HDD

int wait_success(Driver * drvthis, int options, char **other);
int cli_scrn(Driver * drvthis); 
int read_connect_string(Driver * drvthis);
int split( char * str, char delim, char * parts[], int maxparts ); 
#endif // CLIENT_CHILD

void          EmptyKeyRing(KeyRing *kr);
int           AddKeyToKeyRing(KeyRing *kr, unsigned char key);
unsigned char GetKeyFromKeyRing(KeyRing *kr);
void          EmptyReceiveBuffer(ReceiveBuffer *rb);
void          SyncReceiveBuffer(ReceiveBuffer *rb, int fd, unsigned int number);
int           BytesAvail(ReceiveBuffer *rb);
unsigned char GetByte(ReceiveBuffer *rb);
int           PeekBytesAvail(ReceiveBuffer *rb);
void          SyncPeekPointer(ReceiveBuffer *rb);
void          AcceptPeekedData(ReceiveBuffer *rb);
unsigned char PeekByte(ReceiveBuffer *rb);
int  get_crc(char *buf, int len, int seed);

#ifdef TIMING_REPORT
#ifdef LIVE_TIMING 
// bad form, but how to pass a Driver * to 
// a function that cannot be passed specific arguments ?
Driver *driver_pointer_global;
#endif
#endif

/*
 * Opens com port and sets baud correctly...
 */
MODULE_EXPORT int
CFontz633_init (Driver *drvthis)
{
  struct termios portset;
  int tmp, w, h;
  int reboot = 0;
  int speed = DEFAULT_SPEED;
  char size[200] = DEFAULT_SIZE_CF633;
  char *default_size = DEFAULT_SIZE_CF633;

  PrivateData *p;

  /* Allocate and store private data */
  p = (PrivateData *) calloc(1, sizeof(PrivateData));
  if (p == NULL)
    return -1;
  if (drvthis->store_private_ptr(drvthis, p))
    return -1;

  /* Initialize the PrivateData structure */
  p->cellwidth = DEFAULT_CELL_WIDTH;
  p->cellheight = DEFAULT_CELL_HEIGHT;
  p->ccmode = standard;
  p->LEDstate = 0xFFFF;

  p->listen_stack=NULL;
  p->fan_stack=NULL;
  p->init_completed=0;
  p->sending=0;
  p->q_depth=0;
  p->q_recover=0;
  p->deepies=0;
  p->sendq[0].command=CF633_Dummy_Unused;
  p->endq=0;

  report(RPT_INFO, "%s init(%p)", __FUNCTION__, drvthis );

  EmptyKeyRing(&p->keyring);
  EmptyReceiveBuffer(&p->receivebuffer);

  /* Which device should be used */
  strncpy(p->device, drvthis->config_get_string (drvthis->name, "Device", 0, DEFAULT_DEVICE), sizeof(p->device));
  p->device[sizeof(p->device)-1] = '\0';
  report(RPT_INFO,"%s Device (in config) is '%s'", __FUNCTION__, p->device);

  /* Which speed ? CF633 support 19200 only, CF631 & CF635 USB use 115200. */
  tmp = drvthis->config_get_int (drvthis->name, "Speed", 0, DEFAULT_SPEED);
  report(RPT_INFO,"%s Speed (in config) is '%d'", __FUNCTION__, tmp);
  if (tmp == 19200) speed = B19200;
  else if (tmp == 115200) speed = B115200;
  else {
    report(RPT_WARNING, "%s Speed must be 19200 or 11500. Using default value %d.",
        DEFAULT_SPEED, __FUNCTION__);
    speed = DEFAULT_SPEED;
  }

  /* Am I USB or not? */
  p->usb = drvthis->config_get_bool(drvthis->name, "USB", 0, 0);
  if (p->usb)
    report(RPT_INFO, "%s USB is indicated (in config).",  __FUNCTION__);

  /* Set up io port correctly, and open it... */
  report(RPT_DEBUG, "%s Opening device: %s",  __FUNCTION__, p->device);
  p->fd = open(p->device, (p->usb) ? (O_RDWR | O_NOCTTY) : (O_RDWR | O_NOCTTY | O_NDELAY));
  int errno_sav=errno;
  if (p->fd == -1) {
    report(RPT_CRIT, "%s open (%s) failed (%s)",
        __FUNCTION__, p->device, strerror (errno_sav));
    return -1;
  }

  tmp=tcgetattr (p->fd, &portset);
  errno_sav=errno;
  if (tmp==-1) {
    report(RPT_CRIT, "%s tcgetattr (p->fd, &portset) failed (%s)",
         __FUNCTION__, strerror (errno_sav));
    return -1;
  }

  /* We use RAW mode */
  if (p->usb) {
    // The USB way
    portset.c_iflag &= ~( IGNBRK | BRKINT | PARMRK | ISTRIP
          | INLCR | IGNCR | ICRNL | IXON );
    portset.c_oflag &= ~OPOST;
    portset.c_lflag &= ~( ECHO | ECHONL | ICANON | ISIG | IEXTEN );
    portset.c_cflag &= ~( CSIZE | PARENB | CRTSCTS );
    portset.c_cflag |= CS8 | CREAD | CLOCAL ;
    portset.c_cc[VMIN] = 0;
    portset.c_cc[VTIME] = 0;
  } else {
#ifdef HAVE_CFMAKERAW
    /* The easy way */
    cfmakeraw(&portset);
#else
    /* The hard way */
    portset.c_iflag &= ~( IGNBRK | BRKINT | PARMRK | ISTRIP
                | INLCR | IGNCR | ICRNL | IXON );
    portset.c_oflag &= ~OPOST;
    portset.c_lflag &= ~( ECHO | ECHONL | ICANON | ISIG | IEXTEN );
    portset.c_cflag &= ~( CSIZE | PARENB | CRTSCTS );
    portset.c_cflag |= CS8 | CREAD | CLOCAL ;
#endif
  }

  /* Set port speed */
  tmp=cfsetospeed (&portset, speed);
  errno_sav=errno;
  if (tmp==-1) {
    report(RPT_CRIT, "%s cfsetospeed (&portset, speed) failed (%s)",
         __FUNCTION__, strerror (errno_sav));
    return -1;
  }
  tmp=cfsetispeed (&portset, B0);
  errno_sav=errno;
  if (tmp==-1) {
    report(RPT_CRIT, "%s cfsetispeed (&portset, B0) failed (%s)",
         __FUNCTION__, strerror (errno_sav));
    return -1;
  }

  /* Do it... */
  tmp=tcsetattr (p->fd, TCSANOW, &portset);
  errno_sav=errno;
  if (tmp==-1) {
    report(RPT_CRIT, "%s tcgetattr (p->fd, TCSANOW, &portset) failed (%s)",
         __FUNCTION__, strerror (errno_sav));
    return -1;
  }
  
  /* Turn off reports & any future attempts if not hardware supported */
  CFontz633_no_temp_report(drvthis);
  
  /* needs simple flag to skip fan_init & future attempts if not hardware supported */
  CFontz633_no_fan_report(drvthis);
  
  /* Reboot display? takes 2 seconds! get it done first */
  reboot = drvthis->config_get_bool(drvthis->name, "Reboot", 0, 0);
  if (reboot) {
    CFontz633_reboot (drvthis, &p->tdid_tlog);
  }

  /* Which contrast */
  tmp = drvthis->config_get_int (drvthis->name, "Contrast", 0, DEFAULT_CONTRAST);
  report(RPT_INFO, "%s Contrast (in config) is '%d'", __FUNCTION__, tmp);
  if ((tmp < 0) || (tmp > 1000)) {
    report(RPT_WARNING, "%s Contrast must be between 0 and 1000. Using default value.",  __FUNCTION__);
    tmp = DEFAULT_CONTRAST;
  }
  p->contrast = tmp;

  /* Which backlight brightness */
  tmp = drvthis->config_get_int (drvthis->name, "Brightness", 0, DEFAULT_BRIGHTNESS);
  report(RPT_INFO,"CF_init: Brightness (in config) is '%d'", tmp);
  if ((tmp < 0) || (tmp > 1000)) {
    report(RPT_WARNING, "%s Brightness must be between 0 and 1000. Using default value.",  __FUNCTION__);
    tmp = DEFAULT_BRIGHTNESS;
  }
  p->brightness = tmp;

  /* Which backlight-off "brightness" */
  tmp = drvthis->config_get_int (drvthis->name, "OffBrightness", 0, DEFAULT_OFFBRIGHTNESS);
  report(RPT_INFO,"%s OffBrightness (in config) is '%d'", __FUNCTION__, tmp);
  if ((tmp < 0) || (tmp > 1000)) {
    report(RPT_WARNING, "%s OffBrightness must be between 0 and 1000. Using default value.",  __FUNCTION__);
    tmp = DEFAULT_OFFBRIGHTNESS;
  }
  p->offbrightness = tmp;

  /* New firmware version?
   * I will try to behave differently for firmware 0.6 or above.
   * Currently this is not in use.
   */
  p->newfirmware = drvthis->config_get_bool(drvthis->name, "NewFirmware", 0, 0);

  /* Which model is it (CF633, CF631 or CF635)? */
  tmp = drvthis->config_get_int (drvthis->name, "Model", 0, 0);
  report(RPT_INFO,"%s Model (in config) is '%d'",  __FUNCTION__, tmp);
  p->model = tmp;
  
  if (reboot) {
    CFontz633_rbooted (drvthis, &p->tdid_tlog);
  }
  
  /* if this returns error, can I ignore & deconfigure all fan / temp funcs */
  CFontz633_no_live_report(drvthis);

  /* Do temperatures or no from config*/
  p->tempcount = drvthis->config_get_int(drvthis->name, "Temperatures", 0, DEFAULT_TEMP_REPORT);

  /* with a viable query_hw_ver what can be discarded ? */
   query_hw_ver (drvthis);
  /* observation: power on module get 1 hw querry
      subsequent querries return gibberish
      cycle lcd power -> 1 more querry
      subsequent querries return gibberish
      ...
      2005 07 16 swp  
      hmmm... not true 631 ...  more time after reboot() helps ?
      lots more time? 30 sec+
      A:some other packet (fans ?) mistaken as hw querry
      */
  
  /* contrast hints */
  if ( (p->model==635) && ((p->contrast < 300) ||  (p->contrast > 700)) ){
    // 65-126 hw native by 635.pdf =~ 254 -500 ; 300 - 700 by experience
    report(RPT_INFO, "%s Contrast %d outside suggested normal range 300 - 700.", 
         __FUNCTION__, p->contrast);
  }
  else if ( ((p->model==631)||(p->model==633) ) && ((p->contrast < 50) ||  (p->contrast > 500)) ){
      // 0-29 hw native by 633.pdf =~ 0-600 on 1-1000 config scale 
    report(RPT_INFO, "%s Contrast %d outside suggested normal range 50 - 500.", 
         __FUNCTION__, p->contrast);
  }

  /* only que fan management & let everything else fly */
  p->que_fans_only=drvthis->config_get_bool(drvthis->name, "QueFansOnly", 0, 0);
  if (p->que_fans_only) 
    report(RPT_INFO, "%s Only fan management will be qued and gauranteed.",
         __FUNCTION__);

  /*  Temp and Fan syslog reporting or no */
  p->log_sens = drvthis->config_get_int(drvthis->name, "SyslogSensors", 0, DEFAULT_SYSLOG_SENSORS );
  if (p->log_sens > 86400) {
    report(RPT_NOTICE, "%s SyslogSensors = %d shortened to 1 day (86400 seconds.", 
         __FUNCTION__, p->log_sens);
    p->log_sens=86400;
  }
  strncpy(p->logfile, drvthis->config_get_string(drvthis->name, "logfile", 0, ""), sizeof(p->logfile) );
  if (0==strncmp(p->logfile,"", 1)) {
    report(RPT_INFO, "%s Did not read an output logfile.",  __FUNCTION__);
  }
  else {
    report(RPT_INFO, "%s Read write to logfile %s",  __FUNCTION__, p->logfile);
    p->log_fd=fopen(p->logfile, "a");
    if (p->log_fd == NULL){
      report(RPT_ERR, "%s Could not open logfile %s, %s.",
           __FUNCTION__, p->logfile, strerror(errno));
    }
  }
  
  /* Which size */
  if (p->model == 631)
    default_size = DEFAULT_SIZE_CF631;
  else if (p->model == 633)
    default_size = DEFAULT_SIZE_CF633;
  else if (p->model == 635)
    default_size = DEFAULT_SIZE_CF635;
  strncpy(size, drvthis->config_get_string (drvthis->name, "Size", 0, default_size), sizeof(size));
  size[sizeof(size)-1] = '\0';
  report(RPT_INFO, "%s Size in config: %s.",  __FUNCTION__, size);
  if ((sscanf(size, "%dx%d", &w, &h) != 2)
      || (w <= 0) || (w > LCD_MAX_WIDTH)
      || (h <= 0) || (h > LCD_MAX_HEIGHT)) {
    report(RPT_WARNING, "%s Cannot read size: %s. Using default value.",  __FUNCTION__,  size);
    sscanf(default_size, "%dx%d", &w, &h);
  }
  if (strcmp(size, default_size)) {
    report(RPT_INFO, "%s Model supports: %s.",  __FUNCTION__, default_size );
  }
  int m_w, m_h;
  sscanf(default_size, "%dx%d", &m_w, &m_h);
  if (   (w > m_w ) || (h > m_h) ) {
    report(RPT_WARNING, "%s Size too large: %s.  Using supported value(s).",  __FUNCTION__, size);
    if (w > m_w) w=m_w;
    if (h > m_h) h=m_h;
  }
  p->width = w;
  p->height = h;
  report (RPT_INFO,"%s Size used: %dx%d.",
       __FUNCTION__, p->width, p->height);

  /* make sure the frame buffer is there... */
  p->framebuf = (unsigned char *) malloc(p->width * p->height);
  if (p->framebuf == NULL) {
    report(RPT_CRIT, "%s unable to create framebuffer.",  __FUNCTION__);
    return -1;
  }
  memset(p->framebuf, ' ', p->width * p->height);

  /* make sure the framebuffer backing store is there... */
  p->backingstore = (unsigned char *) malloc(p->width * p->height);
  if (p->backingstore == NULL) {
    report(RPT_CRIT, "%s Unable to create framebuffer backing store.",  __FUNCTION__);
    return -1;
  }
  memset(p->backingstore, ' ', p->width * p->height);

  /* init timing data */
#ifdef TIMING_REPORT
  clear_timing(drvthis);
#ifdef LIVE_TIMING
  driver_pointer_global=drvthis;
  install_signal(drvthis);
#endif
#endif

  /* and now send other init packets */
  
  /* Set display-specific stuff.. */
  CFontz633_hidecursor (drvthis);

  CFontz633_set_contrast (drvthis, p->contrast);
  
  CFontz633_hardware_clear (drvthis);

  /* turn LEDs off on a CF635 */
  CFontz633_output(drvthis, 0);

  if (p->tempcount >0)
    if (!CFontz633_ds18_init(drvthis)) return -1;
  
  if (!CFontz633_fan_init(drvthis)) return -1;
  
  /* what else to set on completion of ds18 init? */
  p->sending=0;
  p->endq=1;
  p->sendq[0].command=CF633_Dummy_Unused;

/*  report(RPT_CRIT, "%s EXPLICIT EXIT ON DEBUGGING.",  __FUNCTION__);
  usleep(100); // EXPLICIT EXIT ON DEBUGGING INIT\n gets up
  exit(0);
*/
  spawn_clones (drvthis); 
  report(RPT_DEBUG, "%s Done.",  __FUNCTION__);
  return 0;
}


/*
 * Clean-up
 */
MODULE_EXPORT void
CFontz633_close (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;
  int cmd;
  // try to display the Thanks / Goodby screen ...
  CFontz633_flush (drvthis); 
#ifndef TIMING_REPORT
  // under 100% load not getting to report
  // error in report math; / 0 ?; printing format error ?
  usleep( 150000); // let goodby_screen get up
  // 110000 = get 1st line only usually
#endif
  p->log_sens=0;  //  stop all logging by that thread
#ifdef DO_ULT
  // stop any other sytem info reads too
  p->log_uptime = 0;
  p->log_load = 0;
  p->log_load_s = 0;
  p->log_load_u = 0;
#endif // DO_ULT
#ifdef TIMING_REPORT
  timing_report(drvthis);
#else
  report(RPT_QUE, "%s  max que depth = % 3u,  times over 128 = % 3d",
      p->maxq, p->deepies);
#ifdef STACK_USAGE
  report(RPT_CRIT, "Stack: fans %5d  listen %5d log/cli %5d",
      check_clone_stack_usage (drvthis, NULL, p->fan_stack,    FAN_STACK),
      check_clone_stack_usage (drvthis, NULL, p->listen_stack, LSN_STACK),
      check_clone_stack_usage (drvthis, NULL, p->lg_cli_stack, L_C_STACK));
#endif // STACK_USAGE
#endif // TIMING_REPORT
  if (p->fan_stack){
    free (p->fan_stack);
    p->fan_stack = NULL;
  }
  if (p->listen_stack){
    free (p->listen_stack);
    p->listen_stack = NULL;
  }
  if (p->lg_cli_stack){
    free (p->lg_cli_stack);
    p->lg_cli_stack = NULL;
  } 
  for (cmd=0; cmd < 4; cmd ++){
    if (p->fans[cmd].name) free (p->fans[cmd].name);
    if (p->fans[cmd].d_label) free (p->fans[cmd].d_label);
#ifdef CLIENT_CHILD
    if (p->fans[cmd].line) free (p->fans[cmd].line);
    if (p->fans[cmd].line2) free (p->fans[cmd].line2);
#endif
  }
  for (cmd=0; cmd < p->tempcount; cmd ++){
    if (p->ds18[cmd].name) free (p->ds18[cmd].name);
#ifdef CLIENT_CHILD
    if (p->ds18[cmd].line) free (p->ds18[cmd].line);
#endif
  }
    /* going multithreaded can each thread have a serarate exit_program ?
     * when driver unloads module, how are module children closing ?
     * does each thread ceck for the non-NULL ness of these b4 using / writitng ?
     * are all such Null checks adequately reporting & properly aborting ?
     */
#ifdef DO_ULT
    for (cmd=0; cmd < p->topnum; cmd ++){
      if (p->topnumnames[cmd]) free(p->topnumnames[cmd]);
    }
    for (cmd=0; cmd < p->topname; cmd ++){
      if (p->topnames[cmd]) free (p->topnames[cmd]);
#ifdef CLIENT_CHILD
      if (p->topnameline[cmd]) free(p->topnameline[cmd]);
#endif
    }
#endif // DO_ULT
#ifdef DO_LM
    for (cmd=0; cmd < p->lmnum; cmd ++){
      if (p->lmnames[cmd]) free (p->lmnames[cmd]);
      if (p->lmfeatu[cmd]) free (p->lmfeatu[cmd]);
#ifdef CLIENT_CHILD
      if (p->lmline[cmd]) free(p->lmline[cmd]);
#endif
      
      if (p->chipn[cmd]){
        if (p->chipn[cmd]->prefix) free(p->chipn[cmd]->prefix);
        if (p->chipn[cmd]->busname) free(p->chipn[cmd]->busname);
        free (p->chipn[cmd]);
      }
      if (p->featn[cmd]){
        if (p->featn[cmd]->name) 
          free((char*)p->featn[cmd]->name);
          // free(p->featn[cmd]->name);
          //  warning: passing arg 1 of `free' discards qualifiers from pointer target type
          //  because const char* name; typecast here to avoid compile time warning. or am i missing something ?
        free (p->featn[cmd]);
      }
    }
    if (p->lmconf)  free (p->lmconf);
    if (p->lmfile)  {
      sensors_cleanup();
      fclose (p->lmfile);
    }
#endif // DO_LM
#ifdef DO_HDD
    for (cmd=0; cmd < p->hdd_count; cmd ++){
      if (p->hddname[cmd]) free (p->hddname[cmd]);
      if (p->hddtemp[cmd]) free (p->hddtemp[cmd]);
#ifdef CLIENT_CHILD
      if (p->hddline[cmd]) free (p->hddline[cmd]);
#endif
    }
#endif // DO_HDD
#ifdef MY_CUSTOM_LINE
    if (watts!=NULL) free (watts);
#endif
  if (p->log_fd)  fclose (p->log_fd);
  close(p->fd);
  if (p->framebuf)
    free(p->framebuf);
  p->framebuf = NULL;
  if (p->backingstore)
    free(p->backingstore);
  p->backingstore = NULL;
  free(p);
  drvthis->store_private_ptr(drvthis, NULL);
}


/*
 * Returns the display width in characters
 */
MODULE_EXPORT int
CFontz633_width (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  return p->width;
}


/*
 * Returns the display height in characters
 */
MODULE_EXPORT int
CFontz633_height (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  return p->height;
}


/*
 * Returns the width of a character in pixels
 */
MODULE_EXPORT int
CFontz633_cellwidth (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  return p->cellwidth;
}


/*
 * Returns the height of a character in pixels
 */
MODULE_EXPORT int
CFontz633_cellheight (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  return p->cellheight;
}

/*
 * Flushes all output to the lcd...
 */
MODULE_EXPORT void
CFontz633_flush (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;
  int i,j;
  COMMAND_PACKET flush;

  if (p->model == 633) {
  /*
   * For CF633 we don't use delta update yet.
   * The protocol only permit update of full or partial line starting from pos 0.
   */
  unsigned char *xp = p->framebuf;
  unsigned char *xq = p->backingstore;
  flush.command=CF633_Set_LCD_Contents_Line_One;
  flush.data_length=p->width;

    for (i = 0; i < p->width; i++) {
      if (*xp++ != *xq++) {
  //send_bytes_message(p->fd, CF633_Set_LCD_Contents_Line_One, 16, p->framebuf);
        memcpy(flush.data, p->framebuf, p->width);
  send_packet(drvthis, &flush);
  if (p->init_completed==0) {
    CFontz633_await_packet(drvthis, &flush, 0);
  }
        memcpy(p->backingstore, p->framebuf, p->width);
        break;
      }
    }

    xp = p->framebuf + p->width;
    xq = p->backingstore + p->width;
    flush.command=CF633_Set_LCD_Contents_Line_Two;

    for (i = 0; i < p->width; i++) {
      if (*xp++ != *xq++) {
        //send_bytes_message(p->fd, CF633_Set_LCD_Contents_Line_Two, 16, p->framebuf + p->width);
        memcpy(flush.data, p->framebuf + p->width, p->width);
  send_packet(drvthis, &flush);
  if (p->init_completed==0) {
    CFontz633_await_packet(drvthis, &flush, 0);
  }
        memcpy(p->backingstore + p->width, p->framebuf + p->width, p->width);
        break;
      }
    }
  }
  else { /* (p->model != 633) */
  /*
   * CF631 / CF635 protocol is more flexible and we can do real delta update.
   */

    for (i = 0; i < p->height; i++) {
      // set  pointers to start of the line in frame buffer & backing store
      unsigned char *sp = p->framebuf + (i * p->width);
      unsigned char *sq = p->backingstore + (i * p->width);

      report(RPT_DEBUG, "%s Framebuf: '%.*s'",  __FUNCTION__, p->width, sp);
      report(RPT_DEBUG, "%s Backingstore: '%.*s'",  __FUNCTION__, p->width, sq);
      /* Strategy:
       * - not more than one update command per line
       * - leave out leading and trailing parts that are identical
       */
      
      // set  pointers to end of the line in frame buffer & backing store
      unsigned char *ep = sp + (p->width - 1);
      unsigned char *eq = sq + (p->width - 1);
      int length = 0;

      // skip over leading identical portions of the line
      for (j = 0; (sp <= ep) && (*sp == *sq); sp++, sq++, j++)
  ;
    
      // skip over trailing identical portions of the line
      for (length = p->width - j; (length > 0) && (*ep == *eq); ep--, eq--, length--)
  ;

      /* there are differences, ... */
      if (length > 0) {
        
  /* ... send then to the LCD */
  flush.data[0] = j;  // column
  flush.data[1] = i;  // line

  report(RPT_DEBUG, "%s: l=%d c=%d count=%d string='%.*s'",
         __FUNCTION__, flush.data[0], flush.data[1], length, length, sp);

  flush.command=CF633_Send_Data_to_LCD;
  flush.data_length=length + 2;
  memcpy(&flush.data[2], sp, length);
  //send_bytes_message(p->fd, CF633_Send_Data_to_LCD, length + 2, out);
  send_packet(drvthis, &flush);
  if (p->init_completed==0) {
    CFontz633_await_packet(drvthis, &flush, 0);
  }
      }      
    }  // i < p->height
    memcpy(p->backingstore, p->framebuf, p->width * p->height);
  }
}

/*
 * Return one char from the KeyRing
 */
MODULE_EXPORT char *
CFontz633_get_key (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;
  unsigned char key;

  key = GetKeyFromKeyRing(&p->keyring);

  switch (key) {
    case CF633_KEY_LEFT:
      return "Left";
      break;
    case CF633_KEY_UP:
      return "Up";
      break;
    case CF633_KEY_DOWN:
      return "Down";
      break;
    case CF633_KEY_RIGHT:
      return "Right";
      break;
    case CF633_KEY_ENTER:
      return "Enter";
      break;
    case CF633_KEY_ESCAPE:
      return "Escape";
      break;
    case CF631_KEY_UL_PRESS:
      return "Up";
      break;
    case CF631_KEY_UR_PRESS:
      return "Enter";
      break;
    case CF631_KEY_LL_PRESS:
      return "Down";
      break;
    case CF631_KEY_LR_PRESS:
      return "Escape";
      break;
    case CF633_KEY_UP_RELEASE:
    case CF633_KEY_DOWN_RELEASE:
    case CF633_KEY_LEFT_RELEASE:
    case CF633_KEY_RIGHT_RELEASE:
    case CF633_KEY_ENTER_RELEASE:
    case CF633_KEY_ESCAPE_RELEASE:
    case CF631_KEY_UL_RELEASE:
    case CF631_KEY_UR_RELEASE:
    case CF631_KEY_LL_RELEASE:
    case CF631_KEY_LR_RELEASE:
      // report(RPT_INFO, "%s Returning key release 0x%2x",  __FUNCTION__, key);
      return NULL;
      break;
    default:
      if (key != '\0')
        report(RPT_INFO, "%s Untreated unknown key 0x%2x",  __FUNCTION__, key);
      return NULL;
      break;
  }
  return NULL;
}


/*
 * Prints a character on the lcd display, at position (x,y).
 * The upper-left is (1,1), and the lower right should be (16,2).
 */
MODULE_EXPORT void
CFontz633_chr (Driver *drvthis, int x, int y, char c)
{
  PrivateData *p = drvthis->private_data;

  y--;
  x--;

  p->framebuf[(y * p->width) + x] = (p->model == 633)
                              ? c
            : CFontz_charmap[(unsigned) c];
}


/*
 * Prints a character on the lcd display, at position (x,y).
 * The upper-left is (1,1), and the lower right should be (16,2).
 */
static void
CFontz633_raw_chr (Driver *drvthis, int x, int y, unsigned char c)
{
  PrivateData *p = drvthis->private_data;

  y--;
  x--;

  p->framebuf[(y * p->width) + x] = c;
}


/*
 * Returns current contrast (in promille)
 * This is only the locally stored contrast, the contrast value
 * cannot be retrieved from the LCD.
 */
MODULE_EXPORT int
CFontz633_get_contrast (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  return p->contrast;
}


/*
 *  Changes screen contrast (in promille)
 */
MODULE_EXPORT void
CFontz633_set_contrast (Driver *drvthis, int promille)
{
  PrivateData *p = drvthis->private_data;
  COMMAND_PACKET contr;

  /* Check it */
  if (promille < 0 || promille  > 1000)
    return;
  /* store the software value since there is not get */
  p->contrast = promille;
    
  /* map range [0, 1000] to a range that the hardware understands */
  /* on CF633: [0, 50], on CF631 & CF635: [0, 255] */
  contr.data[0] = (p->model == 633)
          ? (ubyte)(p->contrast / 20)
          : (ubyte)((p->contrast * 255) / 1000);
  
  contr.command=CF633_Set_LCD_Contrast;
  contr.data_length=1;
  
  report(RPT_INFO, "%s Contrast set to %d. (CF hardware=%d)",
       __FUNCTION__, p->contrast, contr.data[0]);
   send_packet (drvthis, &contr);
//   adjustments via panel menu once up mould hang on await packet 
  if (p->init_completed==0){ // if during init without queing, wait for ack
    CFontz633_await_packet (drvthis, &contr, 0);
  }
}


/*
 * Retrieves brightness (in promille)
 */
MODULE_EXPORT int
CFontz633_get_brightness(Driver *drvthis, int state)
{
  PrivateData *p = drvthis->private_data;

  return (state == BACKLIGHT_ON) ? p->brightness : p->offbrightness;
}


/*
 * Sets on/off brightness (in promille)
 */
MODULE_EXPORT void
CFontz633_set_brightness(Driver *drvthis, int state, int promille)
{
  PrivateData *p = drvthis->private_data;

  /* Check it */
  if (promille < 0 || promille > 1000)
    return;

  /* store the software value since there is not get */
  if (state == BACKLIGHT_ON) {
    p->brightness = promille;
    CFontz633_backlight(drvthis, BACKLIGHT_ON);
  }
  else {
    p->offbrightness = promille;
    CFontz633_backlight(drvthis, BACKLIGHT_OFF);
  }
}


/*
 * Sets the backlight on or off.
 * The hardware support any value between 0 and 100.
 */
MODULE_EXPORT void
CFontz633_backlight (Driver *drvthis, int on)
{
  PrivateData *p = drvthis->private_data;
  int hardware_value = (on == BACKLIGHT_ON)
           ? p->brightness
           : p->offbrightness;
  
  static COMMAND_PACKET backlight={ 
    CF633_Set_LCD_And_Keypad_Backlight, 
    1, {0}};
  
  static unsigned int redundant_repeats_from_upstream=0;
  
  /* map range [0, 1000] -> [0, 100] that the hardware understands */
  hardware_value /= 10;
  
  if (backlight.data[0] != hardware_value) {
    backlight.data[0]=hardware_value;
    send_packet(drvthis, &backlight);
    if(p->init_completed==0){
      CFontz633_await_packet(drvthis, &backlight, 0);
    }
    report(RPT_DEBUG, "%s CF633_Set_LCD_And_Keypad_Backlight %d is new after %d repeats of %d.",
         __FUNCTION__, hardware_value, redundant_repeats_from_upstream, 
        backlight.data[0]);
    redundant_repeats_from_upstream=0;
  }
  else{
    redundant_repeats_from_upstream++;
    if (redundant_repeats_from_upstream==65535){
      redundant_repeats_from_upstream=0;
      report(RPT_DEBUG, "%s CF633_Set_LCD_And_Keypad_Backlight 65536 repeats of %d.",
         __FUNCTION__, backlight.data[0]);
    }
  }
}


/*
 * Get rid of the blinking cursor
 */
static void
CFontz633_hidecursor (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;
  COMMAND_PACKET cursor;
  cursor.command=CF633_Set_LCD_Cursor_Style;
  cursor.data_length=1;
  cursor.data[0]=0;

  send_packet(drvthis, &cursor);
  if(p->init_completed==0){
    CFontz633_await_packet(drvthis, &cursor, 0);
  }
}

/*
 * Stop live reporting of temperature.
 */
static void
CFontz633_no_live_report(Driver *drvthis)
{
  // 12 19 05 this is getting error responses on 635 w / scab
  // presumably will with 631 as well
  // scab.pdf says deprecated ( & no 'see new command' )
  PrivateData *p = drvthis->private_data;
  if (p->model==635) return;
  COMMAND_PACKET packet;
  packet.command=CF633_Set_Up_Live_Fan_or_Temperature_Display;
  packet.data_length=2;
  packet.data[0]=0;
  packet.data[1]=0;

  for (packet.data[0] = 0; packet.data[0] < 8; packet.data[0]++){
    send_packet(drvthis, &packet);
    if(p->init_completed==0){
      CFontz633_await_packet(drvthis, &packet, 0);
    }
      int top=(0xC0&(p->incoming_command.command)) >> 6;
      int cmd=(0x3F&(p->incoming_command.command));
    if ((top==1) && (cmd==CF633_Set_Up_Live_Fan_or_Temperature_Display)){
      report(RPT_INFO, "%s Live reports stopped sucessfully.",  __FUNCTION__);
    }
    else  if ((top==3) && (cmd==CF633_Set_Up_Live_Fan_or_Temperature_Display)){
      report(RPT_INFO, "%s Stopping live reporting returned error expected from non-temp/fan hardware.",
           __FUNCTION__);
      /* do this first ... to determine if we have fan temp hardware
       * and configure other features as result ?  Hmm, no diff in returns ?*/
    }
  }
}

/*
 * Stop the reporting of any fan.
 */
static void
CFontz633_no_fan_report(Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;
  COMMAND_PACKET packet;
  
  packet.command=CF633_Set_Up_Fan_Reporting;
  packet.data_length=1;
  packet.data[0]=0;
  send_packet (drvthis, &packet);
  if (p->init_completed==0) {
    CFontz633_await_packet(drvthis, &packet, 0);
  }
    int top=(0xC0&(p->incoming_command.command)) >> 6;
    int cmd=(0x3F&(p->incoming_command.command));
  if ((top==1) && (cmd==CF633_Set_Up_Fan_Reporting)){
    report(RPT_INFO, "%s Fan reporting stopped sucessfully.", __FUNCTION__);
  }
  else  if ((top==3) && (cmd==CF633_Set_Up_Fan_Reporting)){
    // 631 w/o SCAB is not returning errors, reports fan controll OK ...
    // How can zero fan / temp hardware be detected to turn off spurious portions of driver ?
    report(RPT_INFO, "%s Stopping fan reporting returned error expected from non-temp hardware.",
         __FUNCTION__);
    report(RPT_INFO, "%s    ( ... configure things based on this.",
         __FUNCTION__);
    /* might still read fan settings from config and send useless fan packets ? */
  }
}

/*
 * Stop the reporting of any temperature.
 */
static void
CFontz633_no_temp_report(Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;
  COMMAND_PACKET packet;
  
  packet.command=CF633_Set_Up_Temperature_Reporting;
  packet.data_length=4;
  packet.data[0]=packet.data[1]=packet.data[2]=packet.data[3]=0;
  CFontz633_await_packet(drvthis, &packet, 1);
    int top=(0xC0&(p->incoming_command.command)) >> 6;
    int cmd=(0x3F&(p->incoming_command.command));
  if ((top==1) && (cmd==CF633_Set_Up_Temperature_Reporting)){
    report(RPT_INFO, "%s Temperature reporting stopped sucessfully.",  __FUNCTION__);
  }
  else  if ((top==3) && (cmd==CF633_Set_Up_Temperature_Reporting)){
    report (RPT_WARNING, "%s Stopping temperature reporting returned error.",  __FUNCTION__);
    p->tempcount=0;
  }
}

/*
 * Reset the display bios
 */
static void 
CFontz633_reboot (Driver *drvthis, struct timeval *sent_reboot )
{
  COMMAND_PACKET reboo;
  
  report(RPT_INFO, "%s module reboot requested.",  __FUNCTION__);
  reboo.command=CF633_Reboot;
  reboo.data_length=3;
  reboo.data[0] = 8;
  reboo.data[1] = 18;
  reboo.data[2] = 99;
  gettimeofday(sent_reboot,NULL);
  send_packet(drvthis, &reboo);
}

static void 
CFontz633_rbooted (Driver *drvthis, struct timeval *sent_reboot)
{
  PrivateData *p = drvthis->private_data;
  COMMAND_PACKET reboo;
  struct timeval ret;

  /* same as in CFontz633 _ reboot */
  reboo.command=CF633_Reboot;
  reboo.data_length=3;
  reboo.data[0] = 8;
  reboo.data[1] = 18;
  reboo.data[2] = 99;

  sleep(1); // wait for CF633_Reboot
  usleep(800000); // wait for CF633_Reboot
   CFontz633_await_packet (drvthis, &reboo, 0);
  gettimeofday(&ret, NULL);
  report(RPT_INFO, "%s module reboot completed in %f seconds.", __FUNCTION__, 
      (float)time_diff(drvthis, *sent_reboot, ret)/1000000);
  if ( (0x3F&(p->incoming_command.command))!=CF633_Reboot) {
    report(RPT_WARNING, "%s module reboot Failed in %f seconds. All downhill from here :)", 
       __FUNCTION__, (float)time_diff(drvthis, *sent_reboot, ret)/1000000);
    show_packet(drvthis, RPT_WARNING, &p->incoming_command);
  }
}

/*
 * Sets up for vertical bars.
 */
static void
CFontz633_init_vbar (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;
  char a[] = {
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 1, 1, 1, 1, 1,
  };
  char b[] = {
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
  };
  char c[] = {
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
  };
  char d[] = {
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
  };
  char e[] = {
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
  };
  char f[] = {
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
  };
  char g[] = {
    0, 0, 0, 0, 0, 0,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
  };

  if (p->ccmode != vbar) {
    if (p->ccmode != standard) {
      /* Not supported(yet) */
      report(RPT_WARNING, "%s Cannot combine two modes using user defined characters",
           __FUNCTION__);
      return;
    }
    p->ccmode = vbar;

    CFontz633_set_char (drvthis, 1, a);
    CFontz633_set_char (drvthis, 2, b);
    CFontz633_set_char (drvthis, 3, c);
    CFontz633_set_char (drvthis, 4, d);
    CFontz633_set_char (drvthis, 5, e);
    CFontz633_set_char (drvthis, 6, f);
    CFontz633_set_char (drvthis, 7, g);
  }
}

/*
 * Inits horizontal bars...
 */
static void
CFontz633_init_hbar (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;
  char a[] = {
    1, 0, 0, 0, 0, 0,
    1, 0, 0, 0, 0, 0,
    1, 0, 0, 0, 0, 0,
    1, 0, 0, 0, 0, 0,
    1, 0, 0, 0, 0, 0,
    1, 0, 0, 0, 0, 0,
    1, 0, 0, 0, 0, 0,
    1, 0, 0, 0, 0, 0,
  };
  char b[] = {
    1, 1, 0, 0, 0, 0,
    1, 1, 0, 0, 0, 0,
    1, 1, 0, 0, 0, 0,
    1, 1, 0, 0, 0, 0,
    1, 1, 0, 0, 0, 0,
    1, 1, 0, 0, 0, 0,
    1, 1, 0, 0, 0, 0,
    1, 1, 0, 0, 0, 0,
  };
  char c[] = {
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
  };
  char d[] = {
    1, 1, 1, 1, 0, 0,
    1, 1, 1, 1, 0, 0,
    1, 1, 1, 1, 0, 0,
    1, 1, 1, 1, 0, 0,
    1, 1, 1, 1, 0, 0,
    1, 1, 1, 1, 0, 0,
    1, 1, 1, 1, 0, 0,
    1, 1, 1, 1, 0, 0,
  };
  char e[] = {
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 1, 0,
  };
  char f[] = {
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
  };

  if (p->ccmode != hbar) {
    if (p->ccmode != standard) {
      /* Not supported(yet) */
      report(RPT_WARNING, "%s CFontz633_init_hbar: Cannot combine two modes using user defined characters",
           __FUNCTION__);
      return;
    }
    p->ccmode = hbar;

    CFontz633_set_char (drvthis, 1, a);
    CFontz633_set_char (drvthis, 2, b);
    CFontz633_set_char (drvthis, 3, c);
    CFontz633_set_char (drvthis, 4, d);
    CFontz633_set_char (drvthis, 5, e);
    CFontz633_set_char (drvthis, 6, f);
  }
}


/*
 * Draws a vertical bar...
 */
MODULE_EXPORT void
CFontz633_vbar (Driver *drvthis, int x, int y, int len, int promille, int options)
{
/* x and y are the start position of the bar.
 * The bar by default grows in the 'up' direction
 * (other direction not yet implemented).
 * len is the number of characters that the bar is long at 100%
 * promille is the number of promilles (0..1000) that the bar should be filled.
 */
  PrivateData *p = drvthis->private_data;

  CFontz633_init_vbar(drvthis);
  lib_vbar_static(drvthis, x, y, len, promille, options, p->cellheight, 0);
}


/*
 * Draws a horizontal bar to the right.
 */
MODULE_EXPORT void
CFontz633_hbar (Driver *drvthis, int x, int y, int len, int promille, int options)
{
/* x and y are the start position of the bar.
 * The bar by default grows in the 'right' direction
 * (other direction not yet implemented).
 * len is the number of characters that the bar is long at 100%
 * promille is the number of promilles (0..1000) that the bar should be filled.
 */
  PrivateData *p = drvthis->private_data;

  CFontz633_init_hbar(drvthis);
  lib_hbar_static(drvthis, x, y, len, promille, options, p->cellwidth, 0);
}


/*
 * Inits big numbers
 */
void
CFontz633_init_num(Driver *drvthis)
{
PrivateData *p = drvthis->private_data;

char bignum_ccs[8][CELLWIDTH*CELLHEIGHT] = {
  { 1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0 },

  { 0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0 },

  { 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0 },

  { 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0 },

  { 1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1 },

  { 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1 },

  { 1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0 },

  { 0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0 }
};

  if (p->ccmode != bignum) {
    int i;

    if (p->ccmode != standard) {
      /* Not supported (yet) */
      report(RPT_WARNING, "%s Cannot combine two modes using user defined characters",
           __FUNCTION__);
      return;
    }

    p->ccmode = bignum;

    for (i = 0; i < NUM_CCs; i++)
      CFontz633_set_char(drvthis, i, bignum_ccs[i]);
  }
}


/*
 * Writes a big number.
 * Only works on 4-line displays
 */
MODULE_EXPORT void
CFontz633_num(Driver *drvthis, int x, int num)
{
PrivateData *p = drvthis->private_data;

/* each bignum is constructed in a 3 x 4 matrix and consists
 * of only the 8 characters defined above as well as ' '
 *
 * The following table defines the 11 big numbers '0'-'9', ':'
 * and the custom base characters they consist of
 */
char bignum_map[11][4][3] = {
  { /* 0: */
    {  1,  2,  3 },
    {  6, 32,  6 },
    {  6, 32,  6 },
    {  7,  2, 32 } },
  { /* 1: */
    {  7,  6, 32 },
    { 32,  6, 32 },
    { 32,  6, 32 },
    {  7,  2, 32 } },
  { /* 2: */
    {  1,  2,  3 },
    { 32,  5,  0 },
    {  1, 32, 32 },
    {  2,  2,  0 } },
  { /* 3: */
    {  1,  2,  3 },
    { 32,  5,  0 },
    {  3, 32,  6 },
    {  7,  2, 32 } },
  { /* 4: */
    { 32,  3,  6 },
    {  1, 32,  6 },
    {  2,  2,  6 },
    { 32, 32,  0 } },
  { /* 5: */
    {  1,  2,  0 },
    {  2,  2,  3 },
    {  3, 32,  6 },
    {  7,  2, 32 } },
  { /* 6: */
    {  1,  2, 32 },
    {  6,  5, 32 },
    {  6, 32,  6 },
    {  7,  2, 32 } },
  { /* 7: */
    {  2,  2,  6 },
    { 32,  1, 32 },
    { 32,  6, 32 },
    { 32,  0, 32 } },
  { /* 8: */
    {  1,  2,  3 },
    {  4,  5,  0 },
    {  6, 32,  6 },
    {  7,  2, 32 } },
  { /* 9: */
    {  1,  2,  3 },
    {  4,  3,  6 },
    { 32,  1, 32 },
    {  7, 32, 32 } },
  { /* colon: (only 1st column used) */
    { 32, 32, 32 },
    {  0, 32, 32 },
    {  0, 32, 32 },
    { 32, 32, 32 } }
};

  if ((num < 0) || (num > 10))
    return;

  if ((p->width >= 20) && (p->height >= 4)) {
    int y = (p->height - 2) / 2;  /* center vertically */
    int x2, y2;

    CFontz633_init_num(drvthis);

    for (x2 = 0; x2 < 3; x2++) {
      for (y2 = 0; y2 < 4; y2++) {
        CFontz633_chr(drvthis, x+x2, y+y2, bignum_map[num][y2][x2]);
      }
      if (num == 10)
        x2 = 2; /* = break, for colon only */
    }
  }
  else {
    CFontz633_chr(drvthis, x, 1 + (p->height - 1) / 2,
            (num == 10) ? ':' : (num + '0'));
  }
}


/*
 * Sets a custom character from 0 - (NUM_CCs-1)
 *
 * For input, values > 0 mean "on" and values <= 0 are "off".
 *
 * The input is just an array of characters...
 */
MODULE_EXPORT void
CFontz633_set_char (Driver *drvthis, int n, char *dat)
{
  PrivateData *p = drvthis->private_data;
  static COMMAND_PACKET st_char= { CF633_Set_LCD_Special_Character_Data,
    9, {0,0,0,0,0,0,0,0,0}};
  static char p_dat[48]; // [p->cellheight*p->cellwidth]

  if ((n < 0) || (n >= NUM_CCs) || (!dat)   || 
      ((st_char.data[0]==n) &&
       ( bcmp(dat, p_dat, p->cellheight*p->cellwidth) == 0)))
    return;
  int row, col;
  st_char.data[0] = n;  /* Custom char to define. xxx */
  for (row = 0; row < p->cellheight; row++) {
    int letter = 0;

    for (col = 0; col < p->cellwidth; col++) {
      p_dat[(row * p->cellwidth) + col]=dat[(row * p->cellwidth) + col];
      letter <<= 1;
      letter |= (dat[(row * p->cellwidth) + col] > 0);
    }
    st_char.data[row+1] = letter;
  }
  
  send_packet(drvthis, &st_char);
  if(p->init_completed==0){
    CFontz633_await_packet(drvthis, &st_char, 0);
    return;
  }
}


/*
 * Places an icon on screen
 */
MODULE_EXPORT int
CFontz633_icon (Driver *drvthis, int x, int y, int icon)
{
  PrivateData *p = drvthis->private_data;
  char icons[8][6 * 8] = {
  /* Empty Heart */
    {
     0, 1, 1, 1, 1, 1,
     0, 1, 0, 1, 0, 1,
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 1, 0, 0, 0, 1,
     0, 1, 1, 0, 1, 1,
     0, 1, 1, 1, 1, 1,
     },
  /* Filled Heart */
    {
     0, 1, 1, 1, 1, 1,
     0, 1, 0, 1, 0, 1,
     0, 0, 1, 0, 1, 0,
     0, 0, 1, 1, 1, 0,
     0, 0, 1, 1, 1, 0,
     0, 1, 0, 1, 0, 1,
     0, 1, 1, 0, 1, 1,
     0, 1, 1, 1, 1, 1,
     },
  /* arrow_up */
    {
     0, 0, 0, 1, 0, 0,
     0, 0, 1, 1, 1, 0,
     0, 1, 0, 1, 0, 1,
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 0, 0, 0,
     },
  /* arrow_down */
    {
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 1, 0, 0,
     0, 1, 0, 1, 0, 1,
     0, 0, 1, 1, 1, 0,
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 0, 0, 0,
     },
  /* checkbox_off */
    {
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 1, 1, 1, 1, 1,
     0, 1, 0, 0, 0, 1,
     0, 1, 0, 0, 0, 1,
     0, 1, 0, 0, 0, 1,
     0, 1, 1, 1, 1, 1,
     0, 0, 0, 0, 0, 0,
     },
  /* checkbox_on */
    {
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 1, 0, 0,
     0, 1, 1, 1, 0, 1,
     0, 1, 0, 1, 1, 0,
     0, 1, 0, 1, 0, 1,
     0, 1, 0, 0, 0, 1,
     0, 1, 1, 1, 1, 1,
     0, 0, 0, 0, 0, 0,
     },
  /* checkbox_gray */
    {
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 1, 1, 1, 1, 1,
     0, 1, 0, 1, 0, 1,
     0, 1, 1, 0, 1, 1,
     0, 1, 0, 1, 0, 1,
     0, 1, 1, 1, 1, 1,
     0, 0, 0, 0, 0, 0,
     },
   /* Ellipsis */
    {
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 1, 0, 1, 0, 1,
     },
  };

  /* Yes we know, this is a VERY BAD implementation :-) */
  switch (icon) {
    case ICON_BLOCK_FILLED:
      if (p->model == 633)
        CFontz633_chr(drvthis, x, y, 255);
      else
        CFontz633_raw_chr(drvthis, x, y, 31);
      break;
    case ICON_HEART_FILLED:
      CFontz633_set_char(drvthis, 0, icons[1]);
      CFontz633_chr(drvthis, x, y, 0);
      break;
    case ICON_HEART_OPEN:
      CFontz633_set_char(drvthis, 0, icons[0]);
      CFontz633_chr(drvthis, x, y, 0);
      break;
    case ICON_ARROW_UP:
      if (p->model == 633) {
        CFontz633_set_char(drvthis, 1, icons[2]);
        CFontz633_chr(drvthis, x, y, 1);
      }
      else
        CFontz633_raw_chr(drvthis, x, y, 0xDE);
      break;
    case ICON_ARROW_DOWN:
      if (p->model == 633) {
        CFontz633_set_char(drvthis, 2, icons[3]);
        CFontz633_chr(drvthis, x, y, 2);
      }
      else
        CFontz633_raw_chr(drvthis, x, y, 0xE0);
      break;
    case ICON_ARROW_LEFT:
      if (p->model == 633)
        CFontz633_raw_chr(drvthis, x, y, 0x7F);
      else
        CFontz633_raw_chr(drvthis, x, y, 0xE1);
      break;
    case ICON_ARROW_RIGHT:
      if (p->model == 633)
        CFontz633_raw_chr(drvthis, x, y, 0x7E);
      else
        CFontz633_raw_chr(drvthis, x, y, 0xDF);
      break;
    case ICON_CHECKBOX_OFF:
      CFontz633_set_char(drvthis, 3, icons[4]);
      CFontz633_chr(drvthis, x, y, 3);
      break;
    case ICON_CHECKBOX_ON:
      CFontz633_set_char(drvthis, 4, icons[5]);
      CFontz633_chr(drvthis, x, y, 4);
      break;
    case ICON_CHECKBOX_GRAY:
      CFontz633_set_char(drvthis, 5, icons[6]);
      CFontz633_chr(drvthis, x, y, 5);
      break;
    case ICON_SELECTOR_AT_LEFT:
      if (p->model == 633)
        return -1;
      CFontz633_raw_chr(drvthis, x, y, 0x10);
      break;
    case ICON_SELECTOR_AT_RIGHT:
      if (p->model == 633)
        return -1;
      CFontz633_raw_chr(drvthis, x, y, 0x11);
      break;
    default:
      return -1; /* Let the core do other icons */
  }
  return 0;
}


/*
 * Clears the LCD screen
 */
MODULE_EXPORT void
CFontz633_clear (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  memset(p->framebuf, ' ', p->width * p->height);
  p->ccmode = standard;
}


/*
 * Hardware clears the LCD screen
 */
static void
CFontz633_hardware_clear (Driver *drvthis)
{
  COMMAND_PACKET clear;
  clear.command=CF633_Clear_LCD_Screen;
  clear.data_length=0;
   CFontz633_await_packet (drvthis, &clear,1 );
}


/*
 * Prints a string on the lcd display, at position (x,y).  The
 * upper-left is (1,1), and the lower right is (p->width, p->height).
 */
MODULE_EXPORT void
CFontz633_string (Driver *drvthis, int x, int y, char string[])
{
  PrivateData *p = drvthis->private_data;
  int i;

  /* Convert 1-based coords to 0-based... */
  x--;
  y--;

  for (i = 0; string[i] != '\0'; i++) {
    /* Check for buffer overflows... */
    if ((y * p->width) + x + i > (p->width * p->height))
      break;
    p->framebuf[(y * p->width) + x + i] = (p->model == 633)
                  ? string[i]
                  : CFontz_charmap[(unsigned) string[i]];
  }
}


/*
 * Output values using the LEDs of a CF635
 */
MODULE_EXPORT void
CFontz633_output(Driver *drvthis, int state)
{
  PrivateData *p = drvthis->private_data;
  if (p->model != 635)
    return;
  int lednum;
  static const unsigned char CFontz635_LEDs[CF635_NUM_LEDs] = {
    11, 9, 7, 5,  // Green LEDs first, Top first
    12,10, 8, 6,  // Red LEDs next, Top first
  };

  COMMAND_PACKET leds;
  leds.command=CF633_Set_GPIO_Pin;
  leds.data_length=2;
  

  for (lednum = 0; lednum < CF635_NUM_LEDs; lednum++) {
    unsigned int mask = (1 << lednum);
    int on_off = (state & mask);
    
    if ((p->LEDstate & mask) != on_off) {
      leds.data[0]=CFontz635_LEDs[lednum];  
      leds.data[1]=(on_off == 0) ? 0 : 100;
      send_packet (drvthis, &leds);
      if (p->init_completed==0){
        CFontz633_await_packet(drvthis, &leds, 0);  
      }
    }
  }
  p->LEDstate = state;
}


/*
 * KeyRing handling functions.
 * This separates the producer from the consumer.
 * It is just a small fifo of unsigned char.
 */

/** initialize/empty key ring by resetting its read & write pointers */
void EmptyKeyRing(KeyRing *kr)
{
  kr->head = kr->tail = 0;
}


/** add byte to key ring; return success (byte added) / failure (key ring is full) */
int AddKeyToKeyRing(KeyRing *kr, unsigned char key)
{
  if (((kr->head + 1) % KEYRINGSIZE) != (kr->tail % KEYRINGSIZE)) {
     /*  fprintf(stderr,"We add key: %d\n", key); */

          kr->contents[kr->head % KEYRINGSIZE] = key;
      kr->head = (kr->head + 1) % KEYRINGSIZE;
    return 1;
  }
  
  /* KeyRing overflow: do not accept extra key */
  return 0;
}


/** get byte from key ring (or '\0' if key ring is empty) */
unsigned char GetKeyFromKeyRing(KeyRing *kr)
{
  unsigned char retval = '\0';

  kr->tail %= KEYRINGSIZE;

  if ((kr->head % KEYRINGSIZE) != kr->tail) {
    retval = kr->contents[kr->tail];
          kr->tail = (kr->tail + 1) % KEYRINGSIZE;
  }

  /*  if (retval) fprintf(stderr,"We remove key: %d\n", retval); */
  return retval;
}

/** send outgoing_response to the given handle; calc & send CRC when doing so */
void
send_packet(Driver *drvthis, COMMAND_PACKET *out)
{
  PrivateData *p = drvthis->private_data;
  
  if ( (p->init_completed<2) || (p->q_depth==0) || (p->sendq[p->sending].command==CF633_Dummy_Unused) ) {
    // q_d ==0 && pk==Dummy checked in opposite order in which set on acks in listen / sensors ...
/*
May  6 22:00:41 localhost LCDd: send_packet on q     07 Set_LCD_Special_Character_Data adding a Send_Data_to_LCD to que 07 0c (005)
May  6 22:00:41 localhost LCDd: sensors_packet ack q 07 (005) Set_LCD_Special_Character_Data & send q 08 Send_Data_to_LCD : 08 0c (004)
May  6 22:00:42 localhost LCDd: sensors_packet ack q 08 (004) Send_Data_to_LCD & send q 09 Send_Data_to_LCD : 09 0c (003)
May  6 22:00:42 localhost LCDd: sensors_packet ack q 09 (003) Send_Data_to_LCD & send q 0a Send_Data_to_LCD : 0a 0c (002)
May  6 22:00:42 localhost LCDd: sensors_packet ack q 0a (002) Send_Data_to_LCD & send q 0b Send_Data_to_LCD : 0b 0c (001)
May  6 22:00:42 localhost LCDd: send_packet on q     0b Send_Data_to_LCD adding a Set_LCD_Special_Character_Data to que 0b 0d (002)
May  6 22:00:42 localhost LCDd: send_packet on q     0b Send_Data_to_LCD adding a Send_Data_to_LCD to que 0b 0e (003)
May  6 22:00:42 localhost LCDd: sensors_packet ack q 0b Send_Data_to_LCD & none q 0c Dummy_Unused : 0c 0c (000)
May  6 22:00:42 localhost LCDd: send_packet on q     0b Dummy_Unused adding a Set_LCD_Special_Character_Data to que 0b 0f (003)
May  6 22:00:42 localhost kernel: NVRM: memory caching type 0x2 not supported for memory space 3!
May  6 22:00:42 localhost LCDd: send_packet on q     0b Dummy_Unused adding a Set_LCD_Special_Character_Data to que 0b 10 (004)
May  6 22:00:42 localhost LCDd: send_packet on q     0b Dummy_Unused adding a Set_LCD_Special_Character_Data to que 0b 11 (005)
May  6 22:00:43 localhost LCDd: send_packet on q     0b Dummy_Unused adding a Set_LCD_Special_Character_Data to que 0b 12 (006)
May  6 22:00:43 localhost LCDd: send_packet on q     0b Dummy_Unused adding a Set_LCD_Special_Character_Data to que 0b 13 (007)
May  6 22:00:43 localhost LCDd: send_packet on q     0b Dummy_Unused adding a Send_Data_to_LCD to que 0b 14 (008)
May  6 22:00:43 localhost LCDd: send_packet on q     0b Dummy_Unused adding a Set_LCD_Special_Character_Data to que 0b 15 (009)
May  6 22:00:43 localhost LCDd: send_packet on q     0b Dummy_Unused adding a Set_LCD_Special_Character_Data to que 0b 16 (010)
May  6 22:00:44 localhost LCDd: send_packet on q     0b Dummy_Unused adding a Set_LCD_Special_Character_Data to que 0b 17 (011)

que depth += 2 from where !!?
fresh install of debian etch on k6 300, 1st 2.6.8.15 kernel, what are those NVRM messages .... ?
*/ 
    commit_packet(drvthis, out);
    /**** TEST STUFF ****/
    show_packet(drvthis, RPT_DEBUG, out);
    if (p->que_fans_only) {
      unsigned char cmd=(0x3F&(out->command));
      if ( (cmd!=CF633_Set_Fan_Power) 
          && (cmd!=CF633_Set_Up_Fan_Reporting) 
	  && (cmd!=CF633_Set_LCD_And_Keypad_Backlight)){
        report(RPT_QUE, "%s on q     %02x %s BLITHELY sending a %s %02x %02x (%03d)", __FUNCTION__,
            p->sending, pk_nm[p->sendq[p->sending].command]+6, pk_nm[out->command]+6,
            p->sending,p->endq, p->q_depth);
        usleep(25);
        return;
      }
    }
    gettimeofday (&p->tsent, NULL);
    report(RPT_QUE, "%s on q     %02x %s sending a %s %02x %02x (%03d)", __FUNCTION__,
        p->sending, pk_nm[p->sendq[p->sending].command]+6, pk_nm[out->command]+6,
        p->sending, p->endq, p->q_depth+1);
    memcpy(&p->sendq[p->sending].command, &out->command, out->data_length+2);
    p->q_depth++;
    return;
  } // end send now OK
  else {  // packets are qued
    memcpy(&p->sendq[p->endq].command, &out->command, out->data_length+2);
    p->endq++;
    p->q_depth++;
    report(RPT_QUE, "%s on q     %02x %s adding a %s to que %02x %02x (%03d)",
         __FUNCTION__, p->sending, pk_nm[p->sendq[p->sending].command]+6,
        pk_nm[out->command]+6, 
        p->sending, p->endq, p->q_depth);
    if ( p->q_depth > p->maxq){
      p->maxq=p->q_depth;
      report(RPT_QUE, "%s new maximum que depth: %u.",
          __FUNCTION__, p->maxq );
    }
    if (p->q_depth> 128) {
      p->q_recover=p->q_depth;
      if (p->q_depth==UCHAR_MAX-10){
        purge_que(drvthis);
        return; // avoid check b4 exit here
      }
    }
    if (p->q_depth == 1) {
      report(RPT_QUE, "%s suspect ack q to 1 & will see resend soon ? commit_packet now?",
          __FUNCTION__);
      //commit_packet(drvthis, out);
    } 
  } // end packets are cued
}

/* empty que because it is overflowing / full
 * or a key has been pressed and user wants zero
 * lag.
 */
static void purge_que (Driver *drvthis){
  PrivateData *p = drvthis->private_data;
#ifndef TIMING_REPORT
  report(RPT_INFO, "%s PURGING QUE  %02x %02x (%03d)",
      __FUNCTION__, p->sending, p->endq, p->q_depth);
#endif
  /* Let p->sending and  p->sending+1 proceed narmally so
   * acks and send nexts and resends from sending end
   * of que will not be altered by this cleanup */
        unsigned char i=p->sending+9, cmd, pcount=0, ps=p->sending, endque=p->endq;
  	p->endq=p->sending+9;
	p->q_depth=9;
        unsigned char pe=p->endq;
	int fpinq=-1, frinq=-1, blinq=-1;
        do {
	  if (i%40==0) usleep(1); // fast loop preventing an incoming ack
              // um, no switching to listen loop on uniprocessor,
              // without %10 purge is too slow ...
          cmd=(0x3F&(p->sendq[i].command));
	  switch (cmd) { // preserve some commands in que
            case CF633_Set_Fan_Power:			  
              // move any fan powers to front so they will not be
	      // resubmitted to que by timing out
              if (fpinq==-1){
                fpinq=p->endq;
		p->endq++;
		p->q_depth++;
              }
              memcpy(&p->sendq[fpinq].command, &p->sendq[i].command, p->sendq[i].data_length+2);
              report(RPT_QUE, "%s sending %02x, Set_Fan_Power %02x preserved to %02x pared to (%03d)",
                 __FUNCTION__, p->sending, i, fpinq, p->q_depth );
              break;
            case CF633_Set_Up_Fan_Reporting:
              if (frinq==-1){
                frinq=p->endq;
		p->endq++;
		p->q_depth++;
              }
              memcpy(&p->sendq[frinq].command, &p->sendq[i].command, p->sendq[i].data_length+2);
              report(RPT_QUE, "%s sending %02x, Fan_Reporting %02x preserved to %02x pared to (%03d)",
                 __FUNCTION__, p->sending, i, frinq, p->q_depth );
              break;
            case CF633_Set_LCD_And_Keypad_Backlight:
              // & do not leave LCD dark ! 
              if (blinq==-1){
                blinq=p->endq;
		p->endq++;
		p->q_depth++;
              }
              memcpy(&p->sendq[blinq].command, &p->sendq[i].command, p->sendq[i].data_length+2);
              report(RPT_QUE, "%s sending %02x,   Backlight   %02x preserved to %02x pared to (%03d)",
                 __FUNCTION__, p->sending, i, blinq, p->q_depth );
              break;
            default:
              //report(RPT_ERR, "%s sending %02x,   %02x %s removed (%03d)",
              //   __FUNCTION__, p->sending, i, pk_nm[cmd], p->q_depth );
	      break;
	  } // end switch to preserve important commands
#ifdef TIMING_REPORT
          p->ack_c[p->sendq[i].command][1]++;
#endif
          p->sendq[i].command=CF633_Dummy_Unused;
          i++;
	  pcount++;
        } while (i != endque);
        if ((pe!=p->endq) || (ps!=p->sending))
          report(RPT_QUE, "%s p->sending %02x -> %02x     p->endq %02x -> %02x",
              __FUNCTION__, ps, p->sending, pe, p->endq);
#ifdef TIMING_REPORT
        report(RPT_CRIT, "%s %010d.%06d Purged now@  %02x %02x (%03d),   %8lu g, %d p, %f %% over.",
            __FUNCTION__, p->tsent.tv_sec , p->tsent.tv_usec,
            p->sending, p->endq, p->q_depth, p->ack_c_t - p->prevp, pcount,
            (float)pcount / ((float)p->ack_c_t - (float)p->prevp ) *100);
        p->prevp=p->ack_c_t;
        p->q_purged+=pcount;
#endif
        p->q_recover=0; // not a 'natural' recovery from depth
} // end purge_que 

/** calculate CRC over given buffer with given length */
int
get_crc(char *buf, int len, int seed)
{

  /* CRC lookup table to avoid bit-shifting loops. */
  static const word crcLookupTable[256] = {
    0x00000, 0x01189, 0x02312, 0x0329B, 0x04624, 0x057AD, 0x06536, 0x074BF,
    0x08C48, 0x09DC1, 0x0AF5A, 0x0BED3, 0x0CA6C, 0x0DBE5, 0x0E97E, 0x0F8F7,
    0x01081, 0x00108, 0x03393, 0x0221A, 0x056A5, 0x0472C, 0x075B7, 0x0643E,
    0x09CC9, 0x08D40, 0x0BFDB, 0x0AE52, 0x0DAED, 0x0CB64, 0x0F9FF, 0x0E876,
    0x02102, 0x0308B, 0x00210, 0x01399, 0x06726, 0x076AF, 0x04434, 0x055BD,
    0x0AD4A, 0x0BCC3, 0x08E58, 0x09FD1, 0x0EB6E, 0x0FAE7, 0x0C87C, 0x0D9F5,
    0x03183, 0x0200A, 0x01291, 0x00318, 0x077A7, 0x0662E, 0x054B5, 0x0453C,
    0x0BDCB, 0x0AC42, 0x09ED9, 0x08F50, 0x0FBEF, 0x0EA66, 0x0D8FD, 0x0C974,
    0x04204, 0x0538D, 0x06116, 0x0709F, 0x00420, 0x015A9, 0x02732, 0x036BB,
    0x0CE4C, 0x0DFC5, 0x0ED5E, 0x0FCD7, 0x08868, 0x099E1, 0x0AB7A, 0x0BAF3,
    0x05285, 0x0430C, 0x07197, 0x0601E, 0x014a1, 0x00528, 0x037B3, 0x0263A,
    0x0DECD, 0x0CF44, 0x0FDDF, 0x0EC56, 0x098E9, 0x08960, 0x0BBFB, 0x0AA72,
    0x06306, 0x0728F, 0x04014, 0x0519D, 0x02522, 0x034AB, 0x00630, 0x017B9,
    0x0EF4E, 0x0FEC7, 0x0CC5C, 0x0DDD5, 0x0A96A, 0x0B8E3, 0x08A78, 0x09BF1,
    0x07387, 0x0620E, 0x05095, 0x0411C, 0x035A3, 0x0242A, 0x016B1, 0x00738,
    0x0FFCF, 0x0EE46, 0x0DCDD, 0x0CD54, 0x0B9EB, 0x0A862, 0x09AF9, 0x08B70,
    0x08408, 0x09581, 0x0A71A, 0x0B693, 0x0C22C, 0x0D3A5, 0x0E13E, 0x0F0B7,
    0x00840, 0x019C9, 0x02B52, 0x03ADB, 0x04E64, 0x05FED, 0x06D76, 0x07CFF,
    0x09489, 0x08500, 0x0B79B, 0x0A612, 0x0D2AD, 0x0C324, 0x0F1BF, 0x0E036,
    0x018C1, 0x00948, 0x03BD3, 0x02A5A, 0x05EE5, 0x04F6C, 0x07DF7, 0x06C7E,
    0x0A50A, 0x0B483, 0x08618, 0x09791, 0x0E32E, 0x0F2A7, 0x0C03C, 0x0D1B5,
    0x02942, 0x038CB, 0x00A50, 0x01BD9, 0x06F66, 0x07EEF, 0x04C74, 0x05DFD,
    0x0B58B, 0x0A402, 0x09699, 0x08710, 0x0F3AF, 0x0E226, 0x0D0BD, 0x0C134,
    0x039C3, 0x0284A, 0x01AD1, 0x00B58, 0x07FE7, 0x06E6E, 0x05CF5, 0x04D7C,
    0x0C60C, 0x0D785, 0x0E51E, 0x0F497, 0x08028, 0x091A1, 0x0A33A, 0x0B2B3,
    0x04A44, 0x05BCD, 0x06956, 0x078DF, 0x00C60, 0x01DE9, 0x02F72, 0x03EFB,
    0x0D68D, 0x0C704, 0x0F59F, 0x0E416, 0x090A9, 0x08120, 0x0B3BB, 0x0A232,
    0x05AC5, 0x04B4C, 0x079D7, 0x0685E, 0x01CE1, 0x00D68, 0x03FF3, 0x02E7A,
    0x0E70E, 0x0F687, 0x0C41C, 0x0D595, 0x0A12A, 0x0B0A3, 0x08238, 0x093B1,
    0x06B46, 0x07ACF, 0x04854, 0x059DD, 0x02D62, 0x03CEB, 0x00E70, 0x01FF9,
    0x0F78F, 0x0E606, 0x0D49D, 0x0C514, 0x0B1AB, 0x0A022, 0x092B9, 0x08330,
    0x07BC7, 0x06A4E, 0x058D5, 0x0495C, 0x03DE3, 0x02C6A, 0x01EF1, 0x00F78
  };

      /* initialize the new crc with the given seed */
  int newCrc = seed;

  /* This algorithm is based on the IrDA LAP example */
  while (len-- > 0)
    newCrc = (newCrc >> 8) ^ crcLookupTable[(newCrc ^ *buf++) & 0xff];

  /* Make this crc match the one's complement that is sent in the packet */
  return (~newCrc);
}




/*============================================================================
 * 635 WinTest Code.
 * Copyright 2001, Crystalfontz America, Inc. Written by Brent A. Crosby
 * www.crystalfontz.com, brent@crystalfontz.com
 *============================================================================
 */

/*---------------------------------------------------------------------------*/
/* This is some code that kind of makes the windows stuff look a little
 * like the DOS/633 interrupt driven serial stuff. Basically there is a
 * circular buffer, and SyncReceiveBuffer() uses ReadFile() to put data
 * into the circular buffer much like the DOS stuff uses an ISR to put
 * the data into the buffer. Then the rest of the functions work like
 * the counterparts in the 633,
 */



/*                  v TailPeek                        */
/* -------------------------------------------------- */
/*           ^ Tail              ^ Head               */


/** initialize/empty receive buffer by resetting its pointers */
void EmptyReceiveBuffer(ReceiveBuffer *rb)
{
  rb->head = rb->tail = rb->peek = 0;
}


/** read given number of bytes from given file handle into receive buffer */
void SyncReceiveBuffer(ReceiveBuffer *rb, int fd, unsigned int number)
{
  unsigned char  buffer[MAX_DATA_LENGTH];
  int    BytesRead;

  if (number > MAX_DATA_LENGTH)
    number = MAX_DATA_LENGTH;
  BytesRead = read(fd, buffer, number);

  if (BytesRead == -1) {
    /* fprintf(stderr,"~~~Problem reading: %s .\n", strerror(errno)); */
  }
  else {
    int  i;

    /* fprintf(stderr,"Read %d Bytes:", BytesRead); */

    /* wrap write pointer to the receive buffer */
    rb->head %= RECEIVEBUFFERSIZE;

    /* store the bytes read */
    for (i = 0; i < BytesRead; i++) {
      /* fprintf(stderr," %02x", buffer[i]); */
      rb->contents[rb->head] = buffer[i];

      /* increment write pointer (wrap if needed) */
      rb->head = (rb->head + 1) % RECEIVEBUFFERSIZE;
    }
    /* fprintf(stderr,"\n"); */
  }
}


/** return number of bytes available for reading in receive buffer */
int BytesAvail(ReceiveBuffer *rb)
{
  int avail_bytes = rb->head - rb->tail;

  if (avail_bytes < 0)
    avail_bytes += RECEIVEBUFFERSIZE;

  return(avail_bytes % RECEIVEBUFFERSIZE);
}


/** get next byte from receive buffer (return '\0' if buffer is empty) */
unsigned char GetByte(ReceiveBuffer *rb)
{
  unsigned char    return_byte = '\0';
  
  /* wrap read pointer to the receive buffer */
  rb->tail %= RECEIVEBUFFERSIZE;

  /* See if there are any more bytes available. */
  if (rb->tail != (rb->head % RECEIVEBUFFERSIZE)) {
    /* There is at least one more byte. */
    return_byte = rb->contents[rb->tail];

    /* Increment read pointer (wrap if needed) */
    rb->tail = (rb->tail + 1) % RECEIVEBUFFERSIZE;
  }

  return(return_byte);
}


/** return number of bytes available for peeking in receive buffer */
int PeekBytesAvail(ReceiveBuffer *rb)
{
  int avail_bytes = rb->head - rb->peek;
  
  if (avail_bytes < 0)
    avail_bytes += RECEIVEBUFFERSIZE;

  return(avail_bytes % RECEIVEBUFFERSIZE);
}


/** sync peek pointer with read pointer */
void SyncPeekPointer(ReceiveBuffer *rb)
{
  rb->peek = rb->tail;
}


/** accept ppeked data by syncing the read pointer to the peek pointer */
void AcceptPeekedData(ReceiveBuffer *rb)
{
  rb->tail = rb->peek;
}


/** peek next byte from receive buffer (return '\0' if buffer is empty) */
unsigned char PeekByte(ReceiveBuffer *rb)
{
  unsigned char return_byte = '\0';

  /* wrap peek pointer to the receive buffer */
  rb->peek %= RECEIVEBUFFERSIZE;

  /* See if there are any more bytes available. */
  if (rb->peek != (rb->head % RECEIVEBUFFERSIZE)) {
    /* There is at least one more byte. */
    return_byte = rb->contents[rb->peek];

    /* Increment the peek pointer (wrap if needed). */
    rb->peek = (rb->peek + 1) % RECEIVEBUFFERSIZE;
  }

  return(return_byte);
}

/*============================================================================
 *                              check_for_packet()
 *
 * check_for_packet() will see if there is a valid packet in the input buffer.
 * If there is, it will copy it into p->incoming_command and return 1. If there
 * is not it will return 0. incoming_packet may get partially filled with
 * garbage if there is not a valid packet available.
 *----------------------------------------------------------------------------
 */

/* Let's return
 * O if we have no message but we should try again immediatly
 * 1 if we have a message correctly identified
 * 2 if we have no message and we should not retry until new input
 * So a loop should run as long as we have no 0
 * If we have a 2 we should avoid comming back there.
 */
int
check_for_packet(Driver *drvthis, unsigned char expected_length)
{
  PrivateData *p = drvthis->private_data;
  int i;
  int testcrc;

  SyncReceiveBuffer(&p->receivebuffer, p->fd, expected_length);

  //First off, there must be at least 4 bytes available in the input stream
  //for there to be a valid command in it (command, length, no data, CRC).
  if (BytesAvail(&p->receivebuffer) < 4) {
    /* fprintf(stderr,"Not enough byte available for even the smallest message.\n"); */
    return(GIVE_UP); /* We don't need to retry before more byte are received */
  }

  /* Look into the buffer without removing the data. */
  SyncPeekPointer(&p->receivebuffer);

  p->incoming_command.command = PeekByte(&p->receivebuffer);

  /* Only commands 0 through MAX_COMMAND are valid */
  if (MAX_COMMAND < (0x3F & p->incoming_command.command)) {
    /* Throw out one byte of garbage. Next pass through should re-sync. */
    GetByte(&p->receivebuffer);
    /* fprintf(stderr,"###: Unknown command.\n"); */
    return(TRY_AGAIN);
  }
  
    /* There is a valid command byte. Get the data_length. */
  p->incoming_command.data_length = PeekByte(&p->receivebuffer);
  
    /* The data length must be within reason. */
  if(MAX_DATA_LENGTH < p->incoming_command.data_length) {
    //Throw out one byte of garbage. Next pass through should re-sync.
    GetByte(&p->receivebuffer);
    /* fprintf(stderr,"###: Too long packet: %d.\n", p->incoming_command.data_length); */
    return(TRY_AGAIN);
  }

  // Now there must be at least p->incoming_command.data_length + sizeof(CRC) bytes
  // still available for us to continue.
  if ((int) PeekBytesAvail(&p->receivebuffer) < (p->incoming_command.data_length + 2)) {
    //It looked like a valid start of a packet, but it does not look
    //like the complete packet has been received yet.
    /* fprintf(stderr,"Not enough read to check the complete message.\n"); */
    return(GIVE_UP); /* Let's not return until more byte are available */
  }

  /* There is enough data to make a packet. Transfer over the data. */
  for (i = 0; i < p->incoming_command.data_length; i++)
    p->incoming_command.data[i] = PeekByte(&p->receivebuffer);

  //Now move over the CRC.
  p->incoming_command.crc.as_bytes[0] = PeekByte(&p->receivebuffer);
  p->incoming_command.crc.as_bytes[1] = PeekByte(&p->receivebuffer);
  //Now check the CRC.

  //Compute the expected CheckSum
  testcrc = get_crc((unsigned char *) &p->incoming_command,
        p->incoming_command.data_length+2, 0xFFFF);
  testcrc = testcrc & 0xFFFF; /* This is TRICKY */

  if(p->incoming_command.crc.as_word == testcrc) {
    //This is a good packet. I'll be horn swaggled. Remove the packet
    //from the serial buffer.
    AcceptPeekedData(&p->receivebuffer);
    //Let our caller know that p->incoming_command has good stuff in it.
    report(RPT_DEBUG, "%s p->incoming_command has good stuff in it:", __FUNCTION__);
    show_packet(drvthis, RPT_DEBUG, &p->incoming_command); 

    return(GOOD_MSG);
    // note: p->incoming_command.data[p->incoming_command.datalength+1]
    // is unknown, so printing p->incoming_command.data, if it is text
    // is not gauranteed a string terminator \0 \n ...
  }

  /* The CRC did not match. Toss out one byte of garbage.
  * Next pass through should re-sync.
  */
  GetByte(&p->receivebuffer);
  /* fprintf(stderr,"###: Wrong CheckSum. computed/real %04x:%04x \n",
     testcrc, p->incoming_command.crc.as_word); */
  return(TRY_AGAIN);
}

//============================================================================
void show_packet(Driver* drvthis, int rep_level,  COMMAND_PACKET *packet)
{
  PrivateData *p = drvthis->private_data;
        char target[8];
        char packstr[155]; // MAX_DATA_LENGTH 22 *3 + name + ~36 + slop
            // longest I've seen is ~ 126 ?
        int usedlength=0;
        int i, cmd, top, len;

        top=(0xC0&(packet->command)) >> 6;
        cmd=(0x3F&(packet->command));
        len=packet->data_length;
  if (top == 2){
    if (cmd == CF633_Write_User_Flash_Area) // a temp report
          sprintf(packstr, "Packet (top %02d, %s Report) len %02d [", top, p->ds18[packet->data[0]].name, len);
    if (cmd == CF633_Get_Hardware_And_Firmware_Version ) // a fan report
          sprintf(packstr, "Packet (top %02d, %s Report) len %02d [", top, p->fans[packet->data[0]].name, len);
  }
  else { // is a normal response packet or error packet
          sprintf(packstr, "Packet (top %02d, cmd %02d %s ) len %02d [", top, cmd, pk_nm[cmd]+6, len);
  }
  usedlength =strlen(packstr);
  char istext [MAX_DATA_LENGTH];
  if (packet->data_length!=0) {
    for(i=0 ;i < packet->data_length; i++){
      sprintf(target, "%02x ", packet->data[i]);
      sprintf(packstr+usedlength , "%s ",target );
      usedlength+=strlen(target);
      if ((cmd==CF633_Send_Data_to_LCD) && ( i > 1 )){
        istext[i-2]=packet->data[i];
        istext[i-1]=0;
      }
    }
  }
  sprintf(packstr+usedlength, "]\n");
        report(rep_level, "%s", packstr);
  if (cmd==CF633_Send_Data_to_LCD){
    report(rep_level, "     sData_to_LCD     %02d at %02d x %d '%s'",
        packet->data_length-2, packet->data[0],
        packet->data[1], istext);
  }
  report(RPT_DEBUG, "%s strlen is %d, data_length*3 is %d",
      __FUNCTION__ , strlen(packstr), packet->data_length *3);
}

//============================================================================
/* multithreaded : 
 * do not change anything in p -> outgoing command -- might be in use elsewhere
 * do only one write(), no chopping up any remaining incremental writes elsewhere
 */
void commit_packet(Driver* drvthis,  COMMAND_PACKET *packet)
{
  PrivateData *p = drvthis->private_data;
  word tmp_crc;         
  ubyte send_this[2+MAX_DATA_LENGTH]; 

  send_this[0]=packet->command;  
  send_this[1]=packet->data_length;  
      
  if (packet->data_length > 0){
    //write(p->fd, packet.data, packet.data_length);
    memcpy(&send_this[2], &packet->data, packet->data_length);
      }
  tmp_crc= get_crc((unsigned char *) packet, 
        packet->data_length
        + 2, 0xFFFF);
  memcpy(&send_this[2+packet->data_length], &tmp_crc, 2);
  write(p->fd, send_this, 4 + packet->data_length);
}

int
fan_loop(Driver * drvthis)
{
  PrivateData *p = drvthis->private_data;
  // Main loop
  int Quit=0;
  while(!Quit)
  {
    struct timeval now;
    gettimeofday(&now,NULL);
    // 31-( estimated maximum possible que depth in seconds ) 
    // keep fans from spinning up when que gets way behind
    if  (time_diff(drvthis, p->tdid_power, now) > 28000000){
      p->need_fan_pow_ack+=1;
      report(RPT_FANS, "%s need fan pow ack+1   failsafe %ld.%06ld  %f", 
          __FUNCTION__, now.tv_sec- 1145688595 , now.tv_usec,
          (float)time_diff(drvthis, p->tdid_power, now)/1000000
          );
    }
    manage_fans  (drvthis);
    if ( p->need_fan_pow_ack!=0  ){
      // could put multiple fan pow packets into que ...
      p->treq_power=now; 
      set_fan_pow (drvthis);
      if (p->need_fan_pow_ack%8==0)
        report(RPT_FANS, "%s need fan pow ack=%d",
            __FUNCTION__, p->need_fan_pow_ack);
      // reporting settings do not time out
      // and only change when power changes
      if ( p->need_fan_rep_ack!=0  ){
        //  if report settings have changed
        //  or no _ack of previous set
        set_fan_rpt (drvthis);
        if (p->need_fan_rep_ack%8==0)
          report(RPT_FANS, "%s need fan rep ack=%d",
            __FUNCTION__, p->need_fan_rep_ack);
      }
    }
    usleep(125000); // 1/8 sec fan loop
  } // end while(!Quit)
  return 1;
} // end fan_loop

static void
textlog (Driver *drvthis) 
{
    PrivateData * p = drvthis->private_data; 
  //char b[formula maximum possible length];  
  char b[1024];
  int counter;
  char buf[255];
  int lineused=0;

  if (p->log_sens==0) {
    report(RPT_CLIUP, "%s return on syslogsensors==0", __FUNCTION__);
    return;
  }
  if (p->bad_time_to_log==1){ // if sensors currently updating
    counter=time_diff(drvthis, p->ttemps, p->tdid_tlog);
    report(RPT_CLIUP, "%s return on bad_time_to_log", __FUNCTION__);
    report(RPT_CLIUP, "%s tdiff        = %06d %06d", __FUNCTION__,
         counter/1000000,  counter-1000000*(counter/1000000));
    return;
  }
#ifdef DO_HDD
  if (p->hdd_count != 0 )   hddtemp_temps(drvthis);
#endif // DO_HDD
  gettimeofday(&p->tdid_tlog, NULL);
  sprintf (buf, "CF633data");
  sprintf (b+lineused, buf);
  lineused += strlen(buf);
#ifdef DO_ULT  
  if (p->log_uptime !=0) { 
    double uptime =0;
    double idle =0;
    machine_get_uptime(&uptime, &idle);
    sprintf (buf,",%.0f", uptime);
    sprintf (b+lineused, buf);
    lineused += strlen(buf);
  }
  if ((p->log_load !=0 ) || ( p->log_load_s!=0) || (p->log_load_u!=0)){
    store_load(drvthis);
    if (p->log_load !=0) {
      sprintf (buf,",%03.1f", p->load_t);
      sprintf (b+lineused, buf);
      lineused += strlen(buf);
    }
    if (p->log_load_s !=0) {
      sprintf (buf,",%03.1f", p->load_s);
      sprintf (b+lineused, buf);
      lineused += strlen(buf);
    }
    if (p->log_load_u !=0) {
      sprintf (buf,",%03.1f", p->load_u);
      sprintf (b+lineused, buf);
      lineused += strlen(buf);
    }
  }
  /* top values */
  if ((p->topnum!=0) || (p->topname!=0)){
    store_tops(drvthis);
    int i;
    if (p->topnum!=0) { // if doing top by rank 
      for (i=0; i < p->topnum; i++){ // for each name tracked 
        sprintf(buf, ",%s,%2.2f",p->topnumnames[i], p->topnumval[i] ); 
        sprintf (b+lineused, buf);
        lineused += strlen(buf);
      }
    }
    if (p->topname!=0) { // if doing top by name
      for (i=0; i < p->topname; i++){ // for each name tracked 
        sprintf(buf, ",%2.2f", p->topnameval[i] ); 
        sprintf (b+lineused, buf);
        lineused += strlen(buf);
      }
    }
  }
#endif // DO_ULT
#ifdef DO_LM
  if (p->lmnum !=0) {
    store_lm(drvthis);
    for (counter=0; counter < p->lmnum ; counter ++){ // lmsensors data
      if (p->lmval[counter]==666) {
        sprintf(buf, ",%derr", counter+1);
      }
      else {
        if (  (float) ((int)p->lmval[counter]) == p->lmval[counter]  )
          sprintf(buf, ",%d", (int)p->lmval[counter]);
        else
          sprintf(buf, ",%f", p->lmval[counter]);
      }
      sprintf (b+lineused, buf);
      lineused += strlen(buf);
    }
  }
#endif // DO_LM
#ifdef DO_HDD
  if (p->hdd_count !=0) {
    for (counter=0; counter  < p->hdd_count ; counter ++){  // for each hdd
      sprintf (buf,",%s",p->hddtemp[counter] );
      sprintf (b+lineused, buf);
      lineused += strlen(buf);
    }// end each drive
  }
#endif // DO_HDD
  for (counter=0; counter  < 4; counter ++){  // for each fan  
    if (p->fans[counter].p_log!=0) {
      // if this fan logs power setting
      sprintf (buf,",%03d",p->f_p_mask[counter] );
      sprintf (b+lineused, buf);
      lineused += strlen(buf);
    }
    if (p->fans[counter].autorep!=0) { 
      // if this logs sensing
      if (p->fans[counter].divisor < 20) { // fans 
        sprintf (buf, ",%04d",
          (int)p->fans[counter].rpm_rpt);
      }
      else {  // other divisors & display units
        sprintf (buf, ",%f", 
          p->fans[counter].rpm_rpt );
      }
      sprintf (b+lineused, buf);
      lineused += strlen(buf);
    }
  } // done with fan  info
  for (counter=0; counter < p->tempcount; counter ++){  
    // add temp info  
    sprintf (buf, ",%2.4f",
      (float)p->ds18[counter].ds18 /16 );
    sprintf (b+lineused, buf);
    lineused += strlen(buf);
  } // done with temp info
  /* committing to separate logfile : put time at start of this
   * in b, with time in epoch format ?
   * coordinate with ~1402 data_col
   */
#ifdef MY_CUSTOM_LINE
  update_wattsline(drvthis,0); // only other call to this is when watts is active screen
  sprintf (buf, ",%3.2f", watts_est );
  sprintf (b+lineused, buf);
  lineused += strlen(buf);
#endif
  if (p->log_fd !=NULL) { 
    char log_buf[strlen(b)+27];
    time_t very_now=time(NULL);
    ctime_r(&very_now, log_buf);
    snprintf (&log_buf[strlen(log_buf)-1], strlen("\n"), "\n");
    snprintf (&log_buf[strlen(log_buf)], strlen(b)-8, &b[9]);
    fprintf(p->log_fd, "%s\n", log_buf);
    fflush(p->log_fd);
  }
} // end of textlog

static void manage_fans  (Driver *drvthis) {
  int fan;
  struct timeval now;
    PrivateData * p = drvthis->private_data; 
  gettimeofday(&now,NULL);
  for (fan=0; fan < 4 ; fan ++) {
    if ((  time_diff(drvthis,p->fans[fan].tdid_fan, now) >=
        (double)p->fans[fan].a_sec*1000000)
        && ( p->fans[fan].sensor != 0 ) ){
    // calcs for sensor driven fans
    // adjust this fan now 
      report(RPT_FANS, "%s fan %d timeout             did %ld.%06ld",
        __FUNCTION__, fan+1,
        p->fans[fan].tdid_fan.tv_sec- 1145688595 ,
        p->fans[fan].tdid_fan.tv_usec);  
      report(RPT_FANS, "%s                          now %ld.%06ld  %6.6f",
        __FUNCTION__, now.tv_sec- 1145688595,
        now.tv_usec,
        (float)time_diff (drvthis, p->fans[fan].tdid_fan,now)
            / 1000000);  
      unsigned char tmp_r_mask=
          p->fans[fan].autorep *
          ( p->f_p_mask[fan]!=0 );
      float temperat=(p->fans[fan].sen_ref==0) ?
          p->ds18[p->fans[fan].sensor-1].ds18: 
          p->ds18[p->fans[fan].sensor-1].ds18
          -p->ds18[p->fans[fan].sen_ref-1].ds18;
      float tmp_power=FAN_LOW_POWER_PERCENT+
          (temperat-p->fans[fan].by_s_off) / 
          (p->fans[fan].by_s_100-
          p->fans[fan].by_s_off)
          *(100-FAN_LOW_POWER_PERCENT);
      if ( tmp_power > 100){
        tmp_power=100;
      }
      else if (tmp_power < FAN_LOW_POWER_PERCENT) {
        tmp_power=0;
        p->fans[fan].rpm_rpt=0; 
      }
      if ( (int)tmp_power != p->f_p_mask[fan]) {
        p->treq_power=now;
        p->need_fan_pow_ack+=1;
        report(RPT_FANS, "%s need fan pow ack+=1 fan %d",
          __FUNCTION__, fan+1); 
        p->treq_power=now;
      }
      p->f_p_mask[fan]=(int)tmp_power;
      // turn off reporting for powered off fans
      if ( tmp_r_mask != p->fans[fan].autorep *
            ( p->f_p_mask[fan]!=0 )) {
        // turn off reporting for powered off fans
        p->need_fan_rep_ack+=1;
      }
      report(RPT_FANS, 
        "%s Calc f %d: sens %d(%3.4f), sref %d(%3.4f), temp %3.4f, pow %3d(%3d).",
        __FUNCTION__, fan+1, p->fans[fan].sensor,
        (float)p->ds18[p->fans[fan].sensor-1].ds18/16,
        p->fans[fan].sen_ref, 
        (float)p->ds18[p->fans[fan].sen_ref-1].ds18/16 *
          (p->fans[fan].sen_ref!=0),
        temperat/16, p->f_p_mask[fan], 
          (p->f_p_mask[fan]>FAN_LOW_POWER_PERCENT ?
           (int)((p->f_p_mask[fan]-
          FAN_LOW_POWER_PERCENT)*2.5) :0) );
      p->fans[fan].tdid_fan=now;
    } // end it is this sensor driven fan's time
  } // end check for each fan
}// end  manage_fans

/*
 * Detect ds18 temperature sensors.
 */
int 
CFontz633_ds18_init (Driver* drvthis)
{
  PrivateData *p = drvthis->private_data;
  unsigned char retry=3;
  unsigned char this_retry=0;

  unsigned char sensors=0;
  unsigned char detected=0;
  unsigned char s_able_m[4]={0,0,0,0};
  unsigned char s_m_mult=1;
  
  COMMAND_PACKET dow;
  dow.command=CF633_Read_DOW_Device_Information;
  dow.data_length=1;

  do {
    dow.data[0]=sensors;
    dow.data_length=1;
    send_packet(drvthis, &dow);
    usleep(50000); // probe ds18 -> response, /100 on adding reciepts
    if (sensors==0) usleep(350000); // 1st ds18 probe respond delay
    //  more likely to detect all on 1st try with a pause ?
    // is the long pause for 1st report dependent on # of sensors?
    //  usleep(005000*p->tempcount) perhaps
    report(RPT_DEBUG, 
        "%s sent[%d]", __FUNCTION__ , sensors);
    do {
      usleep(10);  // wait for sensor # wanted
      CFontz633_await_packet(drvthis, &dow, 0);
      //show_packet(drvthis, RPT_INFO, &p->incoming_command);
    } while (p->incoming_command.data[0] !=sensors );
    report(RPT_DEBUG, "%s wait got [%d]", __FUNCTION__, sensors);
    if ((p->incoming_command.data[1]!=0) && 
        (sensors < p->tempcount)) {
      // collect ROM ID's of attatched sensors  
      int counter;
      p->ds18->cfromid[0]=p->incoming_command.data[1];
      p->ds18->romid[0]=p->incoming_command.data[1];
      // id[0] is sensor type, 0x28 =ds18b20
      /* I could not make sense of the CF ROM ID's
       * as returned by CF633_Read_DOW_Device_Information:
       * for id[1|2|3] the following does an 8-bit-byte 
       * endian reversal.  The 633 appears to order 
       * the sensors 0-31 by what is romid[1] here
       * ( then romid[2], then romid[3] I'm guessing ...
       * which may also be reversed  in order from
       * the manufacturer's 6,...3,2,1.  romids[4,5,6] are 
       * apparently as of yet unused ??? Q4 2004 
       * [7] is a CRC. */
      for (counter=2; counter <9; counter++) {
        p->ds18->cfromid[counter-1]=
          p->incoming_command.data[counter];
        }
            for (counter=1; counter <8; counter++) {
                    p->ds18->romid[counter] = 
          ((128&(p->ds18->cfromid[counter]))>>7)+
          (( 64&(p->ds18->cfromid[counter]))>>5)+
          (( 32&(p->ds18->cfromid[counter]))>>3)+
          (( 16&(p->ds18->cfromid[counter]))>>1)+
          (( 8& (p->ds18->cfromid[counter]))<<1)+
          (( 4& (p->ds18->cfromid[counter]))<<3)+
          (( 2& (p->ds18->cfromid[counter]))<<5)+
          (( 1& (p->ds18->cfromid[counter]))<<7);
            }
      report(RPT_INFO, 
        "%s sensor %02d type %02x id %02x %02x %02x %02x %02x %02x %02x : cf %02x %02x %02x %02x %02x %02x %02x",
        __FUNCTION__, sensors, p->ds18->cfromid[0],
        p->ds18->romid[1],  p->ds18->romid[2], 
        p->ds18->romid[3],  p->ds18->romid[4],  
        p->ds18->romid[5],  p->ds18->romid[6],  
        p->ds18->romid[7],
        p->ds18->cfromid[1], p->ds18->cfromid[2],
        p->ds18->cfromid[3], p->ds18->cfromid[4],
        p->ds18->cfromid[5], p->ds18->cfromid[6],  
        p->ds18->cfromid[7]);
      report(RPT_DEBUG, "%s s_enable_mask[%d] += %3d",
          __FUNCTION__, (int)(sensors/8),s_m_mult );
      s_able_m[(int)(sensors/8)]+=s_m_mult;
      if (s_m_mult == 128 ) {
        s_m_mult=1;
      }
      else {
        s_m_mult*=2;
      }
      
      detected ++;
      sensors ++;
    } // endif do not enable unrequested sensors
    else if ( (p->incoming_command.data[1]!=0) && 
        (sensors >= p->tempcount)) {
      detected ++;
      sensors ++;
    }
    else if ( (p->incoming_command.data[1]==0) && 
        (sensors >= p->tempcount)) {
      sensors ++;
    }
    if ((p->incoming_command.data[1]==0) && 
        (p->tempcount > detected ) && 
        (this_retry < retry -1  ) ){
      report(RPT_INFO, 
        "%s detected %d of suggested %d Temperatures, attempt # %d, retry.",
        __FUNCTION__, detected, p->tempcount, this_retry );
      report(RPT_DEBUG, 
        "%s s_able_m built to [%d %d %d %d] ",
        __FUNCTION__, s_able_m[0],s_able_m[1],
        s_able_m[2],s_able_m[3] );
      sensors=0;
      detected=0;
      s_able_m[0]=0;
      s_able_m[1]=0;
      s_able_m[2]=0;
      s_able_m[3]=0;
      report(RPT_DEBUG, "%s s_able_m reset to [%d %d %d %d] ",
        __FUNCTION__, s_able_m[0],s_able_m[1],
        s_able_m[2],s_able_m[3] );
      s_m_mult=1;
      this_retry++;
    }
    else if ( (p->incoming_command.data[1]==0)  && 
        ( this_retry == retry -1 )){
      report(RPT_INFO, 
        "%s detected %d of suggested %d Temperatures, attempt # %d, no retry.",
        __FUNCTION__, detected, p->tempcount, this_retry );
      this_retry++;
    }
    report(RPT_DEBUG, "%s this_retry %d sensors %d", 
      __FUNCTION__, this_retry, sensors);
  } while ( (this_retry < retry) && (detected < p->tempcount) );
  
  if (p->tempcount == detected) { 
    report(RPT_INFO, 
      "%s %d Tmp request; %d Tmp detect, attempt %d.",
      __FUNCTION__, p->tempcount, detected, this_retry);
  }
  else if (p->tempcount > detected) { 
    report(RPT_WARNING, 
      "%s %d temps suggested; cont %d detected.",
      __FUNCTION__, p->tempcount, detected );
    p->tempcount=detected;
  }
  else if (p->tempcount < detected) { 
    report(RPT_WARNING, "%s %d temps avail; cont with %d as requested.",
      __FUNCTION__,  detected, p->tempcount );
  }
  report(RPT_DEBUG, "%s s_able_m[%3d][%3d][%3d][%3d]", __FUNCTION__,
    s_able_m[0], s_able_m[1],
    s_able_m[2], s_able_m[3]);
  if (p->tempcount !=0) {
    char buf[255];
    // BUFSIZE is 255 from ../../shared/configfile.c
    char target[64];
    for (sensors =0; sensors < p->tempcount; sensors++){ 
      sprintf(target, "633T%d_Name", sensors+1);  
      p->ds18[sensors].ds18= DEFAULT_DS18_TEMP;
      snprintf(buf, sizeof(buf), 
        drvthis->config_get_string
        (drvthis->name,target, 0, target));
      if ((strlen(buf)==0) || (*buf==*target)) {
        // empty or absent line from config
        sprintf (buf, DEFAULT_DS18_NAME);
        // add sensor number to default name
        snprintf (buf+ strlen(DEFAULT_DS18_NAME), 3, "%02d", sensors+1);
      }
      if (( p->ds18[sensors].name =
          (char *)malloc(strlen(buf)+1)))
        sprintf(p->ds18[sensors].name, buf);
      else {
        report(RPT_ERR, 
          "%s name malloc fail.", __FUNCTION__);
        return -1;
      }
#ifdef CLIENT_CHILD
      sprintf(target, "633T%d_lcd", sensors+1);  
      int tmp_wid= drvthis->config_get_int
        (drvthis->name, target, 0, -1);
      if (( DEFAULT_WIDGET!=0) && (tmp_wid==-1))
        p->ds18[sensors].dowidget=11000+10*sensors;
      else p->ds18[sensors].dowidget=tmp_wid;
      report(RPT_DEBUG, "%s sensor %d name %s dowidget %d.",
          __FUNCTION__, sensors+1, p->ds18[sensors].name,
           p->ds18[sensors].dowidget);
#endif
    } // end of one sensor
  }
  if (p->tempcount > 0) {
    CFontz633_set_temp_report(drvthis, s_able_m);
  }
  return 1;
} // end ds18_init

/* 
 * parse fan configs, verify, fill *p->fans
 */
int 
CFontz633_fan_init (Driver* drvthis){
  PrivateData *p = drvthis->private_data;  
  char target[64];
  int fans=0;
  int tmp;
  p->need_fan_pow_ack=1;
  p->need_fan_rep_ack=1;
  gettimeofday(&p->tdid_power, NULL);
  p->tdid_power.tv_sec-=32;
  p->treq_power=p->tdid_power;
  char buf[255]; // BUFSIZE is 255 from ../../shared/configfile.c
      for (fans =0; fans <  4; fans++){
      p->fans[fans].tdid_fan=p->treq_power;
      p->fans[fans].name=NULL;
      p->fans[fans].d_label=NULL;
      char fan_safety_checks=1;
      sprintf (target, "633F%d_Name", fans+1);  
      sprintf(buf, drvthis->config_get_string (drvthis->name, target, 0, target));
      if ((strlen(buf)==0) || (*buf==*target) ){  // empty or absent 633F_Name line in config
        sprintf(buf, DEFAULT_FAN_NAME);
        snprintf (buf+ strlen(DEFAULT_FAN_NAME), 2, "%d", fans+1); // add fan number to default name
      }
      if ((p->fans[fans].name =(char *)malloc (strlen (buf)+1)))
        snprintf(p->fans[fans].name, strlen (buf)+1, buf);
      else {
        report(RPT_ERR, "%s name malloc fail.", __FUNCTION__);
        return -1;
      }
      if (strncmp(buf,"^",1)==0) {
        report(RPT_INFO, "%s 633F%d safety checking disabled.",__FUNCTION__, fans+1);
        fan_safety_checks=0;
      }

      sprintf (target, "633F%d_D_label", fans+1);  
      sprintf(buf, drvthis->config_get_string (drvthis->name, target, 0, DEFAULT_FAN_D_LABEL));
      if (strlen(buf)==0)  // empty 633F_Name line in config
        sprintf(buf, DEFAULT_FAN_D_LABEL);
      if ((p->fans[fans].d_label =(char *)malloc (strlen (buf)+1)))
        sprintf(p->fans[fans].d_label, buf);
      else {
        report(RPT_ERR, "%s  malloc RPM label fail.", __FUNCTION__);
        return -1;
      }
                          
      sprintf (target, "633F%d_Rpt", fans+1);  
      p->fans[fans].autorep=drvthis->config_get_bool(drvthis->name, target, 0, DEFAULT_FAN_AUTOREPORT);

      sprintf (target, "633F%d_Power", fans+1);  
      tmp=drvthis->config_get_int(drvthis->name, target, 0, DEFAULT_FAN_POWER);
      if ((tmp < 0) || (tmp > 100)) {
        report(RPT_WARNING, "%s 633F%d_Power=%d invalid, set to 100.", __FUNCTION__, fans+1, tmp);
        tmp=100;
      }
      p->fans[fans].power=tmp;
      p->f_p_mask[fans]=tmp;
      sprintf (target, "633F%d_P_Log", fans+1);  
      p->fans[fans].p_log=drvthis->config_get_bool(drvthis->name, target, 0, DEFAULT_FAN_LOG_POW);

      sprintf (target, "633F%d_Divisor", fans+1);  
      tmp=drvthis->config_get_int(drvthis->name, target, 0, DEFAULT_FAN_DIVISOR);
      if (tmp == 0)  {
        report(RPT_WARNING, "%s 633F%d_Divisor=%d invalid, set to %d.", __FUNCTION__,
            fans+1, tmp, DEFAULT_FAN_DIVISOR);
        tmp=DEFAULT_FAN_DIVISOR;
      }
      p->fans[fans].divisor=tmp;

      sprintf (target, "633F%d_Failsafe", fans+1);  
      p->fans[fans].failsafe=drvthis->config_get_bool(drvthis->name, target, 0, DEFAULT_FAN_FAILSAFE);
    
      sprintf (target, "633F%d_Glitch", fans+1);  
      p->fans[fans].glitch=drvthis->config_get_bool(drvthis->name, target, 0, DEFAULT_FAN_GLITCH);
      
      sprintf (target, "633F%d_Temp", fans+1);  
      p->fans[fans].sensor=drvthis->config_get_int(drvthis->name, target, 0, DEFAULT_FAN_SENSOR);
      if ((p->fans[fans].sensor  > p->tempcount ) || ( p->fans[fans].sensor  < 0 )) {
        report(RPT_WARNING, "%s 633F%d_Temp=%d, no temp %d: disabled. Fan%d now 100%%, non-adjusting."
            , __FUNCTION__, fans+1, p->fans[fans].sensor, p->fans[fans].sensor, fans+1);
        p->fans[fans].sensor=DEFAULT_FAN_SENSOR;
        p->fans[fans].power=100;
        p->f_p_mask[fans]=100;
      }
      sprintf (target, "633F%d_Ref_Temp", fans+1);  
      p->fans[fans].sen_ref=drvthis->config_get_int(drvthis->name, target, 0, DEFAULT_FAN_SENSOR);
      if ((p->fans[fans].sen_ref  > p->tempcount ) || ( p->fans[fans].sen_ref  < 0 )) {
        report(RPT_WARNING, "%s 633F%d_Ref_Temp=%d no such temp %d: feature disabled.", __FUNCTION__
            , fans+1, p->fans[fans].sen_ref, p->fans[fans].sen_ref);
        p->fans[fans].sensor=DEFAULT_FAN_SENSOR;
        p->fans[fans].sen_ref=DEFAULT_FAN_SENSOR;
        p->fans[fans].power=100;
        p->f_p_mask[fans]=100;
      }
      if (( p->fans[fans].sensor!=DEFAULT_FAN_SENSOR)&&(fan_safety_checks)) { 
        if  (p->fans[fans].p_log==0){
          p->fans[fans].p_log=1;
          report(RPT_INFO, "%s 633F%d_Temp=%d causes 633F%d_Plog=yes for safety.",
              __FUNCTION__, fans+1,  p->fans[fans].sensor, fans+1);
        }
        if  (p->fans[fans].autorep==0){
          p->fans[fans].autorep=1;
          report(RPT_INFO, "%s 633F%d_Temp=%d causes 633F%d_Rpt=yes for safety.",
              __FUNCTION__, fans+1, p->fans[fans].sensor, fans+1);
        }
        if  (p->fans[fans].failsafe==0){ 
          p->fans[fans].failsafe=1;
          report(RPT_INFO, "%s 633F%d_Temp=%d causes 633F%d_Failsafe=yes for safety.",
              __FUNCTION__, fans+1, p->fans[fans].sensor, fans+1);
        }

      } // end of simple checks for turning on sensor driven fan speed  
      if ( p->fans[fans].sensor!=DEFAULT_FAN_SENSOR) {  
        sprintf (target, "633F%d_T_off", fans+1);  
        p->fans[fans].by_s_off=drvthis->config_get_float(drvthis->name, target, 0, DEFAULT_FAN_BY_S_OFF);
        if (( p->fans[fans].by_s_off > DEFAULT_FAN_BY_S_OFF+DEFAULT_FAN_SAFETY) && (fan_safety_checks)) {
          report(RPT_INFO, "%s 633F%d_T_off=%f -> %d for safety.",
            __FUNCTION__, fans+1,  p->fans[fans].by_s_off, DEFAULT_FAN_BY_S_OFF);
          p->fans[fans].by_s_off=DEFAULT_FAN_BY_S_OFF*16;
        }
        else p->fans[fans].by_s_off=p->fans[fans].by_s_off*16;

        sprintf (target, "633F%d_T_100", fans+1);  
        p->fans[fans].by_s_100=drvthis->config_get_float(drvthis->name, target, 0, DEFAULT_FAN_BY_S_100);
        if ((p->fans[fans].by_s_100 > DEFAULT_FAN_BY_S_100+DEFAULT_FAN_SAFETY)&&(fan_safety_checks)) {
          report(RPT_INFO, "%s 633F%d_T_100=%f -> %d for safety.", __FUNCTION__, 
            fans+1, p->fans[fans].by_s_100, DEFAULT_FAN_BY_S_100);
            p->fans[fans].by_s_100 =DEFAULT_FAN_BY_S_100*16;
        }
        else p->fans[fans].by_s_100=p->fans[fans].by_s_100*16;
        if ((p->fans[fans].by_s_100 <= p->fans[fans].by_s_off)
            &&(fan_safety_checks)) {
          report(RPT_INFO, "%s 633F%d_T_100=%f is amusingly less than 633F%d_T_off=%f.",
            __FUNCTION__, fans+1, p->fans[fans].by_s_100/16, fans+1, 
            p->fans[fans].by_s_off/16);
        }
        sprintf (target, "633F%d_A_Sec", fans+1);  
        p->fans[fans].a_sec=drvthis->config_get_int(drvthis->name, target, 0, DEFAULT_FAN_A_SEC);
        if ( ((p->fans[fans].a_sec > 31) || (p->fans[fans].a_sec < 1)) &&(fan_safety_checks) ) {
          report(RPT_INFO, "%s 633F%d_A_Sec=%d  not 1-31, reset to default %d.",
            __FUNCTION__, fans+1, p->fans[fans].a_sec, DEFAULT_FAN_A_SEC);
          p->fans[fans].a_sec=DEFAULT_FAN_A_SEC;
        }
      } // end secondary temp adjust data
      else {
        p->fans[fans].by_s_off=DEFAULT_FAN_BY_S_OFF*16;
        p->fans[fans].by_s_100 =DEFAULT_FAN_BY_S_100*16;
        p->fans[fans].a_sec=DEFAULT_FAN_A_SEC;
      }
#ifdef CLIENT_CHILD  
      sprintf(target, "633F%d_lcd", fans+1);  
      tmp=drvthis->config_get_int(drvthis->name, target, 0, -1);
      if ((DEFAULT_WIDGET!=0) && (tmp==-1))
          p->fans[fans].dowidget=10000+10*fans;
      else p->fans[fans].dowidget=tmp;
#endif      
      report(RPT_INFO, 
        "%s 633F%d, %s, u %s, rpt %d, pow %d, p_log %d, div %d, fs %d, gch %d, T %s, TA %s, off %f, full %f, a_sec %d.", 
            __FUNCTION__, fans+1, p->fans[fans].name,  p->fans[fans].d_label, 
            p->fans[fans].autorep, p->fans[fans].power, p->fans[fans].p_log,
            p->fans[fans].divisor, p->fans[fans].failsafe, p->fans[fans].glitch,
            (p->fans[fans].sensor!=0) ? p->ds18[p->fans[fans].sensor-1].name: NULL,
            (p->fans[fans].sen_ref!=0) ? p->ds18[p->fans[fans].sen_ref-1].name : NULL,
            p->fans[fans].by_s_off/16, 
            p->fans[fans].by_s_100/16,
            p->fans[fans].a_sec  )  ;   
    } // end for each fan
  set_fan_pow(drvthis);
  /* set fan pow is used when up * has no await */
  COMMAND_PACKET packet;
  packet.command=CF633_Set_Fan_Power;
  CFontz633_await_packet (drvthis, &packet, 0);
  set_fan_rpt(drvthis);
  packet.command=CF633_Set_Up_Fan_Reporting;
  CFontz633_await_packet (drvthis, &packet, 0);
  /* fsg used only in init, has internal await */
  set_fan_fsg (drvthis); 
  return 1;
} //end CFontz633_fan_init

/*
 * Set fan options to values in *p->fans.
 */
void  
set_fan_fsg ( Driver* drvthis)  
{
  PrivateData *p = drvthis->private_data;  
  unsigned char fans;
  COMMAND_PACKET fan_packet;
  unsigned char bitmask;

  // set failsafe first seems safe
  fan_packet.command=CF633_Set_Fan_Power_Failsafe;   
  fan_packet.data_length=2;   
  fan_packet.data[0]=0;   
  fan_packet.data[1]=255; // 31 7/8 seconds failsafe reset
  bitmask=1;
  for (fans =0; fans <4 ;fans ++){ 
    if (p->fans[fans].failsafe==1) 
      fan_packet.data[0]+=bitmask;
    bitmask=bitmask*2;  
  }  
  report(RPT_FANS, "%s failsafe bitmask=%d, %d.", __FUNCTION__,
      fan_packet.data[0], fan_packet.data[1] );
  CFontz633_await_packet (drvthis, &fan_packet, 1);

  // glitch filtering 
  fan_packet.command=CF633_Set_Fan_Tachometer_Glitch_Filter;   
  fan_packet.data_length=1;   
  fan_packet.data[0]=0;   
  
  bitmask=1;
  for (fans =0; fans <4 ;fans ++){ 
    if (p->fans[fans].glitch==1) fan_packet.data[0]+=bitmask;
    bitmask=bitmask*2;  
  }  
  report(RPT_FANS, "%s glitch bitmask=%d.",
      __FUNCTION__, fan_packet.data[0]);
  CFontz633_await_packet (drvthis, &fan_packet, 1 );
} // end of set_fan_fsg

/*
 * Set the reporting of temperatures.
 */
static void 
CFontz633_set_temp_report(Driver *drvthis, unsigned char out[4])
{
  COMMAND_PACKET temp_report;
  temp_report.command=CF633_Set_Up_Temperature_Reporting;
  temp_report.data_length=4;
  temp_report.data[0]=out[0];
  temp_report.data[1]=out[1];
  temp_report.data[2]=out[2];
  temp_report.data[3]=out[3];
  CFontz633_await_packet (drvthis, &temp_report, 1);
}  

/* 
 *  show a line of syslog column names 
 */
void CFontz633_data_col (Driver *drvthis) {
  PrivateData *p = drvthis->private_data;  
  char syslogline[1023];
  int tmp;
  int lineused=0;

  gettimeofday(&p->tdid_tlog, NULL);
  p->tdid_tlog.tv_sec-=32;
  if (p->tempcount == 0) 
    p->bad_time_to_log=0;  // set this elsewhere?
  else
    p->bad_time_to_log=1;
  gettimeofday(&p->tdid_tlog, NULL);
  p->tdid_tlog.tv_sec-=32;
#ifdef DO_ULT  
  char buf[64], target[64]; // without lm or tops these are unused
  /* Log uptime in syslog with fan / temp info or no ? */
  p->log_uptime=drvthis->config_get_bool(drvthis->name, "LogUptime",0,0);
  if (p->log_uptime!=0) {
    snprintf (&syslogline[lineused], 8, ",uptime");
    lineused +=7;
  }
  /* Log % cpu usage in syslog with fan / temp info or no ? */
  p->log_load=drvthis->config_get_bool(drvthis->name, "LogLoad", 0, 0);
  /* Log % system cpu usage in syslog with fan / temp info or no ? */
  p->log_load_s=drvthis->config_get_bool(drvthis->name, "LogLoad_System", 0, 0);
  /* Log % user cpu usage in syslog with fan / temp info or no ? */
  p->log_load_u = drvthis->config_get_bool(drvthis->name, "LogLoad_User", 0, 0);
  if (p->log_load!=0) {
    snprintf (&syslogline[lineused], 7, ",load");
    lineused +=5;
  }
  if (p->log_load_s!=0) {
    snprintf (&syslogline[lineused],8, ",sload");
    lineused +=6;
  }
  if (p->log_load_u!=0) {
    snprintf (&syslogline[lineused], 8, ",uload");
    lineused +=6;
  }
  
  /* Log top cpu usage & how many or no ? */
  p->topnum = drvthis->config_get_int(drvthis->name, "TopNum", 0, 0);
  if (p->topnum > MAXTOP){
    report(RPT_ERR, "%s TopNum %d over maximim of %d, set to %d",
        __FUNCTION__, p->topnum, MAXTOP);
    p->topnum=MAXTOP;
  }
  for (tmp=0; tmp <p->topnum; tmp++) {
    snprintf (&syslogline[lineused], 9, ",top%d,%%", tmp+1);
    lineused +=7;
  }

  /* Top track by name */
  p->topname=0;
  for (tmp=0; tmp < MAXTOP; tmp++){ 
    sprintf(target, "TopName%d", tmp+1);  
    p->topnames[tmp]=NULL;
#ifdef CLIENT_CHILD
    p->topnamelcd[tmp]=0;
    p->topnameline[tmp]=NULL; // default no lcd line  
#endif // CLIENT_CHILD
    snprintf(buf, sizeof(buf), 
      drvthis->config_get_string
      (drvthis->name,target, 0, NULL));
    if (strlen(buf)==0)
        continue;
    if (( p->topnames[tmp] =
        (char *)malloc(strlen(buf)+1))){
      p->topname++;
      sprintf(p->topnames[tmp], buf);
      }
    else {
      report(RPT_ERR, "%s log named tops name malloc fail.", __FUNCTION__);
      p->topnames[tmp]=NULL;
      continue;
    }
    snprintf (&syslogline[lineused], 3+strlen(p->topnames[tmp]),
        ",%s%%", p->topnames[tmp]);
    lineused +=2+strlen(p->topnames[tmp]);
#ifdef CLIENT_CHILD
    sprintf(target, "TopName%d_Lcd", tmp+1);
    p->topnamelcd[tmp]=drvthis->config_get_int(drvthis->name, target, 0, 0);
    if (p->topnamelcd[tmp]!=0){
      report(RPT_CLICRE, "%s topnamelcd[%d]=%d", __FUNCTION__, tmp, p->topnamelcd[tmp] );
      p->topnameline[tmp]=malloc(p->width+1);
      if (!p->topnameline[tmp]) {
        report(RPT_ERR, "%s named tops LCD line malloc fail.", __FUNCTION__);
        p->topnameline[tmp]=NULL;
        p->topnamelcd[tmp]=0;
      }
    }
#endif // CLIENT_CHILD
  } // end of one possible topname 
#ifdef CLIENT_CHILD
  sprintf(target, "FastLcd");
  p->prettylcd=drvthis->config_get_int(drvthis->name, target, 0, 0);
#endif // CLIENT_CHILD
#endif // DO_ULT
  
#ifdef DO_LM
  /* lm-sensors */
  p->lmconf=NULL;
  snprintf(buf, sizeof(buf), drvthis->config_get_string
      (drvthis->name,"lmconf", 0, NULL));
  if (strlen(buf)==0){
    report(RPT_ERR, "%s did not read a lmconf.", __FUNCTION__);
  }
  else {
    if (( p->lmconf = (char *)malloc(strlen(buf)+1))){
      sprintf(p->lmconf, buf);
      report(RPT_LMS < RPT_CLIUP ? RPT_LMS : RPT_CLIUP, 
        "%s read lmconf=%s", __FUNCTION__, p->lmconf);
      }
    else {
      report(RPT_ERR, "%s lmconf malloc fail.", __FUNCTION__);
      p->lmconf=NULL;
    }
  }
  if (p->lmconf !=NULL) {
    if((p->lmfile = fopen(p->lmconf, "r")) == NULL)
    {
      report(RPT_ERR, "%s failed to open %s", __FUNCTION__, p->lmconf);
      free (p->lmconf);
      p->lmconf=NULL;
      p->lmnum=0;
      p->lmfile=NULL;
    }
    else{
      report(RPT_LMS < RPT_CLIUP ? RPT_LMS : RPT_CLIUP,
          "%s opened %s", __FUNCTION__, p->lmconf);
      if (sensors_init(p->lmfile) == 0){
        report(RPT_LMS < RPT_CLIUP ? RPT_LMS : RPT_CLIUP,
            "%s lm sensors_init (%s) good.",
            __FUNCTION__, p->lmconf);
      }
      else {
        report(RPT_ERR, "%s lm sensors_init(%s) failed.",
            __FUNCTION__, p->lmconf);
        free (p->lmconf);
        p->lmconf=NULL;
        p->lmnum=0;
        p->lmfile=NULL;
      }
    }
  }

  if (p->lmfile!=NULL) for (tmp=0; tmp < MAXLM; tmp++){ 
    sprintf(target, "lmName%d", tmp+1);  
    p->lmnames[tmp]=NULL;
    snprintf(buf, sizeof(buf), 
      drvthis->config_get_string
      (drvthis->name,target, 0, NULL));
    if (strlen(buf)==0){
      report(RPT_LMS < RPT_CLIUP ? RPT_LMS : RPT_CLIUP, "%s end read conf lm sensors.", __FUNCTION__);
      break;
    }
    if (( p->lmnames[tmp]=(char *)malloc(strlen(buf)+1))){
      sprintf(p->lmnames[tmp], buf);
      }
    else {
      report(RPT_ERR, "%s lmname malloc fail.", __FUNCTION__);
      p->lmnames[tmp]=NULL;
      exit(EXIT_FAILURE);
    }
    sprintf(target, "lmFeature%d", tmp+1);
    p->lmfeatu[tmp]=NULL;
    snprintf(buf, sizeof(buf), 
      drvthis->config_get_string
      (drvthis->name,target, 0, NULL));
    if(strlen(buf)==0) {
      report(RPT_ERR, "%s lacking lmsensor feature name %s for %s. End lm config.",
          __FUNCTION__, target, p->lmnames[tmp]);
      free(p->lmnames[tmp]);
      p->lmnames[tmp]=NULL;
      break;
    }
    if (( p->lmfeatu[tmp]=(char *)malloc(strlen(buf)+1))){
      sprintf(p->lmfeatu[tmp], buf);
      }
    else {
      report(RPT_ERR, "%s lmfeatu malloc fail.", __FUNCTION__);
      p->lmnames[tmp]=NULL;
      p->lmfeatu[tmp]=NULL;
      exit(EXIT_FAILURE);
    }
#ifdef CLIENT_CHILD
    sprintf(target, "lm%d_Lcd", tmp+1);
    p->lmlcd[tmp]=drvthis->config_get_int(drvthis->name, target, 0, 0);
    if (p->lmlcd[tmp]!=0){
      report(RPT_CLICRE, "%s read lm sensors lcd order[%d]=%d",
          __FUNCTION__, tmp, p->lmlcd[tmp]*10 );
    }
#endif // CLIENT_CHILD
    report(RPT_LMS < RPT_CLICRE ? RPT_LMS : RPT_CLICRE, "%s read lmName%d=%s:%s",
        __FUNCTION__, tmp+1, p->lmnames[tmp], p->lmfeatu[tmp]);
    if (!find_build_req_lm(drvthis)){
      report(RPT_ERR, "%s could not find requsted lm sensors, exit.", __FUNCTION__);
      exit(EXIT_FAILURE);
    }
    if (!strcmp(p->lmnames[tmp],"not_found"))
        continue;
    snprintf (&syslogline[lineused], 3+strlen(p->lmnames[tmp]),
        ",%s:", p->lmnames[tmp]);
    lineused +=2+strlen(p->lmnames[tmp]);
    snprintf (&syslogline[lineused], 1+strlen(p->lmfeatu[tmp]),
        "%s", p->lmfeatu[tmp]);
    lineused +=strlen(p->lmfeatu[tmp]);
  } // end of one possible lm-sensors name / feature pair 
#endif // DO_LM  
#ifdef DO_HDD  
  /* Log all hddtemps  or no ? */
  tmp = drvthis->config_get_bool(drvthis->name, "loghddtemp", 0, 0);
  if (tmp !=0){
    hddtemp_names(drvthis);
  }
  else p->hdd_count=0;
  if (p->hdd_count !=0) {
    for (tmp=0; tmp  < p->hdd_count ; tmp ++){  // for each hdd
      snprintf (&syslogline[lineused],
        strlen(p->hddname[tmp]) +2, ",%s",
        p->hddname[tmp]);
      lineused += strlen (p->hddname[tmp]) +1;
    }// end each drive
  }
#endif // DO_HDD  
  for (tmp=0; tmp <4; tmp++) {
    if (p->fans[tmp].p_log) {
      snprintf (&syslogline[lineused],
        strlen(p->fans[tmp].name) +6, ",%s:pwr",
        p->fans[tmp].name);
      lineused += strlen (p->fans[tmp].name) +5;
    }
    if (p->fans[tmp].autorep) {
      snprintf (&syslogline[lineused], strlen (p->fans[tmp].name) + strlen (p->fans[tmp].d_label) +3,
          ",%s:%s", 
          p->fans[tmp].name,p->fans[tmp].d_label);
      lineused += strlen (p->fans[tmp].name);
      lineused += strlen (p->fans[tmp].d_label) +2;
    }
  } // end of build syslogline for each fan
  if (p->tempcount > 0) {
    for (tmp=0; tmp < p->tempcount; tmp++) {
        snprintf (&syslogline[lineused],  strlen (p->ds18[tmp].name) +2,
            ",%s", p->ds18[tmp].name);
        lineused += strlen (p->ds18[tmp].name) +1;
    } // end of build syslogline for each Temperature  
  } // end of build syslogline for all Temperatures
#ifdef MY_CUSTOM_LINE
  watts=malloc(p->width+1);
  snprintf (&syslogline[lineused], 7, ",watts");
  lineused += 6;
#endif
  snprintf (&syslogline[lineused],  strlen (" \n"), " \n");
  lineused += strlen ("\n");
  //report(RPT_CRIT, "%s CF633data%s", __FUNCTION__, syslogline);
  if (p->log_fd !=NULL) {
    char log_buf[strlen(syslogline)+29];
    time_t very_now=time(NULL);
    ctime_r(&very_now, log_buf);
    snprintf (&log_buf[strlen(log_buf)-1], strlen("\n"), "\n");
    snprintf (&log_buf[strlen(log_buf)], strlen(syslogline), syslogline);
    snprintf (&log_buf[strlen(log_buf)], strlen(" \n"), " \n");
    fprintf(p->log_fd, "%s \n", log_buf);
    fflush(p->log_fd);
  }
} // end of CFontz633_data_col


/* 
 * there is a built in way to compare two structures, yes?
 * but if data_length < MAX_DATA_LENGTH then
 * unused parts of PACKETs are compared by builtin ==
 * return 1 = packets differ
 * return -1 = total equality
*/
static int CFontz633_packet_neq
  (COMMAND_PACKET *packet1, COMMAND_PACKET *packet2) {
  int counter;
/*  if (packet1->CRC.as_word != packet2->CRC.as_word) return 1;  
  if (packet1->CRC.as_bytes[0] != packet2->CRC.as_bytes[0]) return 1;  
  if (packet1->CRC.as_bytes[1] != packet2->CRC.as_bytes[1]) return 1;  
*if the client / server bit of ->command is different, CRC different
*/
  if ((0x3F&(packet1->command)) != (0x3F&(packet2->command))) return 1;  
  if (packet1->data_length != packet2->data_length) return 1;  
  for (counter=0; counter < packet1->data_length; counter++)
    if (packet1->data[counter] != packet2->data[counter]) return 1;  
   return -1;
}

/* 
 * idle until packet of type command arrives
 * packet to send, bool=send it here or no
 * (permits sending and evaluating response externally)
 */
void CFontz633_await_packet (Driver *drvthis, COMMAND_PACKET *out, int send)  {
  PrivateData *p = drvthis->private_data;  
  
  fd_set rfds;
  int retval=0;
  unsigned long spin_past_what=0;  // how many others received while waiting
  struct timeval wait, sent, check;
  int resend_repeat=0;
  ubyte command=(ubyte)out->command;
  int is_msg=-1;
  // passing &p->incoming_command to show_packet when check_for_packet returns 0
  // was crashing in show_packet ... FIXME in show_packet 

  if (p->init_completed>0) { 
    report(RPT_AWAIT, "%s %s after init_completed > 0 ", 
       __FUNCTION__, pk_nm[0x3F&(command)]+6);
    report(RPT_AWAIT, "%s programming error -> exit().", __FUNCTION__);
    exit(-1);
  }
  if (send!=0) {
    send_packet(drvthis, out);
  }
  gettimeofday (&sent, NULL);
  do {
    is_msg=check_for_packet(drvthis, MAX_DATA_LENGTH);
    // stuff is incoming ... no elaborate waiting 
    if (is_msg == TRY_AGAIN ) {
      continue;
    }
    // not a GOOD_MSG
    else if (is_msg == GIVE_UP ) {
      report(RPT_AWAIT, "%s send rep. %ld, others %ld, %s GIVE_UP",
          __FUNCTION__, resend_repeat, spin_past_what, 
          pk_nm[0x3F&(command)]+6);
    }
    // response desired
    else if ((0x3F&(p->incoming_command.command))==command) {
      report(RPT_AWAIT, 
        "%s rep. %ld, other %ld; %s -> return OK.", 
        __FUNCTION__, resend_repeat, spin_past_what,
        pk_nm[0x3F&(command)]+6);
      // show_packet(drvthis, RPT_AWAIT, &p->incoming_command);
      p->sendq[p->sending].command=CF633_Dummy_Unused;
      p->q_depth=0;
      return;
    }
    // some other good packet
    else if ((0x3F&(p->incoming_command.command))!=command) {
      spin_past_what++;
      report(RPT_AWAIT, 
        "%s send rep. %ld, ++others %ld,  awaiting %s rec %s.", 
        __FUNCTION__, resend_repeat, spin_past_what, pk_nm[0x3F&(command)]+6
        , pk_nm
        [0x3F&(p->incoming_command.command)]+6 );
      // show_packet(drvthis, RPT_AWAIT, &p->incoming_command);
      if (spin_past_what > 50) {
        report(RPT_AWAIT, 
          "%s send rep. %ld, others %ld, want %s rec %s and gave up.", 
          __FUNCTION__, resend_repeat, spin_past_what, pk_nm[0x3F&(command)],
          pk_nm [0x3F&(p->incoming_command.command)]+6 );
        p->sendq[p->sending].command=CF633_Dummy_Unused;
        p->q_depth=0;
        return;
      }
    } // end some other good packet
    
    // time to resend ?
    gettimeofday (&check, NULL);
    double since_send = time_diff (drvthis, sent, check);
    if (since_send  > 750000+1000000*(command == CF633_Reboot) ){
      report(RPT_AWAIT, "%s since send %1.6f", __FUNCTION__,
          (float)since_send /1000000);
      p->q_depth=0;
      send_packet(drvthis, out);
      gettimeofday (&sent, NULL);
      report(RPT_AWAIT, 
        "%s timeout. wait send rep. %ld, others %ld, want %s resend %d.", 
        __FUNCTION__, resend_repeat, spin_past_what, pk_nm[0x3F&(command)],
         resend_repeat+1);
      resend_repeat++;
      continue;
    }
    // did not receive anything usefull & not time to resend
    // so lets try select wait on port ...
    else {
      report(RPT_AWAIT, 
        "%s linger;  select (fd %d)", __FUNCTION__, p->fd); 
      if (command == CF633_Reboot) {
        wait.tv_sec= 2;
        wait.tv_usec=125000;
      }
      else {
        wait.tv_sec= 1;
        wait.tv_usec=750000;
      }
      FD_ZERO(&rfds);
      FD_SET(p->fd, &rfds);
      retval = select(p->fd+1, &rfds, NULL, NULL, &wait);
      if (retval == -1){
        report(RPT_CRIT, "%s select() error.  EXIT_FAILURE", __FUNCTION__);
        CFontz633_close (drvthis);  
        exit(EXIT_FAILURE);
      }
    } // end wait on fd
  } while (resend_repeat < 50);
  
  report(RPT_AWAIT, "%s resend_repeat = %d, returning.", __FUNCTION__, resend_repeat); 
  p->sendq[p->sending].command=CF633_Dummy_Unused;
  p->q_depth=0;
  return;
} // end CFontz633_await_packet

/*
 * Listen for packets 
 */
int 
listen_packets(Driver * drvthis)
{
  PrivateData * p = drvthis->private_data;
  struct timeval now;
  static int resend_timeout=250000;
  int is_msg;
  while (1) {
    //if (p->init_completed <2){
    //  continue;
    //}
    is_msg = check_for_packet(drvthis, MAX_DATA_LENGTH);
    // stuff is incoming ... no elaborate waiting 
    if (is_msg == TRY_AGAIN ) { continue; }
    else if (is_msg == GOOD_MSG ){ 
      sensors_packet(drvthis);
      // this slows down everything 5 fold on k6 300
      // is_msg=GIVE_UP; // fake it to sleep
    }
    gettimeofday (&now, NULL);
    if ( (p->sendq[p->sending].command != CF633_Dummy_Unused) 
          && (time_diff(drvthis, p->tsent, now) > (double)resend_timeout ) ){
      //by the pdf is 750000 ? for all CFx 250000 for CF635 ( .. on usb 2.0, not 1.0 ) ?
      static int resend_repeats=0;
      static COMMAND_PACKET last_repeat;
      if (0 != memcmp(&p->sendq[p->sending].command, &last_repeat.command, p->sendq[p->sending].data_length+2)) {
        resend_repeats=0;
        memcpy(&last_repeat.command, &p->sendq[p->sending].command, p->sendq[p->sending].data_length+2);
	report (RPT_INFO, "%s packet first resend %d, resend_timeout now %d",
               __FUNCTION__, resend_repeats, resend_timeout);
      } // end this is resending a new packet
      else { // multiple resends of same packet
        resend_repeats++;
	if ( resend_timeout < 950000) {
          int current_icrease = (int) ( resend_timeout - 250000 ) / 70000;
          if (current_icrease <= resend_repeats){
            resend_timeout+=70000;
	    report (RPT_INFO, "%s packet resent %d times, resend_timeout now %d",
               __FUNCTION__, resend_repeats, resend_timeout);
	    if (resend_timeout>=950000)
              report (RPT_INFO, "%s resend_timeout == maximum %d.",
                __FUNCTION__, resend_timeout);
	  } // end increase timeout now
	} // end have not hit max timeout
      } // end multiple resend of same packet
      report (RPT_INFO, "RESEND q %02x %s %02x (%03d)",
        p->sending, pk_nm[p->sendq[p->sending].command]+6,
        p->endq,
        p->q_depth);
      show_packet(drvthis, RPT_QUE, &p->sendq[p->sending]);
      commit_packet(drvthis, &p->sendq[p->sending]);
      gettimeofday (&p->tsent, NULL);
#ifdef TIMING_REPORT
      p->ack_c[p->sendq[p->sending].command][2]++;
#endif
      if (resend_repeats == 100) {
        report (RPT_CRIT, "%s resend_repeats == 100 -> CFontz633_close", __FUNCTION__);
        sleep(2); // b4 exit on too many repeats
        CFontz633_close(drvthis);
        exit(EXIT_FAILURE);
      }
      continue;
    } // end check for lost packet

    if (is_msg == GIVE_UP ){
      fd_set rfds;
      int retval;
      FD_ZERO(&rfds);
      FD_SET(p->fd, &rfds);
#ifdef TIMING_REPORT
      struct timeval maxelapse;
      gettimeofday (&now, NULL);
#endif
      struct timeval waiting;
      waiting.tv_sec=0;
      waiting.tv_usec=80000;

      retval = select(p->fd+1, &rfds, NULL, NULL, &waiting);
      if (retval == -1){
        report(RPT_CRIT, "%s select() error.  EXIT_FAILURE", __FUNCTION__);
        sleep(2); // b4 exit on elect() error 
        CFontz633_close (drvthis);  
        exit(EXIT_FAILURE);
      }
      else if (retval){
#ifdef TIMING_REPORT
        gettimeofday (&maxelapse, NULL);
        double sel_s=time_diff(drvthis, now, maxelapse);
        p->listen[2]=(p->listen[2]*(double)p->listen_t + sel_s) / (double)(++p->listen_t);
        if (sel_s <  p->listen[0]){
          p->listen[0]= sel_s;
          report(RPT_QUE, "%s new max wait %1.6f",
              __FUNCTION__, (float)p->listen[0]/1000000 );
        }
        else if (sel_s >  p->listen[1]){
          p->listen[1]= sel_s;
          report (RPT_QUE, "listen_packets new max wait %1.6f",
              (float)p->listen[1]/1000000 );
        }
#endif
      }
    } // end GIVE UP 
  } // end while(1)
  return 1;
// fixme : count resends somehow & bail when > ?
// fixme : are there any non-malformed packets that can be resent forever w/o any ack ?
// fixme : do error acks currently count as acks i& increment to next in que, or would they hang the que
} // end listen packets

static void
sensors_packet(Driver *drvthis)
  /*
   *   calling check _ for _ packet from within treat _ packet 
   *   creates infinitely deepening recursion, necessarily....
   *   is this, or something similar still so ?  
   *   threaded separation of sending & listening, with outgoing packet que ...
   */
{
  PrivateData * p = drvthis->private_data;
    int top=(0xC0&(p->incoming_command.command)) >> 6;
    int cmd=(0x3F&(p->incoming_command.command));
  if (top ==1) {
    if (p->que_fans_only) {
      if ( (cmd!=CF633_Set_Fan_Power) 
          && (cmd!=CF633_Set_Up_Fan_Reporting) 
	  && (cmd!=CF633_Set_LCD_And_Keypad_Backlight)){
        report(RPT_QUE, "%s BLITHELY ack %s & wanting q %02x %s : %02x %02x (%03d)",
          __FUNCTION__, pk_nm[cmd]+6,
          p->sending, pk_nm[p->sendq[p->sending].command]+6,
	  p->sending,  p->endq, p->q_depth-1);
        return;
      } // end is blithely ackd
    } // end que_fans_only
    else if  (cmd == p->sendq[p->sending].command) {  // is an expected ack packet
#ifdef TIMING_REPORT
      gettimeofday(&p->t_ack,NULL);
      double this_ack=time_diff(drvthis, p->tsent, p->t_ack);
      if (this_ack < p->ack_d[cmd][0]) p->ack_d[cmd][0]=this_ack;
      if (this_ack > p->ack_d[cmd][1]) p->ack_d[cmd][1]=this_ack;
      if (p->ack_d[cmd][2]==0) p->ack_d[cmd][2]=this_ack;
      else {
        p->ack_d[cmd][2]=(double)(
          ( p->ack_c[cmd][0]*p->ack_d[cmd][2] + this_ack )
          / (p->ack_c[cmd][0]+1)  ) ;
      }
    
      p->ack_c[cmd][0]++;
      p->ack_c_t++;
      p->ack_c_t_qd+=p->q_depth;  
#endif
      unsigned char next=p->sending+1;
      if (p->q_depth > 1){
        report(RPT_QUE, "%s ack q %02x (%03d) %s & send q %02x %s : %02x %02x (%03d)",
          __FUNCTION__, p->sending, p->q_depth, pk_nm[cmd]+6,
          next, pk_nm[p->sendq[next].command]+6,
	  next,  p->endq, p->q_depth-1);
        /* send the next packet */
        commit_packet(drvthis, &p->sendq[next]);
        /* say sendq[this] is now unimportant  */
        p->sendq[p->sending++].command=CF633_Dummy_Unused;
        p->q_depth--; 
        gettimeofday (&p->tsent, NULL);
      }
      else if (p->q_depth == 1 ){
        report(RPT_QUE, "%s ack q %02x %s & none q %02x %s : %02x %02x (%03d)",
          __FUNCTION__, p->sending, pk_nm[cmd]+6,
          next, pk_nm[p->sendq[next].command]+6,
	  next,  p->endq, p->q_depth-1);
//sensors_packet ack q c6 Send_Data_to_LCD & none q c7 Dummy_Unused : c6 c7 (000)
//send_packet on q     c6 Dummy_Unused adding a Dummy_Unused to que c6 c7 (001)
// is this setting  p->sendq[p->sending].command while send packet is going ?
        p->sendq[p->sending].command=CF633_Dummy_Unused;
        p->q_depth--; 
        if (p->q_recover!=0) {
          report(RPT_QUE, "%s q_depth=%d after high of %d", __FUNCTION__,
              p->q_depth, p->q_recover);
          p->q_recover=0;
          p->deepies++;
        }
      }
    } // rec'd wanted ack
    
    if (cmd == CF633_Set_Fan_Power) {
      gettimeofday(&p->tdid_power,NULL);
      report (RPT_FANS, 
        "%s fan pow upd ack rec, last req %ld.%06ld",
        __FUNCTION__, p->treq_power.tv_sec- 1145688595 ,
        p->treq_power.tv_usec);  
      report (RPT_FANS, 
        "%s fan pow upd want x %02d did %ld.%06ld  %6.6f",
        __FUNCTION__, p->need_fan_pow_ack, p->tdid_power.tv_sec- 1145688595,
        p->tdid_power.tv_usec,
        (float)time_diff(drvthis, p->treq_power,p->tdid_power)/1000000 );  
      p->need_fan_pow_ack=0;
      return;
    }
    else if (cmd == CF633_Set_Up_Fan_Reporting) {
      p->need_fan_rep_ack=0;
      report (RPT_FANS, 
        "%s fan rpt upd   success ack received.", __FUNCTION__);
      return;
    }
    else if ( // is an ack report & common cmd
      (cmd==CF633_Set_LCD_Special_Character_Data) ||
      (cmd==CF633_Set_LCD_Contents_Line_One) ||
      (cmd==CF633_Set_LCD_Contents_Line_Two) ||
      (cmd==CF633_Send_Data_to_LCD) ||
      (cmd==CF633_Clear_LCD_Screen) ||
      (cmd==CF633_Set_LCD_And_Keypad_Backlight) ||
      (cmd==CF633_Configure_Key_Reporting) ||
      (cmd==CF633_Set_LCD_Contrast) ||
      (cmd==CF633_Set_LCD_Cursor_Style) ||
      (cmd==CF633_Reboot) ||
      (cmd==CF633_Set_GPIO_Pin) ||
      (cmd==CF633_Ping_Command)  ) 
      return;
  } // end top ==1
  else if (top==2) {
    if (p->incoming_command.command == 0x80) {
      // key press or release
#ifdef TIMING_REPORT
      p->ack_rep[0]++;
#endif
      AddKeyToKeyRing(&p->keyring, p->incoming_command.data[0]);
      if (p->q_depth>25){
        purge_que(drvthis);
      }
      return;
    }
    if (p->incoming_command.command==0x82) {
      // temperature report
#ifdef TIMING_REPORT
      p->ack_rep[2]++;
#endif
      if (p->tempcount==0) return; // unsolicited this config, ignore them.
      if (p->incoming_command.data[0]==0){
        p->bad_time_to_log=1;
      }
      if (p->incoming_command.data[0]>=p->tempcount){
        report (RPT_TEMPS, 
          "%s sens_p rec sens %d, > known %d.",
          __FUNCTION__, p->incoming_command.data[0]+1, p->tempcount );
        return;  
      }
      unsigned char  msb=p->incoming_command.data[2] % 8; 
        // only 3 least significant bits are temp values
      char sign = (p->incoming_command.data[2] > 7) ? -1 : 1;
        // > 7 means subzero   
      if (sign == -1) msb = 7 - msb; 
        // maxim-ic.com's ds18b20.pdf p.3 for bit usage 
      if (p->incoming_command.data[3] == 0) 
        p->ds18[p->incoming_command.data[0]].ds18=
          DEFAULT_DS18_TEMP; // denotes bad CRC 
      else  
        p->ds18[p->incoming_command.data[0]].ds18=
          sign*(p->incoming_command.data[1]+256*msb);
        // (ds1820 / ds18b20 "abstracto-increments") / 16 
        // == degrees celcius 
      if (p->incoming_command.data[0]<p->tempcount-1) return; 
        // if not last temp report this second, skip the rest
      p->bad_time_to_log=0;
      gettimeofday(&p->ttemps,NULL);
      // time to the epoch u_sec of last temp report
      // humbly: should not claim more accurate than 1 sec
      return;
    } // end command==0x82 temperature report 
  
    else if (p->incoming_command.command==0x81)   { 
      // fan autoreport packet
#ifdef TIMING_REPORT
      p->ack_rep[1]++;
#endif
      unsigned char fan=p->incoming_command.data[0];
      if (p->f_p_mask[fan]==0) { // fan recently told to stop 
        p->incoming_command.data[1]=0;
        p->incoming_command.data[2]=0;
      }
      if (p->incoming_command.data[1]<3) {
        // device too slow to report properly, 0=complete stop
        p->fans[fan].rpm_rpt= p->incoming_command.data[1];
      }
      else {  // "27692308" is a constant 
        p->fans[fan].rpm_rpt=((
          (float)p->incoming_command.data[1]-3)*27692308) 
          / ((float)(*(unsigned short *)
          (&p->incoming_command.data[2]))
          *(float)p->fans
          [fan].divisor);
      }
      return;
    } // end of incoming_command==0x81  store  reported fan speeds  
    else {
      report(RPT_WARNING, "%s rec. a ? report packet: %d %s",
          __FUNCTION__, cmd, pk_nm[cmd]);
      show_packet (drvthis, RPT_ERR, &p->incoming_command);
      return;
    }
   }// end top == 2, ongong report packets  
  else if (top==0) { // outgoing packet here?
    if (cmd==0x52){
      /* a CF633_Read_DOW_Device_Information packet during init
      *  is always sensor 0, and happens less frequently when the
       * usleep() in ds18_init is high, above 70000
       * I'm not sure why one winds up here only sometimes.
       * Resend it and sensors get detected on 1st try, (always?)
       * with cloning & cueing, this should be somewhere else or
      * is it obviated? check fixme
      */
      report (RPT_ERR, "%s rec Read_DOW for sens %d, resend.",
        __FUNCTION__, p->incoming_command.data[0]+1 );
      COMMAND_PACKET dow;
      dow.command=CF633_Read_DOW_Device_Information;
      dow.data_length=1;
      dow.data[0]=p->incoming_command.data[0];
      send_packet(drvthis, &dow);
      show_packet(drvthis, RPT_ERR, &p->incoming_command);
      return;
    } // end DOW_Device_Information packet during init
    report (RPT_ERR, 
      "%s rec unknown outgoing packet, resend commit packet:", __FUNCTION__ );
    commit_packet(drvthis, &p->incoming_command);
    show_packet(drvthis, RPT_ERR, &p->incoming_command);
    return;
  } // end (top==0)
  else if (top==3) { // error, malformed packet 
#ifdef TIMING_REPORT
    p->ack_c[cmd][3]++;
#endif
    report (RPT_ERR, "%s rec %d %s, ERROR malformed packet.",
      __FUNCTION__, cmd, pk_nm[cmd] );
      show_packet(drvthis, RPT_ERR, &p->incoming_command);
    return;
  } // end is error report
  if (p->init_completed < 2) { // still initing 
    report(RPT_QUE, "%s rec ack (p->init_completed<2). %u %s ",
        __FUNCTION__, pk_nm[cmd] );
    return;
  } // if still p initing with await _packet ... should not happen 
  // is not a key, temp, or fan  ....
  report (RPT_ERR, "%s Unknown / Unhandled incoming packet:",
      __FUNCTION__);
  show_packet (drvthis, RPT_ERR, &p->incoming_command);
  return;  // and nothing else to do... 
} // end of sensors packet

/*
 * send fan autoreport mask
 * presumes:   p->fans[fan].autorep 
 *     p->f_p_mask[fan]
 * are properly set.
 */
void set_fan_rpt (Driver *drvthis)
{
    PrivateData * p = drvthis->private_data; 
  COMMAND_PACKET set_f_rpt;
  set_f_rpt.command=CF633_Set_Up_Fan_Reporting;
  set_f_rpt.data_length=1;
  
  unsigned char mask[4];
  unsigned char fan;
  for (fan=0; fan < 4 ; fan ++) {
    mask[fan]=( p->fans[fan].sensor==0) ? p->fans[fan].autorep :
      // non temp-adjusting fan set at .power
      p->fans[fan].autorep*(p->f_p_mask[fan]!=0) ;
      // adjusting fan that should not report 
      // if power==0
    if (mask[fan]==0) {
      // a fan told to stop reporting
      // will not report zero rpm
      p->fans[fan].rpm_rpt=0;
    }
  }
  set_f_rpt.data[0]=mask[0] +mask[1]*2 + mask[2]*4 + mask[3]*8;
  if ((p->need_fan_rep_ack > 1) && (p->q_depth > 1 )) {
    unsigned char que_scan; 
    unsigned char fr_in_q=0; 
    for ( que_scan =p->sending; que_scan != p->endq; que_scan ++) {
      if (p->sendq[que_scan].command==CF633_Set_Up_Fan_Reporting) {
        memcpy(p->sendq[que_scan].data, set_f_rpt.data ,4);
        // copied new settings into packet already in que
        fr_in_q++;
        report(RPT_FANS, "%s depth %d, fr in Q %d",
            __FUNCTION__, p->q_depth, fr_in_q  );
        if ( que_scan != p->sending){
          return;
        }
      } // end this command is fan rep
    } // end scanning que
  } // end new fan settings & existing que
  send_packet(drvthis, &set_f_rpt);
  report(RPT_FANS, "%s fan rep mask upd -> %d %d %d %d.", 
    __FUNCTION__, mask[0] ,mask[1],  mask[2], mask[3] );
} // end of set_fan_rpt

/*
 * send fan power mask
 * presumes:   p->f_p_mask[fan]
 * is properly set.
 */
void set_fan_pow (Driver *drvthis)
{
    PrivateData * p = drvthis->private_data; 
  report(RPT_FANS, "%s fan pow upd -> %3d %3d %3d %3d               %ld.%06ld", 
    __FUNCTION__, p->f_p_mask[0], p->f_p_mask[1], 
    p->f_p_mask[2], p->f_p_mask[3],
    p->tdid_power.tv_sec- 1145688595,
    p->tdid_power.tv_usec); // ~s since I wrote this... 
  if ((p->need_fan_pow_ack > 1) && (p->q_depth > 1 )) {
    unsigned char que_scan; 
    unsigned char fp_in_q=0; 
    for ( que_scan =p->sending; que_scan != p->endq; que_scan ++) {
      if (p->sendq[que_scan].command==CF633_Set_Fan_Power) {
        memcpy(p->sendq[que_scan].data, p->f_p_mask,4);
        // copied new settings into packet already in que
        fp_in_q++;
        report(RPT_FANS, "%s depth %d, fp in Q %d",
            __FUNCTION__, p->q_depth, fp_in_q  );
        if ( que_scan != p->sending){
          return;
        }
      } // end this command is fan pow
    } // end scanning que
  } // end new fan settings & existing que
  COMMAND_PACKET set_f_pow;
  set_f_pow.command=CF633_Set_Fan_Power;
  set_f_pow.data_length=4;
  memcpy(set_f_pow.data, p->f_p_mask,4);
  send_packet(drvthis, &set_f_pow);
} // end of set_fan_pow

/*
 * returns (now - past) in usec, seconds * 1000000
 */
double
time_diff (Driver *drvthis, struct timeval past, struct timeval now)
{
  //  PrivateData * p = drvthis->private_data; 
  struct timeval tdiff;
  // time_t  tv_sec;
  //   ./linux/types.h:typedef __kernel_time_t         time_t;
  //     ./asm/posix_types.h:typedef long                __kernel_time_t;
  // suseconds_t     tv_usec;
  //   ./linux/types.h:typedef __kernel_suseconds_t    suseconds_t;
  //     ./asm/posix_types.h:typedef long                __kernel_suseconds_t;
  // LONG_MAX     2 147 483 647L
  // if __WORDSIZE == 64 LONG_MAX     9223372036854775807L
  // 2 147 sec = 35 min, 47 sec @ 1000000 multi  use 1000 ?
  // why did I choose type double to return ?
  
  tdiff.tv_sec=now.tv_sec-past.tv_sec;
  if (now.tv_usec < past.tv_usec){
    tdiff.tv_sec --;
    tdiff.tv_usec = 1000000L + now.tv_usec- past.tv_usec;
  }
  else tdiff.tv_usec=now.tv_usec-past.tv_usec;
/*
  report(RPT_DEBUG, "%s %f = time_diff ( %ld.%06ld, %ld.%06ld)", __FUNCTION__,
      ((float)tdiff.tv_usec + (float)tdiff.tv_sec * 1000000)/ 1000000,
      past.tv_sec, past.tv_usec,
      now.tv_sec, now.tv_usec);
*/
  return (double)(tdiff.tv_usec + tdiff.tv_sec * 1000000L);
}


void query_hw_ver (Driver *drvthis){
    PrivateData * p = drvthis->private_data;
  COMMAND_PACKET hw_querry;  
  int model=0;
  int top;

  hw_querry.command=CF633_Get_Hardware_And_Firmware_Version;
  hw_querry.data_length=0;

    while (model==0 ){
    send_packet(drvthis, &hw_querry);
    CFontz633_await_packet (drvthis, &hw_querry , 0);
    // why two empty responses, with lag sometimes ?
    top=(0xC0&(p->incoming_command.command)) >> 6;
    if ((top==2) && (p->incoming_command.data_length==4)) {
      report(RPT_ERR, "%s receiving Fan autoreport packets!", __FUNCTION__);
      continue;
    }
    model=atoi(&p->incoming_command.data[3]);
    if (model==0 ) usleep(1000);  // b4 resending hw querry
  }
  snprintf(p->model_str, p->incoming_command.data_length+1,  p->incoming_command.data);  
  report(RPT_INFO, "%s read: (top %d len %d) \"%s\" : %d", __FUNCTION__,
      top, p->incoming_command.data_length, 
      p->model_str, model);
  if ((model != 631) && (model != 633) && (model != 635)){
    report(RPT_ERR, "%s %s (%d) unrecognized. Using .conf value %d.",
        __FUNCTION__, p->model_str, model, p->model);
    return;
  }
  if (model != p->model) {
    report(RPT_ERR, "%s %s != config specified model %d. Using %d",
        __FUNCTION__, p->model_str, p->model, model);
    p->model=model;
  }
}

void spawn_clones (Driver *drvthis) {
    PrivateData * p = drvthis->private_data;
   
  p->init_completed=2;
  //int flags=(CLONE_VM | CLONE_PARENT | CLONE_FS | CLONE_PTRACE );
  int flags=(CLONE_VM | CLONE_PARENT | CLONE_FS | CLONE_PTRACE | CLONE_THREAD | CLONE_SIGHAND );
  // CLONE_THREAD  = only 1 in top; no defunct PIDs spawning from init 
  // CLONE_SIGHAND = for sigqueue-sigaction unique threads  
  // bad flags tried * CLONE_PARENT_SETTID | CLONE_FILES
  
  p->fan_stack=calloc (1, FAN_STACK);
  p->fan_PID=clone ((void *)fan_loop,
      p->fan_stack+FAN_STACK, flags, drvthis);
  if (p->fan_PID==-1) {
      report(RPT_CRIT, "%s fan clone err %d %s.",
        __FUNCTION__, errno, strerror(errno));
    free (p->fan_stack);  
  }
  else report(RPT_INFO, "%s fan child %d = PID %d.",
      __FUNCTION__, p->fan_stack, p->fan_PID);

  p->listen_stack=calloc (1, LSN_STACK );
  p->listen_PID=clone ((void *)listen_packets,
    p->listen_stack+LSN_STACK, flags, drvthis);
  if (p->listen_PID==-1) {
    report(RPT_CRIT, "%s listen child clone err %d %s.",
      __FUNCTION__, errno, strerror(errno));
    free (p->listen_stack);
  }
  else report(RPT_INFO, "%s listen child %d = PID %d.",
    __FUNCTION__, p->listen_stack, p->listen_PID);
    
  if ( (p->log_sens==0) 
      && (p->fans[0].sensor==0) && (p->fans[0].power==100)
      && (p->fans[1].sensor==0) && (p->fans[1].power==100)
      && (p->fans[2].sensor==0) && (p->fans[2].power==100)
      && (p->fans[3].sensor==0) && (p->fans[3].power==100)){
    // power==100 checks ... is there a set fan power b4 log client ?
    report(RPT_INFO, "%s log client has no tasks, not starting.", __FUNCTION__);
  } else {
    // sets up textlog & stores to start
    gettimeofday(&p->tdid_tlog,NULL);
    p->tdid_tlog.tv_sec-=p->log_sens;
#ifdef DO_ULT
    p->tdid_load.tv_sec= p->tdid_tlog.tv_sec;
    p->tdid_load.tv_usec=p->tdid_tlog.tv_usec;
    p->tdid_top.tv_sec= p->tdid_tlog.tv_sec;
    p->tdid_top.tv_usec=p->tdid_tlog.tv_usec;
#endif // DO_ULT
#ifdef DO_LM
    p->tdid_lm.tv_sec= p->tdid_tlog.tv_sec;
    p->tdid_lm.tv_usec=p->tdid_tlog.tv_usec;
#endif // DO_LM
    p->lg_cli_stack=calloc (1, L_C_STACK );
    p->lg_cli_PID=clone ((void *)datalog_cli,
    p->lg_cli_stack+L_C_STACK, flags, drvthis);
    if (p->lg_cli_PID==-1) {
      report(RPT_CRIT, "%s data log client child clone err %d %s.",
        __FUNCTION__, errno, strerror(errno));
      free (p->lg_cli_stack);
    }
    else report(RPT_INFO, "%s data log client child %d = PID %d.",
       __FUNCTION__, p->lg_cli_stack, p->lg_cli_PID);
  }
} // end spawn_clones (p->init_completed==2) start children here

#ifdef TIMING_REPORT
void timing_report(Driver *drvthis) {
  PrivateData * p = drvthis->private_data;
  int cmd;
  
  // 2 147 sec = 35 min, 47 sec wrap 
  struct timeval now;
  gettimeofday (&now, NULL);
  double d_t=time_diff(drvthis, p->timing_mk, now);
  
  if (p->ack_rep[0] >0){   // up & down makes 2 per press
    report(RPT_CRIT, "%8lu      Key Presses", p->ack_rep[0]/2);
  }
  if (p->ack_rep[1] >0){
    report(RPT_CRIT, "%8lu      Fan Reports", p->ack_rep[1]);
  }
  if (p->ack_rep[2] >0){
    report(RPT_CRIT, "%8lu      Temperature Reports", p->ack_rep[2]);
  }
  
  double max=0, avg=0;
  long int total=0, tot_d=0, tot_r=0, tot_m = 0;
  double min=1000000;
  for (cmd =0; cmd<  MAX_COMMAND+2 ; cmd++){
    if (p->ack_c[cmd][0] >0){
      if (min > p->ack_d[cmd][0]) min =  p->ack_d[cmd][0];
      if (max < p->ack_d[cmd][1]) max =  p->ack_d[cmd][1];
      if ( avg==0 ) 
        avg = p->ack_d[cmd][2];
      else 
        avg=(double) (total*avg + p->ack_c[cmd][0]*p->ack_d[cmd][2]) / (total+ p->ack_c[cmd][0]);
      total+=p->ack_c[cmd][0];
      tot_d+=p->ack_c[cmd][1];
      tot_r+=p->ack_c[cmd][2];
      tot_m+=p->ack_c[cmd][3];
    }
  }
  report(RPT_CRIT, "   count     %%       min t     max t     avg      dump resent malf");
  for (cmd =0; cmd< MAX_COMMAND+2; cmd++)
    if (p->ack_c[cmd][0] >0){
      int pct_lh=(int)(((float)p->ack_c[cmd][0]/(float)total)*100);
      int pct_rh=(int)(((float)p->ack_c[cmd][0]/(float)total - (float)pct_lh/100)*10000);
      report(RPT_CRIT, "%8lu   %02d.%02d  %f  %f  %f %5lu %5lu %5lu  %s",
        p->ack_c[cmd][0], pct_lh, pct_rh, 
        p->ack_d[cmd][0]/1000000, p->ack_d[cmd][1]/1000000, p->ack_d[cmd][2]/1000000,
         p->ack_c[cmd][1], p->ack_c[cmd][2], p->ack_c[cmd][3], pk_nm[cmd]+6);
    }
  report(RPT_CRIT, "-----------------------------------------");
  report(RPT_CRIT, "%8lu          %f  %f  %f %5lu %5lu %5lu  All",
      total, min/1000000, max/1000000, avg/1000000, 
      tot_d, tot_r, tot_m);
  // Que Depth Average = summ of all que depth measures / packets recd
  float qda   = (float)p->ack_c_t_qd / (float) p->ack_c_t;
  int   qda_lh= (int)qda;
  int   qda_rh= (int)((((float)p->ack_c_t_qd / (float) p->ack_c_t)-(int)((float)p->ack_c_t_qd / (float) p->ack_c_t))*100);
  // Que Depth Average as seconds of time = qda * average packet time
  float qd_sec = (float)avg/1000000 * qda;
  int qd_seclh= (int) qd_sec;
  int qd_secrh= (int)((qd_sec - (float)qd_seclh)*100);  
  report(RPT_CRIT, "%s avg que depth % 3d.%02d = ~% 2d.%02d sec., max % 3u,  times over 128 % 3d",
      p->model_str, qda_lh, qda_rh,
      qd_seclh, qd_secrh, p->maxq, p->deepies, 
      (float)p->ack_c_t/((float)d_t/1000000) , (float)d_t/1000000);
  report(RPT_CRIT, "listen_packets loop wait:  min %f  max %f  avg %f.",
      p->listen[0]/1000000,  p->listen[1]/1000000,  p->listen[2]/1000000);
  report(RPT_CRIT, "%8lu good, %8lu purged, %f %% overload.     %f packets/sec, elapse %f",
      p->ack_c_t, p->q_purged,
      p->q_purged / (float)p->ack_c_t  *100,
      (float)p->ack_c_t/((float)d_t/1000000) , (float)d_t/1000000);
} // end of timing_report

#ifdef LIVE_TIMING
static void
install_signal(Driver *drvthis)
{
#ifndef WIN32
  /* Installs signal handlers te get live timing report 
   * can also receive a reload signal.
   * sigaction() is favoured over signal() */

  struct sigaction sa;

  debug( RPT_CRIT, "%s, kill -SIGUSR1 %d -> live timing report.", __FUNCTION__, getpid());
  debug( RPT_CRIT, "%s, kill -SIGUSR2 %d -> reset timing start.", __FUNCTION__, getpid());
  sigemptyset ( &(sa.sa_mask) );
  sa.sa_flags = SA_RESTART;

  sa.sa_handler = live_timing;
  sigaction (SIGUSR1, &sa, NULL);    /* kill - 30 10 16*/
  sa.sa_handler = live_timing_reset;
  sigaction (SIGUSR2, &sa, NULL);    /* kill - 30 10 16*/
#else
        /* Win32 does not support POSIX signals i.e. sigaction(). However, it does 
         * support ANSI signals in mingw. */
  sa.sa_handler = live_timing;
  signal (SIGUSR1, &sa, NULL);    /* kill - 30 10 16*/
  sa.sa_handler = live_timing_reset;
  signal (SIGUSR2, &sa, NULL);    /* kill - 30 10 16*/

#endif
} // end install_signal

static void
live_timing(int signal_number){
  timing_report(driver_pointer_global);
#ifdef STACK_USAGE
  PrivateData * p = driver_pointer_global->private_data;
  Driver *drvthis=driver_pointer_global;
  report(RPT_CRIT, "Stack: fans %5d  listen %5d log/cli %5d",
      check_clone_stack_usage (driver_pointer_global, NULL, p->fan_stack,    FAN_STACK),
      check_clone_stack_usage (driver_pointer_global, NULL, p->listen_stack, LSN_STACK),
      check_clone_stack_usage (driver_pointer_global, NULL, p->lg_cli_stack, L_C_STACK));
#endif
  clear_timing(driver_pointer_global);
}  

static void
live_timing_reset(int signal_number){
  clear_timing(driver_pointer_global);
}  

#endif // ifdef LIVE TIMING functions

void
clear_timing(Driver *drvthis){
    PrivateData * p = drvthis->private_data;
  
  gettimeofday (&p->timing_mk, NULL);  
  p->ack_c_t=0;
  p->prevp=0;
  p->ack_c_t_qd=0;
  p->q_purged=0;
  p->listen[0]=10000000;
  p->listen[1]=0;
  p->listen[2]=0;
  p->listen_t=0;
  p->maxq=0;
  p->deepies=0;
  p->q_recover=0;
  p->ack_rep[0]=p->ack_rep[1]=p->ack_rep[2]=0;
  int tmp;
  for (tmp=0; tmp < MAX_COMMAND+2; tmp ++){
    p->ack_d[tmp][0]=999999999;
    p->ack_d[tmp][1]=0;
    p->ack_d[tmp][2]=0;
    p->ack_c[tmp][0]=0;
    p->ack_c[tmp][1]=0;
    p->ack_c[tmp][2]=0;
    p->ack_c[tmp][3]=0;
  }
}  

#endif // ifdef TIMING REPORT functions

#ifdef STACK_USAGE
int check_clone_stack_usage (Driver *drvthis,  char *stack_name, void *start, int block_size){
  //PrivateData * p = drvthis->private_data;
  void *check_malloc;

  for ( check_malloc=start; check_malloc < start+block_size; check_malloc++){
    if (( *(int*)check_malloc != 0)  ) {
      if (stack_name!=NULL)
        report(RPT_CRIT, "%s %d 1st non zero= %d, used ~= %05d    %s",
          __FUNCTION__, start, check_malloc, 
          start + block_size - check_malloc, 
          stack_name );
      return start + block_size - check_malloc + 1; 
    }
  }
  return 0;
}
#endif

int datalog_cli (Driver * drvthis){
    PrivateData * p = drvthis->private_data;

  report(RPT_INFO, "%s started", __FUNCTION__);

  /* print initial column labels  */
  if (p->log_sens!=0) {
    CFontz633_data_col (drvthis);
#ifdef DO_ULT // does DO_LM depend on this too?
    if ((p->log_load !=0 ) || ( p->log_load_s!=0)
        || (p->log_load_u!=0) || (p->log_uptime!=0)){
      machine_init();
    }
#endif // DO_ULT
    textlog (drvthis);
  }
#ifdef CLIENT_CHILD
  int port;
  char host[64];

//   usleep(750000); // let server get up b4 connecting
  port=drvthis->config_get_int( "server", "port", 0, LCDPORT );
  strncpy(host, drvthis->config_get_string( "server", "bind", 0, "localhost" ), sizeof(host));
  cf_dsk = sock_connect( host, port );
  if( cf_dsk <= 0 ) {
    report(RPT_CRIT, "%s bad sock_connect", __FUNCTION__);
    return -1;
  }
  else {
    report(RPT_CLICRE, "%s socket %d opened", __FUNCTION__, cf_dsk);
  }
  report(RPT_DEBUG, "%s sock_send: %d hello", __FUNCTION__, cf_dsk);
  sock_send_string( cf_dsk, "hello\n" );
  if (read_connect_string(drvthis)==-1){
    report(RPT_ERR, "%s hello failed, exiting.", __FUNCTION__);
    exit(EXIT_FAILURE);
  }
  else {
    report(RPT_CLICRE, "%s hello ok", __FUNCTION__);
  }
  
  report(RPT_DEBUG, "%s sock_send: %d client_set -name CF633log",
      __FUNCTION__, cf_dsk );
  sock_send_string( cf_dsk, "client_set -name CF633log\n");
  wait_success(drvthis, 0, NULL);
  
  cli_scrn(drvthis);
#endif // CLIENT_CHILD
  client_loop(drvthis);
  return 1;
} // end datalog_cli

#ifdef DO_ULT 
void store_load(Driver *drvthis){
    PrivateData * p = drvthis->private_data;

  struct timeval now;
  gettimeofday(&now, NULL);
  if (time_diff(drvthis, p->tdid_load, now)<5000000) return;
  p->tdid_load=now;
    
  machine_get_load(&p->load);
  if (p->log_load !=0) {
    p->load_t=((float)p->load.user +(float)p->load.system +(float)p->load.nice)
        /(float)p->load.total*100 ;
  }
  if (p->log_load_s !=0) {
    p->load_s=(float)p->load.system/(float)p->load.total*100;
  }
  if (p->log_load_u !=0) {
    p->load_u=(float)p->load.user/(float)p->load.total * 100;
  }
}

static int
sort_jiffpid (void *a, void *b)
{
        procjiffy_type *one, *two;
        if ((a == NULL) || (b == NULL))
                return 0;
        one = (procjiffy_type *) a;
        two = (procjiffy_type *) b;
        return (two->pid < one->pid);
}

int machine_get_jiffies(Driver *drvthis, LinkedList *procs)
{
  // Much of this code was ripped from "gmemusage"
  char buf[128];
  DIR *proc;
  FILE *StatFile;
  struct dirent *procdir;
  procjiffy_type *pjt=NULL;
    //PrivateData * p = drvthis->private_data;

  char  procName[16], dummychar;
  int pid, dummyint;
  unsigned long  dummyul;
  unsigned long int utime, stime; // , cutime, cstime;
  if((proc = opendir("/proc")) == NULL)
  {
    report(RPT_CRIT, "%s unable to open /proc.", __FUNCTION__);
    return(-1);
  }
  while((procdir = readdir(proc)))
  {
    if(!index("1234567890", procdir->d_name[0]))
      continue;
    sprintf(buf, "/proc/%s/stat", procdir->d_name);
    if((StatFile = fopen(buf, "r")) == NULL)
    {
      // Not a serious error; process has finished before we could
      // examine it:
      continue;
    }
    utime= stime =0; // cutime = cstime=0;
    fscanf(StatFile,
      "%d %s %c %d %d %d %d %d %lu %lu %lu %lu %lu %lu %lu", //  %lu %lu",  /* utime stime cutime cstime */
            &pid, procName, &dummychar, 
      &dummyint, &dummyint, &dummyint, &dummyint, &dummyint,
      &dummyul, &dummyul, &dummyul, &dummyul, &dummyul,
      &utime, &stime); //, &cutime, &cstime); 
    fclose(StatFile);
    LL_Rewind(procs);
    pjt = malloc(sizeof(procjiffy_type));
    if(!pjt) {
      report(RPT_ERR, "%s Error allocating process entry", __FUNCTION__);
      break;
    }
    char *tmpstart, *tmpend;
    tmpstart = strchr(procName, '(') + 1;
    tmpend = strrchr(tmpstart, ')');
    snprintf(pjt->name, tmpend-tmpstart+1, tmpstart);
/*    report(RPT_ERR, "%s jiffy %s  %lu = %lu + %lu + %lu + %lu",
        __FUNCTION__,  pjt->name,
        utime+ stime, // + cutime+ cstime,
        utime, stime); //, cutime, cstime);
*/    pjt->utime=utime;
    pjt->stime=stime;
    //pjt->cutime=cutime;
    //pjt->cstime=cstime;
    pjt->pid=pid;
    // TODO:  Check for errors here?
    LL_Push(procs, (void *) pjt);
    //  free(pjt) happens when  linked list is freed 
  }
  closedir(proc);
  return(1);
}

void store_tops(Driver *drvthis){
    PrivateData * p = drvthis->private_data;
  static LinkedList *jiffies[3]={NULL,NULL,NULL};
  static int newj=0;
  static unsigned long int named_jiffs[2][MAXTOP];
  int named_jiffs_count[MAXTOP];
  int i;
  static struct timeval delta_t[2];
  // compare values with machine_get_load and top
  // procjiffy_type load_sum;

  if (jiffies[1-newj]!=NULL){
    struct timeval now;
    gettimeofday(&now, NULL);
    if (time_diff(drvthis, p->tdid_top, now)<5000000) return;
    p->tdid_top=now;
  }
    
  //load_sum.utime=0;
  //load_sum.stime=0;
  //load_sum.cutime=0;
  //load_sum.cstime=0;
  jiffies[newj] = LL_new();
  if (jiffies[newj] == NULL) {
    report(RPT_ERR, "%s jiffies[%d] : Error allocating list.",
        __FUNCTION__, newj);
    return;
  } 
  else {
    gettimeofday(&delta_t[newj], NULL);
    machine_get_jiffies(drvthis, jiffies[newj]);
    // sort jiffies[newj] by pid: only sort performed on any set
    LL_Rewind(jiffies[newj]);
    LL_Sort(jiffies[newj], sort_jiffpid); 
    LL_Rewind(jiffies[newj]);
    // create jiffies[2] for elapsed values
    jiffies[2] = LL_new();
    if (jiffies[2] == NULL) {
      report(RPT_ERR, "%s jiffies[2] : Error allocating list.", __FUNCTION__);
      return;
    } 
    for (i=0; i < p->topname; i++){
      named_jiffs_count[i]=0;
    }
    if (jiffies[1-newj]==NULL) { // true if first iteration 
      double uptime, idle;
      machine_get_uptime(&uptime, &idle);
      p->elapse=uptime*100; // 100 jiffies / sec.
/* L. T. says in the headers that jiffies wrap artificially at 5 minutes to
 * aid in debugging that wrap.  So if all jiffies wrapped at 5 min uptime,
 * should if ( uptime > 5 min in jiffies ) then uptime -= 5 min in jiffies
 * be used here to take elapse since wrap ?
 */
      for (i=0; i < p->topname; i++){
        named_jiffs[0][i]=named_jiffs[1][i]=0;
      }
    }
    else {  // true all up and running iterations
      LL_Rewind(jiffies[1-newj]);
      p->elapse=time_diff(drvthis,delta_t[1-newj], delta_t[newj])/10000;
    }
    report(RPT_CLIUP, "%s elapse %f seconds.",
        __FUNCTION__, (float)p->elapse / 100);
    do {
      procjiffy_type *computed;
      computed = malloc(sizeof(procjiffy_type));
      if (!computed) {
        report(RPT_ERR, "%s jiffies[2] : Error allocating new jiffies[2] member.", __FUNCTION__);
        return;
      }
      procjiffy_type *fresh = LL_Get(jiffies[newj]);
      computed->pid=fresh->pid;
      memcpy(computed->name, fresh->name, strlen (fresh->name)+1 );
      if (p->topname >0  ) { // if doing top by name
        for (i=0; i < p->topname; i++){ // for each name tracked 
          if (!strncmp(p->topnames[i], fresh->name, strlen(p->topnames[i]))){
            // how to handle jiffies wrap here ?
            named_jiffs[newj][i]+=fresh->utime+fresh->stime; //+fresh->cutime+fresh->cstime;
              // is sum of all procs by that name
              // can wrap -> time data type or no ?
            (named_jiffs_count[i])++;
          }
        }
      }        
      if (jiffies[1-newj]==NULL) { // if 1st read of jiffies
        computed->utime=fresh->utime;
        computed->stime=fresh->stime;
        //computed->cutime=fresh->cutime;
        //computed->cstime=fresh->cstime;
        //report(RPT_ERR, "%s 1stread %d %s %lu %lu",
        //    __FUNCTION__, fresh->pid, fresh->name, fresh->utime, fresh->stime);
      }
      else {  // seach old jiffies for previous value
        procjiffy_type *test;
        int found=0;
        do { 
          test = LL_Get(jiffies[1-newj]);
          if (test->pid == computed->pid){
              // when does jiffies wrap ?
              // /usr/include/linux/time.h:172: * The maximum jiffie value is (MAX_INT >> 1).
              // limits. h:#  define INT_MAX       2147483647
              // 1 sec  =       100 jiff
              // 1 min  =     6 000 jiff
              // 1 hour =   360 000 jiff
              // 1 day  = 8 640 000 jiff
              // 2147483647 - 248 * 8 640 000 = 2147483647 - 2142720000 = 4766647 remainder 
              //    4766647 -  13 *   360 000 =    4766647 -    4680000 =   86647 remainder
              //      86647 -  14 *     6 000 =      86647 -      84000 =    2647 remainder
              //       2647 =  26.47 sec
              // 248 days, 13 hours, 14 min, 26.47 sec 
            computed->utime= fresh->utime-test->utime;
            computed->stime= fresh->stime-test->stime;
            //computed->cutime= fresh->cutime-test->cutime;
            //computed->cstime= fresh->cstime-test->cstime;
            // ->?time are unsigned long int, so if they wrapped in above, throw out 
            // wrapped unsigned long int results & start over
            if (fresh->utime < test->utime)  {
              computed->utime= fresh->utime - test->utime + (INT_MAX >> 1);
            }
            if (fresh->stime < test->stime) {
              computed->stime= fresh->stime - test->stime + (INT_MAX >> 1);
            }
          /*  if (fresh->cutime < test->cutime)  {
              computed->cutime= fresh->cutime - test->cutime + (INT_MAX >> 1);
            }
            if (fresh->cstime < test->cstime) {
              computed->cstime= fresh->cstime - test->cstime + (INT_MAX >> 1);
            }
          */  found=1;
            break;
          }
        } while (LL_Next (jiffies[1-newj]) !=-1);
        if (found==0) {  // did not find a match in previous set
            computed->utime =fresh->utime;
            computed->stime =fresh->stime;
            //computed->cutime =fresh->cutime;
            //computed->cstime =fresh->cstime;
            // search jiffies[1-newj] from beginning next time
             LL_Rewind(jiffies[1-newj]);
        }

      }  // end search old jiffies
      // could sum this jiffies[newj] into load totalizer here
      // see how it compares to values form machine get load
      //load_sum.utime=load_sum.utime+computed->utime;
      //load_sum.stime=load_sum.stime+computed->stime;
      //load_sum.cutime=load_sum.cutime+computed->cutime;
      //load_sum.cstime=load_sum.cstime+computed->cstime;
        
      LL_Push (jiffies[2], (void *)computed);
    } while (LL_Next (jiffies[newj]) != -1 );// end feshly read list into jiffies[2]

/*    float myload =  (float)(load_sum.utime+ load_sum.stime // +load_sum.cutime+load_sum.cstime;
        ) / (float)p->elapse;
    int load_diff = myload*10000 - p->load_t*100;
    if ((p->load_t > 99) || (abs(load_diff)  > 100 )){ 
      report(RPT_ERR, "%s eps %7.2f, tot j %8lu = l %3.2f (%3.2f): %3.2f diff.",
        __FUNCTION__, (float)p->elapse, load_sum.utime+ load_sum.stime, // +load_sum.cutime+load_sum.cstime,
        myload*100, p->load_t, (float)abs(load_diff)/100);
    }
*/                          
    //  store  jiffies[2]
    LL_Rewind(jiffies[2]);
    LL_Sort(jiffies[2], sort_jiffies);  // sort by cpu load
    LL_Rewind(jiffies[2]);
    int j;
    int skipnamed=0;
    int pass_break_to_i;
    for (i = 0; i < p->topnum + skipnamed; i++) {
      pass_break_to_i=0;
      procjiffy_type *pj = LL_Get(jiffies[2]);
      if (p->topname!=0) { // if doing top by name
        for (j=0; j < p->topname; j++){ // for each name tracked 
          if (!strncmp(p->topnames[j], pj->name, strlen(p->topnames[j]))){
            report(RPT_CLIUP, "%s topn% 2d % 16s,  % 8lu % 8lu = % 2.6f  count %d\n",
              __FUNCTION__, j, p->topnames[j], pj->utime, pj->stime, //, pj->cutime, pj->cstime, 
              ((float)named_jiffs[newj][j]-(float)named_jiffs[1-newj][j])
              /((float)p->elapse/100), named_jiffs_count[j]);
            skipnamed++;
            pass_break_to_i=1;
            break; 
          }
        }
      }
      if (pass_break_to_i==0){
        report(RPT_CLIUP, "%s Top % 2d % 16s,  % 8lu % 8lu = % 2.6f", __FUNCTION__, 
            i, pj->name, pj->utime, pj->stime, // pj->cutime, pj->cstime,
          (  // (float)pj->cutime+(float)pj->cstime+
             (float)pj->utime+(float)pj->stime)
          /((float)p->elapse/100));
        p->topnumval[i-skipnamed]=( //(float)pj->cutime+(float)pj->cstime+
            (float)pj->utime+(float)pj->stime)
            /((float)p->elapse/100);
        if (p->topnumnames[i-skipnamed]) {
          free (p->topnumnames[i-skipnamed]);
        }
        p->topnumnames[i-skipnamed]=strdup(pj->name);
      }
      LL_Next (jiffies[2]);
    }
    if (p->topname!=0) { // if doing top by name
      for (i=0; i < p->topname; i++){ // for each name tracked 
        p->topnameval[i]=((float)named_jiffs[newj][i]- (float)named_jiffs[1-newj][i]) /
                                                          ((float)p->elapse/ 100);
        named_jiffs[1-newj][i]=0;
        named_jiffs_count[i]=0;
      }
    }
    newj=1-newj; // new becomes old
    // Now clean up the old one & jiffies[2] ...
    if (jiffies[newj]!=NULL){ // 1st time through is NULL
      int lllength=LL_Length (jiffies[newj]);
      LL_Rewind(jiffies[newj]);
      do {
        procjiffy_type *tofree = (procjiffy_type *) LL_Get(jiffies[newj]);
        if (tofree != NULL) { free(tofree); }
      lllength=LL_Length (jiffies[newj]);
      } while (LL_Next(jiffies[newj]) == 0);
      LL_Destroy(jiffies[newj]);
      jiffies[newj]=NULL;
    }
    LL_Rewind(jiffies[2]);
    do {
      procjiffy_type *tofree=(procjiffy_type *) LL_Get(jiffies[2]);
      if (tofree != NULL) { free(tofree); }
    } while (LL_Next(jiffies[2]) == 0);
    LL_Destroy(jiffies[2]);
    jiffies[2]=NULL;
  } /* end of top  */
}

static int
sort_jiffies (void *a, void *b)
{
        procjiffy_type *one, *two;
        if ((a == NULL) || (b == NULL))
                return 0;
        one = (procjiffy_type *) a;
        two = (procjiffy_type *) b;
  //float onej=(float)one->utime + (float)one->stime + (float)one->cutime + (float)one->cstime;
  //float twoj=(float)two->utime + (float)two->stime + (float)two->cutime + (float)two->cstime;

  // by pid or by pid and children ?
        return (two->utime + two->stime  > one->utime + one->stime );
        //return (twoj  > onej );
}
#endif // DO_ULT

#ifdef DO_LM
/*
 * read a lm sensor p->lmname[ p->lmnum] lmfeat[p->lmnum]
 * if so fill in p->chip[p->lmnum] and p->feat[p->lmnum]
 * and p->lmnum ++
 * return 1 on sucess, 0 otherwise
 * */
int find_build_req_lm(Driver *drvthis){
    PrivateData * p = drvthis->private_data;
  const struct sensors_chip_name *achip=NULL;
  int chip_num=0;
  do {
    achip=sensors_get_detected_chips(&chip_num);
    // warning: assignment discards qualifiers from pointer target type
    if (achip==NULL) continue;
    if (0 != strncmp(p->lmnames[p->lmnum], achip->prefix, strlen(p->lmnames[p->lmnum]))){ 
      continue;
    }
    report(RPT_LMS < RPT_CLICRE ? RPT_LMS : RPT_CLICRE,
        "%s lmName%d=%s matched %s", __FUNCTION__, p->lmnum+1,
        p->lmnames[p->lmnum], achip->prefix);
    const struct sensors_feature_data *afeature=NULL;
    double value;
    int a,b;
    a=b=0;
    while ((afeature=sensors_get_all_features(*achip, &a, &b))) {
    // warning: assignment discards qualifiers from pointer target type
      if (afeature->mode & SENSORS_MODE_R) {
        if (sensors_get_feature(*achip, afeature->number, &value)) {
          // printf ("Can't get feature `%s' data!\n", afeature->name);
          continue;
        }
        if (afeature->mapping == SENSORS_NO_MAPPING){
          char *label=NULL;
          sensors_get_label(*achip, afeature->number, &label);
          if (!strncmp(label,p->lmfeatu[p->lmnum], strlen(p->lmfeatu[p->lmnum]) ) ){
              report(RPT_LMS < RPT_CLICRE ? RPT_LMS : RPT_CLICRE,
                  "%s lmFeature%d=%s matched %s. (%f)", __FUNCTION__, 
                  p->lmnum+1, p->lmfeatu[p->lmnum], label, value);
            /* create, copy into p-> lmstructs */
            
            p->chipn[p->lmnum]=malloc(sizeof(struct sensors_chip_name));
            if (p->chipn[p->lmnum]==NULL) {
              report(RPT_ERR, "%s p->chipn[p->lmnum]=malloc fail", __FUNCTION__);
              return 0;
            }
            p->chipn[p->lmnum]->prefix=strdup(achip->prefix);
            p->chipn[p->lmnum]->bus=achip->bus;
            p->chipn[p->lmnum]->addr=achip->addr;
            p->chipn[p->lmnum]->busname=strdup(achip->busname);
            
            p->featn[p->lmnum]=malloc(sizeof(struct sensors_feature_data));
            if (p->featn[p->lmnum]==NULL) {
              report(RPT_ERR, "%s p->featn[p->lmnum]=malloc fail", __FUNCTION__);
              return 0;
            }
            p->featn[p->lmnum]->number=afeature->number;
            p->featn[p->lmnum]->name=strdup(afeature->name);
            p->featn[p->lmnum]->mapping=afeature->mapping;
            p->featn[p->lmnum]->unused=afeature->unused;
            p->featn[p->lmnum]->mode=afeature->mode;

#ifdef CLIENT_CHILD
            if (p->lmlcd[p->lmnum]!=0){
              p->lmline[p->lmnum]=malloc(p->width+1);
              if (!p->lmline[p->lmnum]) {
                report(RPT_ERR, "%s log lm LCD line malloc fail.", __FUNCTION__);
                p->lmline[p->lmnum]=NULL;
                p->lmlcd[p->lmnum]=0;
              } // yeah, let it try to continue ...
            }
#endif
            report(RPT_LMS < RPT_CLICRE ? RPT_LMS : RPT_CLICRE,
              "%s chip %d, prefix %s, bus %d, addr %d, busname %s feature %d.", __FUNCTION__, 
              chip_num, achip->prefix, achip->bus, achip->addr, achip->busname, afeature->number);
            p->lmnum++;
            return 1;
          }
        } // end SENSORS_NO_MAPPING
      } // end (afeature->mode & SENSORS_MODE_R)
    } // end sensors_get_all_features for a given chip
  } while (achip !=NULL);
  report(RPT_ERR, "%s lmName%d=%s lmFeature%d=%s not found.",
    __FUNCTION__, p->lmnum+1, p->lmnames[p->lmnum],
    p->lmnum+1, p->lmfeatu[p->lmnum]);
  free(p->lmnames[p->lmnum]);
  p->lmnames[p->lmnum]=strdup("not_found");
  free(p->lmfeatu[p->lmnum]);
  p->lmfeatu[p->lmnum]=strdup("not_found");
  return 1;
  // free() achip, afeature, alabel ?
}

void store_lm(Driver *drvthis){
    PrivateData * p = drvthis->private_data;

  struct timeval now;
  gettimeofday(&now, NULL);
  if (time_diff(drvthis, p->tdid_lm, now)<125000) return;
  p->tdid_lm=now;
    
  int counter;
  for (counter=0; counter < p->lmnum ; counter ++){ // lmsensors data
    double value;
    if (sensors_get_feature(*p->chipn[counter], p->featn[counter]->number, &value)) {
      report(RPT_ERR, "%s failed to read sensor %s:%s.", __FUNCTION__, 
          p->chipn[counter]->prefix, p->featn[counter]->name);
      p->lmval[counter]=666;
    }
    else {
      report(RPT_LMS < RPT_CLIUP ? RPT_LMS : RPT_CLIUP,
        "%s lms%d prefix %s, bus %d, addr %d, busname %s feature %d = %f.",
        __FUNCTION__, counter+1, p->chipn[counter]->prefix, p->chipn[counter]->bus, 
        p->chipn[counter]->addr, p->chipn[counter]->busname, p->featn[counter]->number, value );
      p->lmval[counter]=(float)value;
    }
  }
}
#endif // DO_LM

void
client_loop(Driver * drvthis)
{
  PrivateData *p = drvthis->private_data;

  int Quit=0;
  struct timeval now;
  gettimeofday(&now,NULL);
#ifdef CLIENT_CHILD
  int i = 0;
  int connected = 0;
  char buf[8192];
  char *argv[256];
  int argc, newtoken;
  int len;
  // lines and lcd_hgt are globals 
  // assigned values before main _ loop is called
  char foreground[lines/lcd_hgt];  // is a screen visible
  unsigned int updates[lines/lcd_hgt];  // # updates to current screen 
  char active_screen=0;  // which screen # is active
  char listen_ign=0;
  unsigned int want_success=0;

  for(i=0; i<lines/lcd_hgt; i++) 
    updates[i]=foreground[i]=0;
#endif
  // Main loop
  while(!Quit)
  {
    gettimeofday(&now, NULL);
    if ( (time_diff(drvthis, p->tdid_tlog, now)>=
        p->log_sens * 1000000) &&
        (p->bad_time_to_log==0) ){
      textlog (drvthis);
      gettimeofday(&p->tdid_tlog, NULL);
    } //end of if time to display 
    usleep(125000); // 1/8 sec screen update loop
#ifdef CLIENT_CHILD
    // Check for server input...
    len = sock_recv( cf_dsk, buf, 8000);

    // Handle server input...
    while(len > 0)
    {
      report(RPT_CLIUP, "%s tokenizer got: %s", __FUNCTION__, buf);
      // Now split the string into tokens...
      argc = 0;
      newtoken = 1;
      for(i = 0; i < len; i++)
      {
        switch(buf[i])
        {
          case ' ':
            newtoken = 1;
            buf[i] = 0;
          break;
          default:  // regular chars, keep tokenizing
            if(newtoken)
              argv[argc++] = buf + i;
            newtoken = 0;
          break;
          case '\0':
          case '\n':
            buf[i] = 0;
            if(argc > 0)
            { 
             report(RPT_CLIUP, "%s Tokenizer line: %s %s\n", __FUNCTION__, argv[0], argv[1]);
              if(0 == strcmp(argv[0], "listen"))
                {
                active_screen=atoi((char *)argv[1]+6);
                report(RPT_CLIUP, "%s Listen %d", __FUNCTION__, active_screen );
                foreground[(int)active_screen]=1;
                listen_ign=1;
              }
              else if(0 == strcmp(argv[0], "ignore"))
              {
                int ign_scn=atoi((char *)argv[1]+6);
                report(RPT_CLIUP,"%s Ignore %d", __FUNCTION__, ign_scn);
                foreground[ign_scn]=0;
                updates[ign_scn]=0;
                listen_ign=0;
              }
              else if(0 == strncmp(argv[0], "cf_log", 6))
              {    // if "l/i screen l/i" is chopped up by a success ... ?
                // seems to never happen now
                int scrn_num=atoi((char *)argv[1]+6);
                report(RPT_CLIUP,"%s foreground cf_log[%d]=%d",
                  __FUNCTION__, scrn_num, listen_ign);
                foreground[scrn_num]=listen_ign;
              }
              else if (0 == strcmp(argv[0], "key"))
              {
                report(RPT_CLIUP,"%s Key %s", __FUNCTION__,  argv[1]);
              }
              else if(0 == strcmp(argv[0], "menu"))
              {
              }
              else if(0 == strcmp(argv[0], "connect"))
              {
                int a;
                for(a = 1; a < argc; a++)
                {
                  if(0 == strcmp(argv[a], "wid"))
                    lcd_wid = atoi(argv[++a]);
                  else if(0 == strcmp(argv[a], "hgt"))
                    lcd_hgt = atoi(argv[++a]);
                  else if(0 == strcmp(argv[a], "cellwid"))
                    lcd_cellwid = atoi(argv[++a]);
                  else if(0 == strcmp(argv[a], "cellhgt"))
                    lcd_cellhgt = atoi(argv[++a]);
                }
                connected = 1;
              }
              else if(0 == strcmp(argv[0], "bye"))
              {
                report(RPT_CLIUP, "%s got : %s : exit", __FUNCTION__, argv[0]);
                exit(EXIT_SUCCESS);
              }
              else if(0 == strcmp(argv[0], "success"))
              {
                want_success--;
                report(RPT_CLIUP, "%s success: got %s [%u]",
                    __FUNCTION__, argv[0], want_success);
              }
              else
              {
                int j;
                for(j = 0; j < argc; j++)
                  report(RPT_CLIUP, "%s got ?: %s", __FUNCTION__, argv[j]);
              }
            }

            // Restart tokenizing
            argc = 0;
            newtoken = 1;
          break;
        } // switch( buf[i] )
      } // end tokenizig one buf
      len = sock_recv( cf_dsk, buf, 8000);
    } // end while len > 0
    // Update screens...
    for (i=active_screen*lcd_hgt; i< active_screen*lcd_hgt+lcd_hgt; i++)  {
      if ( i == lines ) break;
      int screen=i/lcd_hgt; // which screen # this is,  == active_screen
      int screen_line=(i+lcd_hgt)%lcd_hgt+1; // which line # on this screen
      if (foreground[screen]==1) {  // if this screen is visible now
        if (screen_line==1) updates[screen]++;
        if ( ((updates[screen]+7)%8)==0) { // once every 8 loops
          // ( 1 ) = 8 times a second
          // ((updates[screen]+3)%4)==0 = 2 times / second
          // ((updates[screen]+1)%2)==0 = 4 times / second
          //updateline[displinenum[i]](drvthis, t_or_f_num[i]);
          updateline[i](drvthis, t_or_f_num[i]);
/*#ifdef PROTOCOL_CHEATING
          report(RPT_CLIUP, "%s displayed PCHEAT %d = orig %d = %s, t/f = %d",
            __FUNCTION__, i, displinenum[i], 
            (char *)(*((int*)iowritestohere[i]))
            , t_or_f_num[i]);
#endif*/
#ifndef PROTOCOL_CHEATING    
          report(RPT_CLIUP, "%s displayed %d = orig %d = %s, t/f = %d",
            __FUNCTION__, i, displinenum[i], orderline[i] , t_or_f_num[i]);
          snprintf(buf,sizeof(buf),"widget_set cf_log%d line%d 1 %d \"%s\"\n",
            screen,   // screen # (name)
            i,    // total line # (name)
            screen_line,   // line # this screen (value)
            orderline[i]);   //  text
          report(RPT_CLIUP , "%s line %d (orig %d) rep %u %u update widget: %s",
              __FUNCTION__, i, displinenum[i],
              updates[screen], updates[screen]/8, buf );
          if (sock_send_string( cf_dsk, buf )< strlen(buf) ){
            report(RPT_CRIT, "%s exit on socket write error.", __FUNCTION__);
            exit(EXIT_FAILURE);
          }
          want_success++;
#endif // #ifndef PROTOCOL_CHEATING
        }
      } // end line [i] is foreground
    } // end for each line [i]
#endif // CLIENT_CHILD
  } // end while(!Quit) 125000 usec loop
} // end client_loop

#ifdef DO_HDD
static void hddtemp_names(Driver *drvthis){
    PrivateData * p = drvthis->private_data;
  
  int hdd_sock, count=0;
  do {
    hdd_sock = sock_connect ("localhost", 7634);
    if (hdd_sock <=0 ) {
      report(RPT_CLICRE, "%s waiting for hddtemp daemon ... %d",
          __FUNCTION__, count );
      usleep (500000); // hddtemp connect
      count++;
    }
  } while ((hdd_sock <=0 ) && (count < 200));
  
  if (hdd_sock <=0 ) {
    report(RPT_ERR, "%s failed to open hddtemp on port %d", __FUNCTION__, 7634);
    p->hdd_count=0;
    return;
  }
  else {
    report(RPT_CLICRE, "%s hddtemp daemon good connect on try %d",
          __FUNCTION__, count );
    int len;
    count=0;
    // from daemon_send_message in hddtemp/srs/deamon.c
    // appears 130 bytes hypothetical max per drive
    // I'm getting 26
    char hdd_in[132*MAX_HDD];
    do {
      usleep (50000); // hddtemp wakeup
      len = sock_recv( hdd_sock, hdd_in, sizeof (hdd_in));
      count++;
    } while ((len <= 0) && (count < 200));
    if (len <=0 ) {
      report(RPT_ERR, "%s failed to read from hddtemp on port %d", __FUNCTION__, 7634);
      p->hdd_count=0;
      return;
    } 
    else {
      // "good read on try 1, |/dev/hda|ST3120024A|27|C|2c/devices/0-0290/temp1_input"
      if ( len < 512) hdd_in[len]='\0';
      else hdd_in[511]='\0';
      report(RPT_CLICRE, "%s hddtemp daemon good read on try %d, %s",
          __FUNCTION__, count -1, hdd_in);
    }
    if (len >0 ) p->hdd_rep_len=len+1;
    char * ptr=hdd_in;
    int word_of = 0;
    int drive=0;
    while (strstr (ptr, "|" )!=NULL){
      p->hddname[drive]=NULL; 
#ifdef CLIENT_CHILD
      p->hddline[drive]=NULL; 
      p->hddlcd[drive]=0;
#endif
      ptr=strstr (ptr, "|" );
      do {
        ptr=index(ptr, '|')+1;
/*        if ( (ptr ==NULL) || (ptr > hdd_in + sizeof (hdd_in) ) ){
          report(RPT_ERR, "%s fail on drive # %d", __FUNCTION__, drive );
          ptr=NULL;
          if (p->hddname[drive]!=NULL) free (p->hddname[drive]);
#ifdef CLIENT_CHILD
          if (p->hddline[drive]!=NULL) free (p->hddline[drive]);
          p->hddlcd[drive]=0;
#endif
          continue;
        

        }
*/
        if (word_of==0) { // if this is the name
          char aword[128];
          strncpy(aword, ptr , index(ptr, '|') -ptr);
          aword[index(ptr, '|') -ptr ]='\0';
          p->hddname[drive]=strdup(aword);
        }
        if (word_of==2) { // if this is the temperature 
          char aword[128];
          strncpy(aword, ptr , index(ptr, '|') -ptr);
          aword[index(ptr, '|') -ptr ]='\0';
          p->hddtemp[drive]=strdup(aword);
        }
        word_of ++;
      } while (( word_of != 4 ) && (ptr  !=NULL));
      if (ptr !=NULL) {
#ifdef CLIENT_CHILD
        char target[256];
        sprintf(target, "%s_Lcd", p->hddname[drive]);
        p->hddlcd[drive]=drvthis->config_get_int(drvthis->name, target, 0, 0);
        p->hddline[drive]=malloc(p->width+1);
          if (!p->hddline[drive]) {
            report(RPT_ERR, "%s LCD line malloc fail.", __FUNCTION__);
            p->hddline[drive]=NULL;
            p->hddlcd[drive]=0;
          }
        report(RPT_CLICRE, "%s drive %d name %s temp %s Lcd=%d", 
          __FUNCTION__, drive, p->hddname[drive], p->hddtemp[drive], 
          p->hddlcd[drive] );
#else
        report(RPT_CLICRE, "%s drive %d name %s temp %s", 
          __FUNCTION__, drive, p->hddname[drive], p->hddtemp[drive] );
#endif
        ptr=index(ptr, '|') + 1;
        drive ++;
        word_of=0;
        // end read if storage is full 
        if (drive == MAX_HDD) ptr=NULL;
      } // end bail on partial read or too many drives
    } // end good drive read returned better than ptr==NULL
    p->hdd_count=drive;
  sock_close ( hdd_sock );
  } // end good connect
  report(RPT_CLICRE, "%s read %d drives", __FUNCTION__, p->hdd_count);
} // end hddtemp_ names

static void hddtemp_temps(Driver *drvthis){
    PrivateData * p = drvthis->private_data;
  int hdd_sock, count=0;
  do {
    hdd_sock = sock_connect ("localhost", 7634);
    if (hdd_sock <=0 ) {
      report(RPT_ERR, "%s waiting for hddtemp daemon ... %d",
          __FUNCTION__, count );
      usleep (500000); // hddtemp connect
      count++;
    }
  } while ((hdd_sock <=0 ) && (count < 200));
  
  if (hdd_sock <=0 ) {
    report(RPT_ERR, "%s failed to open hddtemp on port %d", __FUNCTION__, 7634);
    p->hdd_count=0;
    return;
  }
  else {
    report(RPT_CLIUP, "%s hddtemp daemon good connect on try %d",
          __FUNCTION__, count );
    int len;
    count=0;
    char hdd_in[p->hdd_rep_len];
    do {
      usleep (50000); // hddtemp wakeup
      len = sock_recv( hdd_sock, hdd_in, sizeof (hdd_in));
      count++;
    } while (len <= 0 && count < 30);
    if (len <=0 ) {
      report(RPT_ERR, "%s failed to read from hddtemp on port %d", __FUNCTION__, 7634);
      p->hdd_count=0;
      return;
    } 
    else {
      if ( len < p->hdd_rep_len) hdd_in[len]='\0';
      else hdd_in[p->hdd_rep_len-1]='\0';
      report(RPT_CLIUP, "%s hddtemp daemon good read on try %d, %s",
          __FUNCTION__, count -1, hdd_in);
    }
    char * ptr=hdd_in;
    int word_of = 0;
    int drive=0;
    // why was this crashing when (ptr == NULL) ?
    while ((ptr != NULL) && (strstr (ptr, "|" )!=NULL)){
      ptr=strstr (ptr, "|" );
      do {
        ptr=index(ptr, '|')+1;
        if (word_of==2) { // if this is the temperature 
          char aword[128];
          strncpy(aword, ptr , index(ptr, '|') -ptr);
          aword[index(ptr, '|') -ptr ]   ='\0';
          if (p->hddtemp[drive]!=NULL) free(p->hddtemp[drive]);
          p->hddtemp[drive]=strdup(aword);
          report(RPT_CLIUP, "%s drive %d temp %s", 
            __FUNCTION__, drive, p->hddtemp[drive]);
        }  
        word_of ++;
      } while (( word_of != 4 ) && (ptr  !=NULL));
      // end read if storage is full 
      if (drive+1 == p->hdd_count) ptr=NULL;
      if (ptr != NULL) ptr=index(ptr, '|') + 1;
      drive ++;
      word_of=0;
    } // end good drive read
  sock_close ( hdd_sock );
  } // end good connect
} // end hddtemp_ temps 
#endif // DO_HDD

#ifdef MY_CUSTOM_LINE
static void update_wattsline (Driver *drvthis, unsigned char unused) {
    PrivateData * p = drvthis->private_data;
  static int check_once=0;
  if (check_once==0){
    check_once=1;
    if ( p->tempcount != 6) {
      // simple & easy & not gauranteed to catch a change check 
      report(RPT_ERR, "%s #ifdef MY_CUSTOM_LINE; edit update_wattsline() to something meaningful.",
          __FUNCTION__);
      snprintf(watts, 12, "?: undefined");
      return;
    }
  }
  if ( p->tempcount != 6) {
    sprintf(watts, "?: ");
  }
  if ( p->tempcount < 3) {
    snprintf(watts, 12, "?: <3 temps");
    return;
  }
  // average of 3 temp probes measuring water out of computer into radiator
  float in =(  (float)p->ds18[0].ds18+
      (float)p->ds18[1].ds18+
      (float)p->ds18[2].ds18) / 3;  
  // average of 3 temp probes measuring water out of radiator into computer
  float out=(  (float)p->ds18[p->tempcount-1].ds18+
      (float)p->ds18[p->tempcount-2].ds18+
      (float)p->ds18[p->tempcount-3].ds18) / 3;
  float dt=(out-in)/16; // degrees celcius delta temp
  float mlps=p->fans[1].rpm_rpt; // milliliters of water per second
  float w = dt * mlps *4.2; // 1 calorie = raise 1 ml water 1 deg c
        // 4.2 joules / calorie ; joules / second = wattage
  /* cue the laser lights and chorus;
   *   tapdance with the theory of significant digits
   * and manufacturers' specs for accuracy & repeatability;
   * pay the laboratory staff for measuring the thermal conductivity
   * and thermal capacitance of today's coolant mixture;
   * blow big kisses to degree of certainty & margin of error; */
  int w_lh=(int)w;
  int w_rh=(int) ((w-(float)w_lh)*100);
  sprintf(watts+2, "% 3d.%02d        ", w_lh, w_rh);
#ifdef DO_ULT
  float l=p->load_t;
  int l_lh=(int)l;
  int l_rh=(int) ((l-(float)l_lh)*100);
  snprintf(watts+p->width-7, 8, "% 3d.%02d%%",  l_lh, l_rh );
#endif // DO_ULT
//#ifdef PROTOCOL_CHEATING    
//  snprintf((char *)(wattsptr), strlen(watts), "%s", watts);
//#endif
  watts_est =w;
} // end update_wattsline
#endif // MY_CUSTOM_LINE

/*
 * *****************************************************
 * everything below here is in CLIENT_CHILD ifdef
 * *****************************************************
 */
#ifdef CLIENT_CHILD

/*
 * wait on socket for "success" from server
 * other = string that also signals sucess, 
 * when server switches to a new screen, the "listen newscreen" 
 * message knocks out / supresses / replaces an expected success
 */
int wait_success(Driver * drvthis, int options, char **other)
{
  char buf[8192];
  int len=0;
  int a;
  char * argv[20];
  int argc;
  short received = 0;
  short timeout = 5000; /* Give the server 5 secs to respond */
  int timeout_mult = 10000;
  char success=0;
  short repeat=0;
  
  while (repeat < 10000 ) {
    buf[0] = 0;
    while( !received && timeout > 0) {
      len = sock_recv_string ( cf_dsk, buf, sizeof(buf));
      if( len == 0 ) {
        usleep(timeout_mult); // waiting for success from server socket
        timeout --;
      } else {
        received = 1;
      }
    }
    if( len <= 0 ) {
      report(RPT_ERR, "%s Received (len <= 0) repeat [%d].", __FUNCTION__, repeat );
    }
    argc = split( buf, ' ', argv, 20 );
    for (a = 0; a < argc; a++) {
      if (0 == strcmp (argv[a], "success")){
        success=1;
        report(RPT_DEBUG, "%s received  expected: %s repeat %d & return",
            __FUNCTION__, argv[a], repeat);
        return 0;
      }
      else if (other!=NULL) {
        int i;
        for (i=0;  i <= options ; i++){
          // report(RPT_CLICRE, "%s alternate check : %s", __FUNCTION__, other[i]);
          if (0 == strcmp (argv[a], other[i])){
            success=1;
            report(RPT_DEBUG, "%s received alternate: %s  repeat %d & return",
                __FUNCTION__, argv[a], repeat);
            return 0;
          } // end checking alternates
        }
      }
      else if (strlen(argv[a]) !=0) {
        report(RPT_DEBUG, "%s received something else ?: %s repeat %d",
            __FUNCTION__, argv[a], repeat);
      }
    }
    if(success==0){
      repeat ++;
      if (repeat %2000==0) {
        report(RPT_WARNING, "%s did not receive a success, repeat [%d].", __FUNCTION__, repeat);
        }
    }
    else if(success==1){
      return 1;
    }
    usleep(1000); // wait_success, cli waiting on socket response
  } // end of repeat for multiple responses
  report(RPT_ERR, "%s did not receive a success, repeat [%d] and fail.", __FUNCTION__, repeat );
  return 0;
} // end of wait_success 

int cli_scrn(Driver * drvthis){
  PrivateData * p = drvthis->private_data;
  int fan, temp, i, j;
  char *spaces=NULL;
        // add blank line(s) as option to organize lines ?

  spaces=malloc(lcd_wid+1);
  if (spaces == NULL) {
    report(RPT_ERR, "%s client malloc spaces=fail.", __FUNCTION__);
    exit(EXIT_FAILURE); 
  }
  for (i=0; i<lcd_wid; i++)
    snprintf (spaces+i, 2," ");
  
  for (i=0; i<MAXLINES; i++){
     order[MAXLINES]=-1;
     orderline[MAXLINES]=NULL;     
     displinenum[MAXLINES]=-1;
#ifdef PROTOCOL_CHEATING
     iowritestohere[MAXLINES]=0;
     // =NULL was working once what has changed ?
#endif // PROTOCOL_CHEATING
  }
  
  for (fan=0; fan <4; fan++) {
     if (p->fans[fan].dowidget!=0) {
      p->fans[fan].line=malloc(lcd_wid+1);
      if (p->fans[fan].line==NULL) {
        report(RPT_ERR, "%s malloc p->fans[].line=fail.", __FUNCTION__);
        exit(EXIT_FAILURE); 
      }
      snprintf(p->fans[fan].line, lcd_wid+1, "%s%s",
          p->fans[fan].name, spaces);
      snprintf((p->fans[fan].line)+lcd_wid-7, 8, "not-yet");
      order[lines]=(p->fans[fan].dowidget+1)*10;
      displinenum[lines]=lines;
      orderline[lines]=p->fans[fan].line;
      t_or_f_num[lines]=fan;
      updateline[lines]=*set_up_fanline;
      report(RPT_DEBUG, "%s %02d lines: fans[%d].line , order %03d, %s",
          __FUNCTION__, lines, fan, order[lines], p->fans[fan].line);
#ifdef PROTOCOL_CHEATING
      iowritestohere[lines]=(int)&p->fans[fan].line;
      report(RPT_PCHEAT, "%s    fan[%02d].line iowritestohere[%02d]=%d, %s",
          __FUNCTION__, fan, lines, iowritestohere[lines],
          (char *)(*((int*)iowritestohere[lines])));
#endif
      lines++;
      if (p->fans[fan].sensor==0) {
        p->fans[fan].line2=NULL;
      }
      else { // temp associated
        p->fans[fan].line2=malloc(lcd_wid+1);
        if (p->fans[fan].line2==NULL) {
          report(RPT_ERR, "%s  malloc p->fans[].line2=fail.", __FUNCTION__);
          exit(EXIT_FAILURE); 
        }
        order[lines]=(p->fans[fan].dowidget+1)*10+1;
        displinenum[lines]=lines;
        orderline[lines]=p->fans[fan].line2;
        // without separate set_up_fan.line2 func
        t_or_f_num[lines]=fan;
        updateline[lines]=*set_up_fanline;
        if  (p->fans[fan].sen_ref==0) { // simple temp
          snprintf(p->fans[fan].line2, lcd_wid+1, "%s%s",
              p->ds18[p->fans[fan].sensor-1].name, spaces);
          snprintf((p->fans[fan].line2)+lcd_wid-7, 8, "not-yet");
        }
        else { // must be a relative temp
          snprintf(p->fans[fan].line2, lcd_wid+1, "%s%s",  
            p->ds18[p->fans[fan].sensor-1].name, spaces);
          snprintf(p->fans[fan].line2+4, lcd_wid-3, "%s%s",
            p->ds18[p->fans[fan].sen_ref-1].name, spaces);
          snprintf((p->ds18[p->fans[fan].sen_ref-1].name)+lcd_wid-7, 8, "not-yet");
        }
        report(RPT_DEBUG, "%s %02d lines: fans[%d].line2, order %03d, %s",
            __FUNCTION__, lines, fan, order[lines], p->fans[fan].line2);
#ifdef PROTOCOL_CHEATING
        iowritestohere[lines]=(int)&p->fans[fan].line2;
        report(RPT_PCHEAT, "%s   fan[%02d].line2 iowritestohere[%02d]=%d, %s",
          __FUNCTION__, fan, lines, iowritestohere[lines],
          (char *)(*((int*)iowritestohere[lines])));
#endif
        lines++;
      } // end has assocated temp
     } // end fan has LCD line(s)
     else p->fans[fan].line=NULL;
  } // end each fan
  for (temp=0; temp < p->tempcount; temp ++) {
    if ( p->ds18[temp].dowidget!=0) {  
      p->ds18[temp].line= malloc(lcd_wid+1);
      if (p->ds18[temp].line==NULL) {
        report(RPT_ERR, "%s malloc  p->ds18[].line=fail.", __FUNCTION__);
        exit(EXIT_FAILURE); 
      }
      order[lines]=(p->ds18[temp].dowidget+1)*10;
      displinenum[lines]=lines;
      orderline[lines]=p->ds18[temp].line;
      t_or_f_num[lines]=temp;
      updateline[lines]=*set_up_templine;
      snprintf(p->ds18[temp].line, lcd_wid+1, "%s%s",
          p->ds18[temp].name, spaces);
      snprintf((p->ds18[temp].line)+lcd_wid-7, 8, "not-yet");
      report(RPT_DEBUG, "%s %02d lines: ds18[%d].line , order %03d, %s",
        __FUNCTION__, lines, temp, order[lines], p->ds18[temp].line);
#ifdef PROTOCOL_CHEATING
      iowritestohere[lines]=(int)&p->ds18[temp].line;
      report(RPT_PCHEAT, "%s   ds18[%02d].line iowritestohere[%02d]=%d, %s",
        __FUNCTION__, temp,lines, iowritestohere[lines],
        (char *)(*((int*)iowritestohere[lines])));
#endif
      lines++;
      }
    else p->ds18[temp].line=NULL;
  } // end each temp 
#ifdef DO_ULT
  for (temp=0; temp < MAXTOP; temp ++) {
    if(p->topnameline[temp]!=NULL) { // if mallo line during readconf
      order[lines]=p->topnamelcd[temp]*10;
      displinenum[lines]=lines;
      orderline[lines]=p->topnameline[temp];
      t_or_f_num[lines]=temp;
      updateline[lines]=*set_up_topnameline;
      snprintf(p->topnameline[temp], lcd_wid+1, "%s%%%s",
          p->topnames[temp], spaces);
      snprintf((p->topnameline[temp])+lcd_wid-8, 9, " not-yet");
      report(RPT_DEBUG, "%s %02d lines: topname[%d],    order %03d, %s",
          __FUNCTION__, lines, temp, order[lines], p->topnameline[temp]);
#ifdef PROTOCOL_CHEATING
      iowritestohere[lines]=(int)&p->topnameline[temp];
      report(RPT_PCHEAT, "%s topnameline[%02d] iowritestohere[%02d]=%d, %s",
        __FUNCTION__, temp,lines, iowritestohere[lines],
        (char *)(*((int*)iowritestohere[lines])));
#endif
      lines++;
    } // end this topnames[temp] has a line
  } // end for each topname
#endif // DO_ULT
#ifdef DO_LM
  for (temp=0; temp < p->lmnum; temp ++) {
    if(p->lmline[temp]!=NULL) { // if mallo line in find_build_req_lm  
      order[lines]=p->lmlcd[temp]*10;
      displinenum[lines]=lines;
      orderline[lines]=p->lmline[temp];
      t_or_f_num[lines]=temp;
      updateline[lines]=*set_up_lmline;
      snprintf(p->lmline[temp], lcd_wid+1, "%s:%s%s",
          p->lmnames[temp], p->lmfeatu[temp], spaces);
      snprintf((p->lmline[temp])+lcd_wid-8, 9, " not-yet");
      report(RPT_DEBUG, "%s %02d lines: lm[%d].line,     order %03d, %s",
        __FUNCTION__,  lines, temp, order[lines], p->lmline[temp]);
#ifdef PROTOCOL_CHEATING
      iowritestohere[lines]=(int)&p->lmline[temp];
      report(RPT_PCHEAT, "%s      lmline[%02d] iowritestohere[%02d]=%d, %s",
        __FUNCTION__, temp,lines, iowritestohere[lines],
        (char *)(*((int*)iowritestohere[lines])));
#endif
      lines++;
    } // end this topnames[temp] has a line
  } // end for each topname
#endif // DO_LM
#ifdef DO_HDD
  for (temp=0; temp < p->hdd_count; temp ++) {
    if(p->hddline[temp]!=NULL) { // if mallo line during readconf
      order[lines]=p->hddlcd[temp]*10;
      displinenum[lines]=lines;
      orderline[lines]=p->hddline[temp];
      t_or_f_num[lines]=temp;
      updateline[lines]=*set_up_hddline;
      snprintf(p->hddline[temp], lcd_wid+1, "%s%s",
          p->hddname[temp], spaces);
      snprintf((p->hddline[temp])+lcd_wid-8, 9, "       x");
      report(RPT_CLICRE, "%s %02d lines: hdd[%d].line,    order %03d, %s",
        __FUNCTION__, lines, temp, order[lines], p->hddline[temp]);
#ifdef PROTOCOL_CHEATING
      iowritestohere[lines]=(int)&p->hddline[temp];
      report(RPT_PCHEAT, "%s     hddline[%02d] iowritestohere[%02d]=%d, %s",
          __FUNCTION__, temp,lines, iowritestohere[lines],
        (char *)(*((int*)iowritestohere[lines])));
#endif
      lines++;
    } // end this hdd[temp] has a line
  } // end for each hdd
#endif // DO_HDD
  
/*************************************************************
 * Add any more lines here
 * lines = # of current lines: add line [lines] then lines++
 *          order[lines]=numeric order for this line
 *      orderline[lines]=string
 * iowritestohere[lines]=address of pointer that function to
 *           update this line writes to
 *       order[MAXLINES]
 *      orderline[MAXLINES]     globals at top of Cfontz633_c.c
 *    displinenum[MAXLINES]
 * iowritestohere[MAXLINES]  
 *       MAXLINES   must be increasd for more lines 
 * Creation of actual widgets & screens is 
 * done below automatically.
 * **********************************************************/

#ifdef MY_CUSTOM_LINE
  wattsptr=(int)watts;
  snprintf(watts, lcd_wid+1, "w:%s", spaces);
  order[lines]=210;
  displinenum[lines]=lines;
  orderline[lines]=watts;
  t_or_f_num[lines]=0; //unused here
  updateline[lines]=*update_wattsline;
#ifdef PROTOCOL_CHEATING
  //iowritestohere[lines]=(int)&wattsptr;
  iowritestohere[lines]=(int)&watts;
  // wattsptr will be set te point to the widget string stored
  // by the LCDd server: writing to that directly instead of 
  // sending update widgets via socket is PROTOCOL_CHEATING
  // and keeps the socket less loaded & the packet que shorter
#endif
  lines++;
#endif // MY_CUSTOM_LINE
  
  if (lines==0){
    return 1;
  }
  // sort the lcd lines
  for (i=1; i < lines; i ++) {
    for (j=0; j < i+1; j ++) {
      if (order[j] > order[i]){
        int swaporder=order[j];
        int swaplinenum=displinenum[j];
        char * swapline=orderline[j];
        unsigned char t_or_f_tmp= t_or_f_num[j];
        void *udl_tmp=updateline[j];
#ifdef PROTOCOL_CHEATING
        int swapint=iowritestohere[j];
#endif
        
        order[j]=order[i];
        orderline[j]=orderline[i];
        displinenum[j]=displinenum[i];
        t_or_f_num[j]=t_or_f_num[i];
        updateline[j]=updateline[i];
#ifdef PROTOCOL_CHEATING
        iowritestohere[j]=iowritestohere[i];
#endif
        
        order[i]=swaporder;
        orderline[i]=swapline;
        displinenum[i]=swaplinenum;
        t_or_f_num[i]=t_or_f_tmp;
        updateline[i]=udl_tmp;
#ifdef PROTOCOL_CHEATING
        iowritestohere[i]=swapint;
#endif
      }
    }
  } // done sorting
  for (i=0; i < lines; i++){
    if ( order[i] ==  order[(i-1)] )
      report(RPT_INFO < RPT_CLICRE ? RPT_INFO : RPT_CLICRE, 
          "%s duplicate ordering %d %s; %d %s",
          __FUNCTION__, order[i],  orderline[i], order[(i-1)], orderline[(i-1)]);
  }
  for (i=0; i < lines; i++){
      report(RPT_CLICRE, "%s scn %02d l %02d  %07d  %s%s%s",
          __FUNCTION__, i/lcd_hgt, i, order[i], ( (i/lcd_hgt +2) %2 == 0 ) ? "\"" : "|",
          orderline[i], ( (i/lcd_hgt +2) %2 == 0 ) ? "\"" : "|" );
  }
  char *other[]={"fixME", "elsewhere", "listen", "ignore", "cf_log"};
  for (i=0; i < lines; i ++) { // for each line to be created
    char buf[64];
    j=i/lcd_hgt; // j is screen #, starting with 0 
    report(RPT_DEBUG, "%s line %d screen %d", __FUNCTION__, i, j);
    // when server switches to newly added screen,
    // the "listen" appears to knock out an expected success
    if ((i-lcd_hgt*j)==0){ // is 1st in new screen  
      snprintf (buf, sizeof(buf)-1, "screen_add cf_log%d\n", j);
      report(RPT_DEBUG, "%s sock_send: %d %s", __FUNCTION__, cf_dsk, buf);
      sock_send_string( cf_dsk, buf );
      wait_success(drvthis, 5, other);
      // hide all screens untill all done
      //snprintf (buf, sizeof(buf)-1, "screen_set cf_log%d priority hidden\n", j);
      //report(RPT_DEBUG, "%s sock_send: %d %s", __FUNCTION__, cf_dsk, buf);
      //sock_send_string( cf_dsk, buf );
      //wait_success(drvthis, 5, other);
    } // end is new screen
    snprintf( buf, sizeof(buf)-1, "widget_add cf_log%d line%d string\n",j, i );
    report(RPT_DEBUG, "%s sock_send: %d %s", __FUNCTION__, cf_dsk, buf);
    sock_send_string( cf_dsk, buf );
    wait_success(drvthis, 5, other);
    snprintf( buf, sizeof(buf)-1, "widget_set cf_log%d line%d 1 %d \"%s\"\n",
        j,    // screen #
        i,    // line #
        (i+lcd_hgt)%lcd_hgt+1,   // line # on this screen 
        orderline[i] );  // text
    report(RPT_DEBUG, "%s sock_send: %d %s", __FUNCTION__, cf_dsk, buf);
    sock_send_string( cf_dsk, buf );
    wait_success(drvthis, 5, other);
  } // end for each line to be created
#ifdef PROTOCOL_CHEATING
  Client * cf_client_cheat=NULL;
  int tryme=1 ;
  do {
    // changed from sock to sock -1 CVS-stable-0-5-x-20060415 ?
    // if initing during boot & things are slow in server, we want sock, not sock -1 ????
    cf_client_cheat= clients_find_client_by_sock (tryme );
    if ( cf_client_cheat!=NULL ){ 
      report(RPT_ERR, "%s clients_find_client_by_sock (connect %d / find %d) = %s   %d",
          __FUNCTION__, cf_dsk, tryme, (cf_client_cheat==NULL) ? "null" : cf_client_cheat->name, cf_client_cheat); 
      if (strncmp (cf_client_cheat->name, "CF633log", 8 ) != 0 ){
        cf_client_cheat=NULL;
        tryme++;
      }
    }
    else tryme++;
  } while ((cf_client_cheat==NULL) && (tryme != 255)) ;
  report(RPT_PCHEAT, "%s clients_find_client_by_sock (connect %d / find %d) = %s  sock or sock -1 ? %d",
        __FUNCTION__, cf_dsk, cf_client_cheat->sock, cf_client_cheat->name, cf_client_cheat);
  Screen *cf_screen_cheat=NULL;
  for (i=0; i < lines; i ++) { // for each line to be created
    char buf[64];
    j=i/lcd_hgt; // j is screen #, starting with 0 
    //report(RPT_PCHEAT, "%s PROTOCOL_CHEATING line %d screen %d", __FUNCTION__, i, j);
    // when server switches to newly added screen,
    // the "listen" appears to knock out an expected success
    if ((i-lcd_hgt*j)==0){ // is 1st in new screen  
      snprintf (buf, sizeof(buf)-1, "cf_log%d", j);
      cf_screen_cheat=NULL;  
      do {
        cf_screen_cheat=client_find_screen (cf_client_cheat, buf);
        // is hanging when cf_screen_cheat=0, screen not found ?
        // wait_success failed ?
        usleep(10000);
      } while (cf_screen_cheat==NULL);
      report(RPT_PCHEAT, "%s (Screen *)cf_screen_cheat=%d",
          __FUNCTION__, cf_screen_cheat);
    } // end is new screen
    Widget * cf_widget_cheat=NULL;
    snprintf (buf, sizeof(buf)-1, "line%d", i);
    do {
      cf_widget_cheat=screen_find_widget (cf_screen_cheat, buf);
      if (cf_widget_cheat==NULL) usleep(10000); // idle until server creates cf_widget_cheat
    } while ( cf_widget_cheat==NULL);
    while (cf_widget_cheat->text==NULL) {
      usleep(10000); // idle until server creates cf_widget_cheat->text
      report(RPT_PCHEAT, "%s cf_widget_cheat line %02d until server creates cf_widget_cheat->text",
          __FUNCTION__, i);
    }
    while (strncmp(cf_widget_cheat->text, orderline[i], lcd_wid)) {
      usleep(10000);  // idle untill server fills in cf_widget_cheat->text
      report(RPT_PCHEAT, "%s cf_widget_cheat line %02d not NULL but not filled in SUSPICION ?", __FUNCTION__, i);
    }
    report(RPT_PCHEAT, "%s cf_widget_cheat=%d line %02d    |%s|",
        __FUNCTION__, cf_widget_cheat, i, cf_widget_cheat->text);
    // set pointer from private data fan[] ds[] to point to widget.text
    *((int *)iowritestohere[i])=(int)(char *)cf_widget_cheat->text;
    report(RPT_PCHEAT, "%s PCHEAT *((int *)iowritestohere[%02d])=  %d     |%s|",
        __FUNCTION__, i, iowritestohere[i], (char *)(*((int*)iowritestohere[i])) );
  // sleep (2);  // debugging delay @ end of get cf_widget_cheat
  } // end for each line to be created
#endif // PROTOCOL_CHEATING
  report(RPT_CLICRE, "%s creating LCDd %d lines in %d screens.",
      __FUNCTION__, lines, (lines-1)/lcd_hgt+1);
/*  report(RPT_CLICRE, "%s unhiding screens...",__FUNCTION__);
  for (i=0; i < lines; i+=lcd_hgt) {
    j=i/lcd_hgt; // j is screen #, starting with 0 
    char buf[64];
    //  all done
    snprintf (buf, sizeof(buf)-1, "screen_set cf_log%d priority info\n", j);
    report(RPT_DEBUG, "%s sock_send: %d %s", __FUNCTION__, cf_dsk, buf);
    sock_send_string( cf_dsk, buf );
    //wait_success(drvthis, 0, NULL);
    wait_success(drvthis, 5, other);
  }
  for (i=0; i < lines; i+=lcd_hgt) {
    //wait_success(drvthis, 0, NULL);
  }
*/  return 1;  
} // end cli_scrn

int read_connect_string(Driver * drvthis)
{
  char buf[8192];
  int len=-1;
  int a;
  char * argv[20];
  int argc;
  short received = 0;
  short timeout = 50; /* Give the server 5 secs to respond */

  buf[0] = 0;

  while( !received && timeout > 0) {
    len = sock_recv_string ( cf_dsk, buf, sizeof(buf));
    if( len == 0 ) {
      usleep(10000); // read_connect_string len==0, try again
      timeout --;
    } else {
      received = 1;
    }
  }
  report(RPT_CLICRE, "%s read_connect_string: %s", __FUNCTION__, buf);

  if( len <= 0 ) {
    report(RPT_ERR, "%s Did not receive LCDd connect response.", __FUNCTION__);
    return -1;
  }

  argc = split( buf, ' ', argv, 20 );
  for (a = 1; a < argc; a++) {
    if (0 == strcmp (argv[a], "wid"))
      lcd_wid = atoi (argv[++a]);
    else if (0 == strcmp (argv[a], "hgt"))
      lcd_hgt = atoi (argv[++a]);
  }
  if( argc <= 0 || strcmp( argv[0], "connect" ) != 0
      || lcd_wid == 0 || lcd_wid == 0 ) {
          report(RPT_ERR, "%s Received invalid LCDd connect response.", __FUNCTION__);
    return -1;
  }
  report(RPT_CLICRE, "%s Client Connect Received lcd_wid=%d  lcd_hgt=%d.",
        __FUNCTION__, lcd_wid, lcd_hgt);
  return 1;
} // end of read_connect_string

int split( char * str, char delim, char * parts[], int maxparts )
/* Splits a string into parts, to which pointers will be returned in &parts.
 * The return value is the number of parts.
 * maxparts is the maximum number of parts returned. If more parts exist
 * they are (unsplit) in the last part.
 * The parts are split at the character delim.
 * No new space will be allocated, the string str will be mutated !
 */
{
  char * p1 = str;
  char * p2;
  int part_nr = 0;

  /* Find the delim char to end the current part */
  while( part_nr < maxparts - 1 && (p2 = strchr( p1, delim )) ) {

    /* subsequent parts... */
    *p2 = 0;
    parts[part_nr] = p1;

    p1 = p2 + 1; /* Just after the delim char */
    part_nr ++;
  }
  /* and the last part... */
  parts[part_nr] = p1;
  part_nr ++;

  return part_nr;
} // end of split

#ifdef DO_ULT
static void
set_up_topnameline (Driver *drvthis, unsigned char topn)
{
    PrivateData * p = drvthis->private_data;
#ifdef CLIENT_CHILD
  if (p->prettylcd!=0) {
    // could possibly run store tops (p->height*8) times every second
    // instead of once every p->syslogsensors seconds
    store_tops(drvthis);
  }
#endif // CLIENT_CHILD
  if (p->topnameline[topn]!=NULL) { 
    // if this topname[topn] has a string for log / lcd line 
    snprintf(p->topnameline[topn]+strlen(p->topnameline[topn])-7, 8, "%f", 
        p->topnameval[topn] );
    report(RPT_CLIUP, "%s update p->topnameline[%d]: %s",
        __FUNCTION__, topn, p->topnameline[topn]);
  }
  else {
    report(RPT_CLIUP, "%s update p->topnameline[%d]: NULL",
        __FUNCTION__, topn);
  }
} // end set_up_topnameline
#endif // DO_ULT

#ifdef DO_LM
static void
set_up_lmline (Driver *drvthis, unsigned char lmn)
{
  PrivateData * p = drvthis->private_data;
    //report(RPT_ERR, "%s update 0 p->lmline[%d] %f: |%s| %d",
    //    __FUNCTION__, lmn, (float)p->lmval[lmn],  p->lmline[lmn], strlen(p->lmline[lmn]));
  if (p->prettylcd!=0) {
      // could possibly run sensors_get_feature (p->height*8) times every second
      // instead of once every p->syslogsensors seconds
    double value;
    if (sensors_get_feature(*p->chipn[lmn], p->featn[lmn]->number, &value)) {
      value=666;  // error on read
    }
    p->lmval[lmn]=(float)value;
  }
  if (p->lmline[lmn]!=NULL) { 
    // if this lm[lmn] has a string for log / lcd line 
    snprintf(p->lmline[lmn]+strlen(p->lmline[lmn])-7, 8, " %f", 
        (float)p->lmval[lmn] );
    //report(RPT_ERR, "%s update 1 p->lmline[%d] %f: |%s| %d ",
    //    __FUNCTION__, lmn, (float)p->lmval[lmn],  p->lmline[lmn], strlen(p->lmline[lmn]));
    // why no ending / heartbeat space for this ?
    p->lmline[lmn][strlen(p->lmline[lmn])-1]=' ';
    //p->lmline[lmn][lcd_wid]=' ';
    report(RPT_CLIUP, "%s update p->lmline[%d] %f: |%s| %d",
        __FUNCTION__, lmn, (float)p->lmval[lmn],  p->lmline[lmn], strlen(p->lmline[lmn]));
  }
  else {
    report(RPT_CLIUP, "%s update p->lmline[%d]: NULL",
        __FUNCTION__, lmn);
  }
} // end set_up_lmline
#endif // DO_LM

#ifdef DO_HDD
static void
set_up_hddline (Driver *drvthis, unsigned char hdd)
{
    PrivateData * p = drvthis->private_data;
  if (p->hddline[hdd]!=NULL) {
#ifdef CLIENT_CHILD
    if (p->prettylcd!=0) {
      // could possibly run hddtemp_temps (p->height*8) times every second
      // instead of once every p->syslogsensors seconds
      hddtemp_temps(drvthis);
    }
#endif // CLIENT_CHILD
    hddtemp_temps(drvthis);
    // if this has a string for log / lcd line 
    snprintf(p->hddline[hdd]+strlen(p->hddline[hdd])-strlen(p->hddtemp[hdd]) -3,
        strlen(p->hddtemp[hdd])+4, "%s    ", p->hddtemp[hdd] );
    report(RPT_CLIUP, "%s update p->hddline[%d]: %s",
        __FUNCTION__, hdd,   p->hddline[hdd]);
  }
  else {
    report(RPT_CLIUP, "%s update p->hddline[%d]: NULL",
        __FUNCTION__, hdd);
  }
} // end set_up_hddline
#endif // DO_HDD

static void
set_up_templine (Driver *drvthis, unsigned char temp)
{
    PrivateData * p = drvthis->private_data; 
  if (p->ds18[temp].line!=NULL) { 
    // if this temp has a string for log / lcd line 
    snprintf(p->ds18[temp].line+strlen(p->ds18[temp].line)-7, 8, "%f", 
        (float)p->ds18[temp].ds18 /16 );
    report(RPT_CLIUP, "%s update p->ds18[%d].line %d: %s",
        __FUNCTION__, temp, &(p->ds18[temp].line) , p->ds18[temp].line);
  }
  else {
    report(RPT_CLIUP, "%s update p->ds18[%d].line %d: NULL",
        __FUNCTION__, temp, &(p->ds18[temp].line));
  }
} // end set_up_templine

static void
set_up_fanline (Driver *drvthis, unsigned char fan)
{
    PrivateData * p = drvthis->private_data; 
  if (p->fans[fan].line!=NULL) { 
    // if this fan has a string for log / lcd line 
    snprintf((p->fans[fan].line)+strlen(p->fans[fan].line)-7, 8, "%f", 
        p->fans[fan].rpm_rpt );
    report(RPT_CLIUP, "%s update p->fans[%d].line  %d: %s",
        __FUNCTION__, fan, &(p->fans[fan].line), p->fans[fan].line);
  }
  else {
    report(RPT_CLIUP, "%s update p->fans[%d].line  %d: NULL",
        __FUNCTION__, fan, &(p->fans[fan].line));
  }
  if ((p->fans[fan].line2)!=NULL) {
    // if this fan has an asociated temp widget
    if (p->fans[fan].sen_ref==0) { // if simple temp 
       snprintf((p->fans[fan].line2)
         +strlen(p->fans[fan].line2)-7, 8, "%f",(float)p->ds18[p->fans
         [fan].sensor-1].ds18/16 );
    }
    else { // relative temp
       snprintf( (p->fans[fan].line2) +strlen(p->fans[fan].line2)-7, 8, "%f",
        (float)(p->ds18[p->fans[fan].sensor-1].ds18
         -p->ds18[p->fans[fan].sen_ref-1].ds18) /16 );
    }
    report(RPT_CLIUP, "%s update p->fans[%d].line2 %d: %s",
        __FUNCTION__, fan, &(p->fans[fan].line2), p->fans[fan].line2);
  } 
  else {
    report(RPT_CLIUP, "%s update p->fans[%d].line2 %d: NULL",
        __FUNCTION__, fan, &(p->fans[fan].line2));
  }
  // end has a temperature line2
} // end set_up_fanline

#endif // CLIENT_CHILD 
 
/* EOF */
